/****
 *
 * Exemplo: Consultando um Bit
 *
 * Seo: B.8
 *
 * Descrio: Verifica se um bit de um valor inteiro est ligado
 *
 ****/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'
 *
 ****/
int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeInteiro(): L um nmero inteiro via teclado
 *              e deixa o buffer intacto
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero inteiro lido
 *
 ****/
int LeInteiro(void)
{
   int num, /* O nmero lido */
       teste, /* Valor retornado por scanf() */
       nResto = 0; /* Nmero de caracteres */
                   /* excedentes           */

   /* Desvia para c se o valor for invlido */
inicio:
      /* Tenta ler um valor vlido */
   teste = scanf("%d", &num);

      /* Se no ocorreu erro de leitura ou de final */
      /* de arquivo, h caracteres remanescentes    */
      /* que precisam ser removidos                 */
   if (teste != EOF) {
      nResto = LimpaBuffer();
   }

      /* Enquanto o valor retornado por scanf()  */
      /* no indicar que um valor vlido foi     */
      /* lido continua tentando obter esse valor */
   while(teste != 1) {
      printf("\a\n\t>>> O valor digitado e' invalido. "
                      "Tente novamente\n\t> " );
      goto inicio; /* No causa dano algum */
   }

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf("\t>>> %d caracteres foram descartados\n", nResto);
   }

      /* O valor retornado certamente  vlido */
   return num;
}

/****
 *
 * RepresentacaoBinaria(): Apresenta no meio de sada a
 *                         representao binria de um valor
 *                         do tipo int
 *
 * Parmetros:
 *      numero (entrada): valor inteiro cuja representao
 *                        binria ser apresentada
 *
 * Retorno: Nada.
 *
 ****/
void RepresentacaoBinaria(int numero)
{
   int           i, bit, numeroDeBits;
   unsigned int  mascara; /* A mscara deve ser unsigned */

       /* Cada byte contm um nmero de bits igual a CHAR_BIT */
   numeroDeBits  = CHAR_BIT*sizeof(int);

       /* Coloca 1 na posio mais  esquerda da  */
       /* mscara. Todos os outros bits tero 0s. */
   mascara = 0x1 << (numeroDeBits - 1);

     /* Exibe cada bit a partir da esquerda */
   for (i = 1; i <= numeroDeBits; i++) {
      bit = (numero & mascara) ? 1 : 0;

         /* Apresenta um bit na tela */
      printf("%d", bit);

         /* Separa sequncias de bytes com espaos em branco */
         /* para melhor visualizao da representao        */
      if (i % CHAR_BIT == 0)
         printf("  ");

         /* Move bit com valor 1 da mscara */
         /* uma posio para a direita      */
      mascara >>= 1;
   }
}

/****
 *
 * TestaBit(): Verifica se um bit de um valor inteiro est ligado
 *             ou desligado
 *
 * Parmetros:
 *      valor (entrada) - a valor que ter seu bit testado
 *      pos (entrada) - posio do bit que ser testado
 *
 * Retorno: 1, se o bit estiver ligado; 0, em caso contrrio
 *
 ****/
int TestaBit(int valor, int pos)
{
   return 1 & valor >> pos;
}

/****
 *
 * main(): Testa a funo TestaBit()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int numero, pos, resposta;

   printf("\n>>> Introduza um valor inteiro: ");
   numero = LeInteiro();

   printf("\n>>> Representacao binaria de %d:\n\t", numero);
   RepresentacaoBinaria(numero);

   printf("\n\n>>> Bit que deseja consultar: ");
   pos = LeInteiro();

   if (pos < 0 || pos >= (int) (8*sizeof(int))) {
      printf("\nIndice invalido!!!\n");
      return 1;
   }

   resposta = TestaBit(numero, pos);

   printf( "\n>>> O bit na posicao %d esta' %s\n\t",
           pos, resposta ? "ligado" : "desligado");

   return 0;
}

/******************* Exemplo de Execuo ******************

>>> Introduza um valor inteiro: 234

>>> Representacao binaria de 234:
        00000000  00000000  00000000  11101010

>>> Bit que deseja consultar: 9

>>> O bit na posicao 9 esta' desligado

**********************************************************/
