/****
 *
 * Exemplo: Disperso por Mistura
 *
 * Seo: B.10
 *
 * Descrio: Implementao de uma funo disperso usando o
 *            mtodo de mistura
 *
 ****/

#include <stdio.h>   /* printf() */
#include <stdlib.h>  /* calloc() */
#include <limits.h>  /* CHAR_BIT */

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'.
 *
 ****/
int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeInteiro(): L um nmero inteiro via teclado e deixa o buffer
 *              intacto
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero inteiro lido
 *
 ****/
int LeInteiro(void)
{
   int num, /* O nmero lido */
       teste, /* Valor retornado por scanf() */
       nResto = 0; /* Nmero de caracteres excedentes */

   /* Desvia para c se o valor for invlido */
inicio:
      /* Tenta ler um valor vlido */
   teste = scanf("%d", &num);

      /* Se no ocorreu erro de leitura ou de final */
      /* de arquivo, h caracteres remanescentes    */
      /* que precisam ser removidos                 */
   if (teste != EOF) {
      nResto = LimpaBuffer();
   }

      /* Enquanto o valor retornado por scanf()  */
      /* no indicar que um valor vlido foi     */
      /* lido continua tentando obter esse valor */
   while(teste != 1) {
      printf("\a\n\t>>> O valor digitado e' invalido. "
                      "Tente novamente\n\t> " );
      goto inicio; /* No causa dano algum */
   }

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf("\t>>> %d caracteres foram descartados\n", nResto);
   }

      /* O valor retornado certamente  vlido */
   return num;
}

/****
 *
 * RepresentacaoBinaria(): Apresenta no meio de sada a
 *                         representao binria de um valor
 *                         do tipo int
 *
 * Parmetros:
 *      numero (entrada): valor inteiro cuja representao
 *                        binria ser apresentada
 *
 * Retorno: Nada
 *
 ****/
void RepresentacaoBinaria(int numero)
{
   int           i, bit, numeroDeBits;
   unsigned int  mascara; /* A mscara deve ser unsigned aqui */

       /* Cada byte contm um nmero de bits igual a CHAR_BIT */
   numeroDeBits  = CHAR_BIT*sizeof(int);

       /* Coloca 1 na posio mais  esquerda da  */
       /* mscara. Todos os outros bits tero 0s. */
   mascara = 0x1 << (numeroDeBits - 1);

     /* Exibe cada bit a partir da esquerda do nmero (parmetro) */
   for (i = 1; i <= numeroDeBits; i++) {
      bit = (numero & mascara) ? 1 : 0;

         /* Apresenta um bit na tela */
      printf("%d", bit);

         /* Separa bits em sequncias de oito para */
         /* melhor visualizao da representao   */
      if (i % 8 == 0)
         printf("  ");

         /* Move bit com valor 1 da mscara */
         /*  uma posio para a direita     */
      mascara >>= 1;
   }
}

/****
 *
 * DispersaoMisturaJunior(): Calcula o valor de disperso de uma
 *                           chave usando o mtodo de mistura
 *
 * Parmetros:
 *      chave (entrada) - a chave
 *
 * Retorno: O valor de disperso calculado
 *
 * Observaes: Essa verso  dirigida para iniciantes em
 *              programao de baixo nvel em C
 *
 ****/
unsigned DispersaoMisturaJunior(int chave)
{
   unsigned mascara[4] = { 0xFF000000, /* 11111111 00000000 00000000 00000000 */
                           0x00FF0000, /* 00000000 11111111 00000000 00000000 */
                           0x0000FF00, /* 00000000 00000000 11111111 00000000 */
                           0x000000FF, /* 00000000 00000000 11111111 00000000 */
                         };
   char     octetos[4] = { (chave & mascara[0]) >> 3*8,
                           (chave & mascara[1]) >> 2*8,
                           (chave & mascara[2]) >> 1*8,
                           (chave & mascara[3]) >> 0*8
                         };

   unsigned i, xor1, xor2;

   for (i = 0; i < 4; ++i) {
      printf("\n\n>>> Octeto %d:\n\t", i + 1);
      RepresentacaoBinaria(octetos[i]);
   }

   xor1 = octetos[0] ^ octetos[3];
   xor2 = octetos[1] ^ octetos[2];

   return xor1 ^ xor2;
}

/****
 *
 * DispersaoMisturaMedio(): Calcula o valor de disperso de uma
 *                           chave usando o mtodo de mistura
 *
 * Parmetros:
 *      chave (entrada) - a chave
 *
 * Retorno: O valor de disperso calculado
 *
 * Observaes: Essa verso  dirigida para nvel intermedirio
 *              em programao de baixo nvel em C
 *
 ****/
unsigned DispersaoMisturaMedio(int chave)
{
   unsigned mascara = 0xFF000000, /* 11111111 00000000 00000000 00000000 */
            i, xor1, xor2;
   char     octetos[4];

   for (i = 0; i < 4; ++i) {
      octetos[i] = (chave & mascara) >> ((3 - i)*8);

      printf("\n\n>>> Mascara: %X", mascara);
      printf("\n>>> Octeto %d: ", i + 1);
      RepresentacaoBinaria(octetos[i]);

      mascara = mascara >> 8;
   }

   xor1 = octetos[0] ^ octetos[3];
   xor2 = octetos[1] ^ octetos[2];

   return xor1 ^ xor2;
}

/****
 *
 * DispersaoMisturaSenior(): Calcula o valor de disperso de uma
 *                           chave usando o mtodo de mistura
 *
 * Parmetros:
 *      chave (entrada) - a chave
 *
 * Retorno: O valor de disperso calculado
 *
 * Observaes: Essa verso  dirigida para nerds
 *              em programao de baixo nvel em C
 *
 ****/
unsigned DispersaoMisturaSenior(int chave)
{
   unsigned mascara = 0xFF000000,
            dispersao = 0,
           *octetos; /* Um octeto armazena um byte ou um char */
   size_t   i, tam;

   tam = sizeof(chave);

   octetos = calloc(tam, tam);

   for (i = 0; i < tam; ++i) {
      octetos[i] = (chave & mascara) >> ((tam - 1 - i)*CHAR_BIT);

      printf("\n\n>>> Mascara: %X", mascara);
      printf("\n>>> Octeto %d: ", i + 1);
      RepresentacaoBinaria(octetos[i]);

      mascara = mascara >> CHAR_BIT;
   }

   for (i = 0; i < tam/2; ++i)
      dispersao = dispersao ^ octetos[i] ^ octetos[tam - 1 - i];

   return dispersao;
}

/****
 *
 * main(): Testa as funes de dispersao por mistura
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int chave,
       dispersao;

   while (1) {
      printf("\n\n>>> Introduza um valor inteiro: ");
      chave = LeInteiro();

      if (chave <= 0)
         break;

      printf("\n>>> Representacao binaria de %d:\n\t", chave);
      RepresentacaoBinaria(chave);

      dispersao = DispersaoMisturaJunior(chave);

      printf( "\n\n>>> Valor de dispersao (Junior) de %d: %d",
              chave, dispersao );

      printf("\n\n>>> Representacao binaria de %d:\n\t", dispersao);
      RepresentacaoBinaria(dispersao);

      dispersao = DispersaoMisturaMedio(chave);

      printf( "\n\n>>> Valor de dispersao (Medio) de %d: %d",
              chave, dispersao );

      printf("\n\n>>> Representacao binaria de %d:\n\t", dispersao);
      RepresentacaoBinaria(dispersao);

      dispersao = DispersaoMisturaSenior(chave);

      printf( "\n\n>>> Valor de dispersao (Senior) de %d: %d",
              chave, dispersao );

      printf("\n\n>>> Representacao binaria de %d:\n\t", dispersao);
      RepresentacaoBinaria(dispersao);
   }

   printf("\n\n>>> Bye, bye\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

>>> Introduza um valor inteiro: 205405

>>> Representacao binaria de 205405:
        00000000  00000011  00100010  01011101

>>> Octeto 1:
        00000000  00000000  00000000  00000000

>>> Octeto 2:
        00000000  00000000  00000000  00000011

>>> Octeto 3:
        00000000  00000000  00000000  00100010

>>> Octeto 4:
        00000000  00000000  00000000  01011101

>>> Valor de dispersao (Junior) de 205405: 124

>>> Representacao binaria de 124:
        00000000  00000000  00000000  01111100

>>> Mascara: FF000000
>>> Octeto 1: 00000000  00000000  00000000  00000000

>>> Mascara: FF0000
>>> Octeto 2: 00000000  00000000  00000000  00000011

>>> Mascara: FF00
>>> Octeto 3: 00000000  00000000  00000000  00100010

>>> Mascara: FF
>>> Octeto 4: 00000000  00000000  00000000  01011101

>>> Valor de dispersao (Medio) de 205405: 124

>>> Representacao binaria de 124:
        00000000  00000000  00000000  01111100

>>> Mascara: FF000000
>>> Octeto 1: 00000000  00000000  00000000  00000000

>>> Mascara: FF0000
>>> Octeto 2: 00000000  00000000  00000000  00000011

>>> Mascara: FF00
>>> Octeto 3: 00000000  00000000  00000000  00100010

>>> Mascara: FF
>>> Octeto 4: 00000000  00000000  00000000  01011101

>>> Valor de dispersao (Senior) de 205405: 124

>>> Representacao binaria de 124:
        00000000  00000000  00000000  01111100

>>> Bye, bye

**********************************************************/
