/****
 *
 * Exemplo: Funes de Disperso Prontas: Funo JSW
 *
 * Seo: B.5
 *
 * Descrio: Implementao da funo de disperso JSW
 *
 ****/

#include <stdio.h>   /* printf()         */
#include <stdlib.h>  /* srand() e rand() */
#include <time.h>    /* time()           */
#include <limits.h>  /* UCHAR_MAX        */

/****
 *
 * CriaTabela(): Cria uma tabela de valores aleatrios
 *
 * Parmetros:
 *      tab[] (sada) - array que representa a tabela
 *      tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 ****/
static void CriaTabela(int tab[], int tam)
{
   int i;

      /* Inicia o gerador de nmeros aleatrios */
   srand((unsigned)time(NULL));

      /* Cria a tabela de valores aleatrios */
   for (i = 0; i < tam; ++i) {
      tab[i] = rand();
   }
}

/****
 *
 * DispersaoJSW(): Implementao da funo de disperso JSW
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoJSW(const char *chave)
{
   unsigned int dispersao = 16777551; /* Nmero mgico */
   static int   primeiraChamada = 1,
                tabela[UCHAR_MAX + 1];

   if (primeiraChamada) {
      CriaTabela(tabela, UCHAR_MAX + 1);
      primeiraChamada = 0;
   }

   while (*chave)
      dispersao = (dispersao << 1 | dispersao >> 31) ^
                  tabela[(size_t)(*chave++)];

   return dispersao;
}

/****
 *
 * main(): Testa a funo DispersaoJSW()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   printf( "\n   >>>> Usando Dispersao DispersaoJSW <<<<\n");
   printf( "\n>>> Dispersao de \"bola\" = %u",
           DispersaoJSW("bola") );
   printf( "\n>>> Dispersao de \"loba\" = %u\n",
           DispersaoJSW("loba") );

   printf( "\n>>> Dispersao de \"bola\" = %u",
           DispersaoJSW("bola") );
   printf( "\n>>> Dispersao de \"loba\" = %u\n",
           DispersaoJSW("loba") );

   printf("\n>>> Dispersao de \"anticonstitucionalissimamente\""
          " = %u\n", DispersaoJSW("anticonstitucionalissimamente"));

   return 0;
}

/****************** Resultado do Programa *****************

   >>>> Usando Dispersao DispersaoJSW <<<<

>>> Dispersao de "bola" = 268690024
>>> Dispersao de "loba" = 268643032

>>> Dispersao de "bola" = 268690024
>>> Dispersao de "loba" = 268643032

**********************************************************/
