#include <stdio.h>

int BuscaBinaria(int ar[], int n, int chave)
{
   int inf, /* Limite inferior da busca */
       sup, /* Limite superior da busca */
       meio, /* Meio do intervalo */
       cont = 0; /* Resultado da comparao de duas chaves */

      /* limites inferior e superior iniciais */
   inf = 0;
   sup = n - 1;

      /* Efetua a busca binria */
   while (inf <= sup) {
      ++cont;

         /* Calcula o meio do intervalo */
      meio = (inf + sup)/2;

         /* Verifica se o elemento se     */
         /* encontra no meio do intervalo */
      if (chave == ar[meio]) {
         printf("\n\t>>> Numero de comparacoes = %d\n", cont);
         return meio; /* Elemento encontrado */
      }

         /* Ajusta o intervalo de busca */
      if (chave < ar[meio]) {
         sup = meio - 1;
      } else {
         inf = meio + 1;
      }
   }

   printf("\n\t>>> Numero de comparacoes = %d\n", cont);

   return -1; /* Elemento no foi encontrado */
}

/****
 *
 * BuscaInterpolacao(): Efetua uma busca por interpolao numa
 *                      tabela de busca implementada como lista
 *                      indexada ordenada
 *
 * Parmetros:
 *      tabela (entrada) - tabela na qual ser efetuada a busca
 *      chave (entrada) - a chave de busca
 *
 * Retorno: ndice do primeiro elemento encontrado na lista
 *          que casa com a chave de busca. Um valor negativo,
 *          se a chave no for encontrada.
 *
 ****/
int BuscaInterpolacao(int ar[], int n, int chave)
{
   int inf, /* Limite inferior da busca */
       sup, /* Limite superior da busca */
       meio, /* Meio do intervalo */
       chaveInf, chaveSup, /* Chaves dos limites inferior e superior */
       cont = 0;

      /* limites inferior e superior iniciais */
   inf = 0;
   sup = n - 1;

      /* Efetua a busca por interpolao */
   while (inf <= sup) {
      ++cont;

         /* Converte as chaves nos extremos do intervalo */
      chaveInf = ar[inf];
      chaveSup = ar[sup];

         /* Calcula o meio do intervalo */
      meio = inf + (sup - inf)*((chave - chaveInf) /
                   (chaveSup - chaveInf));

         /* Verifica se o elemento se encontra */
         /* no meio do intervalo               */
      if (chave == ar[meio]) {
         printf("\n\t>>> Numero de comparacoes = %d\n", cont);
         return meio; /* Elemento encontrado */
      }

         /* Ajusta o intervalo de busca */
      if (chave < ar[meio])
         sup = meio - 1;
      else
         inf = meio + 1;
   }

   printf("\n\t>>> Numero de comparacoes = %d\n", cont);

   return -1; /* Elemento no foi encontrado */
}

int main()
{
   int  ar[] = {-5, 1, 3, 8, 8, 8, 10, 11, 13, 21, 22, 25},
        tam = sizeof(ar)/sizeof(ar[0]),
        pos;

   printf("\n*** Busca binaria: 8 ***\n");
   pos = BuscaBinaria(ar, tam, 8);
   printf("\n\t>>> pos = %d\n", pos);

   printf("\n*** Busca por interpolacao: 8 ***\n");
   pos = BuscaInterpolacao(ar, tam, 8);
   printf("\n\t>>> pos = %d\n", pos);

   printf("\n*** Busca binaria: 4 ***\n");
   pos = BuscaBinaria(ar, tam, 4);
   printf("\n\t>>> pos = %d\n", pos);

   printf("\n*** Busca por interpolacao: 4 ***\n");
   pos = BuscaInterpolacao(ar, tam, 4);
   printf("\n\t>>> pos = %d\n", pos);

   return 0;
}

/****************************************************************

*** Busca binaria: 8 ***

        >>> Numero de comparacoes = 1

        >>> pos = 5

*** Busca por interpolacao: 8 ***

        >>> Numero de comparacoes = 4

        >>> pos = 3

*** Busca binaria: 4 ***

        >>> Numero de comparacoes = 3

        >>> pos = -1

*** Busca por interpolacao: 4 ***

        >>> Numero de comparacoes = 4

        >>> pos = -1

*****************************************************************/
