/****
 *
 * Ttulo: Checando rvores binrias de busca de modo incorreto
 *
 * Exerccios de Reviso: Questo 31
 *
 ****/

/*********************** Includes *************************/

#include<stdio.h>
#include<stdlib.h>

/****************** Definies de Tipos *******************/

   /* Tipo de n e tipo de ponteiro  */
   /* para n de uma rvore binria  */
typedef struct no {
           struct no *esquerda;
           int        conteudo;
           struct no *direita;
        } tNo, *tArvoreBB;

/***************** Definies de Funes ******************/

/****
 *
 * EhArvoreDeBuscaErrada(): Tenta verificar se uma rvore binria
 *                           uma rvore binria DE BUSCA
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore
 *      compara (entrada) - ponteiro para uma funo que
 *                          compara contedos dos ns de
 *                          acordo com o mesmo critrio de
 *                          comparao de qsort()
 *
 * Retorno: 1, se a rvore for uma rvore binria de busca;
 *          0, em caso contrrio
 *
 ****/
int EhArvoreDeBuscaErrada(tNo *p)
{
   int compEsquerda, /* Resultado da comparao do contedo */
                     /* de um n com o contedo de seu      */
                     /* filho esquerdo e com...             */
       compDireita;  /* ...o contedo de seu filho direito  */

      /* Se a rvore estiver vazia, ela  de busca */
   if (!p)
      return 1;

      /* Obtm o resultado da comparao do contedo do   */
      /* n corrente com o contedo de seu filho esquerdo */
   compEsquerda = !p->esquerda || p->esquerda->conteudo <
                                  p->conteudo;

      /* Obtm o resultado da comparao do contedo do  */
      /* n corrente com o contedo de seu filho direito */
   compDireita = !p->direita || p->direita->conteudo > p->conteudo;

      /* Se os critrios de comparao do contedo do n     */
      /* corrente com os contedos dos seus filhos foram     */
      /* satisfeitos, a rvore ser de busca se suas         */
      /* subrvores esquerda e direita forem tambm de busca */
   if (compEsquerda && compDireita)
      return EhArvoreDeBuscaErrada(p->esquerda) &&
             EhArvoreDeBuscaErrada(p->direita);
   else
      return 0;

      /* No deve chegar at aqui, mas agrada compilador de C++ */
   return 0;
}

int EhArvoreDeBusca(tNo *p)
{
   static tNo *antecessor = NULL;

      /* Efetua um caminhamento em ordem infixa guardando */
      /* o valor do n antecessor imediato do n corrente */
   if (p) {
         /* Verifica se a subrvore esquerda  de busca */
      if (!EhArvoreDeBusca(p->esquerda))
         return 0;

         /* Se o contedo do n corrente for menor do */
         /* que o contedo do n antecessor, no se   */
         /* trata de rvore de busca                  */
      if (antecessor && p->conteudo < antecessor->conteudo)
         return 0;

         /* O antecessor passa a ser o n corrente */
      antecessor = p;

         /* Verifica se a subrvore direita  de busca */
      return EhArvoreDeBusca(p->direita);
   }

      /* Se ainda no houve retorno, a rvore foi aprovada */
   return 1;
}

/****
 *
 * NoNovo(): Cria um n de uma rvore binria com o contedo
 *           especificado e subrvores vazias
 *
 * Parmetros:
 *      conteudo (entrada) - o contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
tNo *NoNovo(int conteudo)
{
   tNo *pNo = malloc(sizeof(tNo));

      /* Verifica se o n foi alocado */
   if (!pNo) {
      printf("\nUm no' da arvore nao foi alocado\n");
      exit(1);
   }

      /* Recheia o n */
   pNo->conteudo = conteudo;
   pNo->esquerda = pNo->direita = NULL;

   return pNo;
}

/****
 *
 * main(): Cria um rvore binria e testa a funo
 *         EhArvoreDeBuscaErrada()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   tNo *raiz = NoNovo(12); /* Cria a raiz da rvore */

      /* Insere os demais ns na rvore */
   raiz->esquerda = NoNovo(10);
   raiz->direita = NoNovo(13);
   raiz->esquerda->esquerda = NoNovo(7);
   raiz->esquerda->direita = NoNovo(11);

      /* Verifica se a rvore binria  de busca */
   if (EhArvoreDeBuscaErrada(raiz))
      printf("\n\t>>> A arvore e' de busca\n");
   else
      printf("\n\t>>> A arvore NAO e' de busca\n");

   return 0;
}

/****************** Resultado do Programa *****************

        >>> A arvore e' de busca

**********************************************************/
