/****
 *
 * Exemplo: RadixSort
 *
 * Seo: 11.4.3
 *
 ****/

/*********************** Includes *************************/

#include <string.h>  /* Funes strxxx() */
#include <stdio.h>   /* Funes de E/S   */
#include <stdlib.h> /* Funes de alocao de memria */

/****************** Definies de Macros ******************/

#define TAM_CHAVE        5  /* Nmero de dgitos na chave */
#define BASE            10  /* A base numrica utilizada  */
#define PRIMEIRO_DIGITO '0' /* O primeiro dgito da base  */

/****************** Definies de Tipos *******************/

typedef char tChave[TAM_CHAVE + 1];

typedef struct no {
           tChave     chave;
           struct no *proximo;
        } tNo, *tLista;

/********************* Funes Globais ********************/

/****
 *
 * RadixSort(): Implementa o mtodo de ordenao RadixSort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser ordenada
 *      nChaves (entrada) - nmero de chaves na tabela
 *      tamChave (entrada) - tamanho da chave (v. Observaes)
 *
 * Retorno: Nada
 *
 * Observaes:
 *      1. Assume-se que os elementos do array a ser ordenado
 *         so strings de mesmo tamanho. Esta funo no faz
 *         nenhum teste para verificar se isso realmente
 *         ocorre
 *      2. Assume-se que os strings so numricos com base 10
 *      3. O tamanho da chave no leva em considerao o
 *         caractere terminal de string ('\0')
 *      4. No foi implementado o devido tratamento de exceo
 *         (p.ex., em alocalo de memria)
 *
 ****/
void RadixSort(tChave tabela[], int nChaves, int tamChave)
{
   int     i, d, b;
   tLista *listas, pNo, p, q;
   char   *pDigito;

   listas = calloc(BASE, sizeof(tLista));

   for (d = 0; d < tamChave; ++d) {
         /* Armazena as chaves em filas de acordo  */
         /* com o dgito na posio d (contando de */
         /* trs para a frente)                    */
      for (i = 0; i < nChaves; ++i) {
            /* Para cada chave  alocado um n que */
            /* ser inserido em alguma das filas   */
         pNo = malloc(sizeof(tNo));
         strcpy(pNo->chave, tabela[i]);
         pNo->proximo = NULL;

            /* Faz pDigito apontar para o dgito */
            /* da corrente coleta de chaves      */
         pDigito = (char *)pNo->chave + tamChave - 1 - d;

            /* Determina a fila na qual a   */
            /* chave corrente ser colocada */
         p = listas[*pDigito - PRIMEIRO_DIGITO];
         q = NULL;

         while(p) { /* O fundo da fila  o final */
                    /* da lista encadeada        */
            q = p;
            p = p->proximo;
         }

         if (!q) /* Esta fila est vazia */
            listas[*pDigito - PRIMEIRO_DIGITO] = pNo;
         else /* A fila j contm pelo menos um elemento */
            q->proximo = pNo;
      }

      i = 0;

         /* Esvazia todas as filas e armazena as chaves */
         /* na tabela comeando-se com a fila associada */
         /* ao menor dgito e terminando-se com a fila  */
         /* associada ao maior dgito                   */
      for (b = 0; b < BASE; ++b) {
         p = listas[b];
         while (p) { /* Esvazia a fila corrente */
            strcpy(tabela[i], p->chave);
            ++i;
            q = p->proximo;
            free(p);
            p = q;
         } /* while */

            /* As filas foram todas esvaziadas.     */
            /* necessrio anular os ponteiros para   */
            /* os incios das filas antes de iniciar */
            /* uma nova rodada do processo.          */
         listas[b] = NULL;

      } /* for */
   } /* for */

   free(listas);
}

/****
 *
 * EstaOrdenada2(): Testa se uma tabela est ordenada em
 *                  ordem crescente (verso para strings)
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada2(tChave tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (strcmp(tabela[i], tabela[i+1]) > 0)
         return 0;

   return 1;
}

/****
 *
 * RadixSortDecrescente(): Implementa o mtodo RadixSort
 *                         para ordenao descrescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser ordenada
 *      nChaves (entrada) - nmero de chaves na tabela
 *      tamChave (entrada) - tamanho da chave (v. Observaes)
 *
 * Retorno: Nada
 *
 * Observaes:
 *      1. Assume-se que os elementos do array a ser ordenado
 *         so strings de mesmo tamanho. Esta funo no faz
 *         nenhum teste para verificar se isso realmente
 *         ocorre
 *      2. Assume-se que os strings so numricos com base 10
 *      3. O tamanho da chave no leva em considerao o
 *         caractere terminal de string ('\0')
 *      4. No foi implementado o devido tratamento de exceo
 *         (p.ex., em alocalo de memria)
 *
 ****/
void RadixSortDecrescente(tChave tabela[], int nChaves, int tamChave)
{
   int     i, d, b;
   tLista *listas, pNo, p, q;
   char   *pDigito;

   listas = calloc(BASE, sizeof(tLista));

   for (d = 0; d < tamChave; ++d) {
         /* Armazena as chaves em filas de acordo  */
         /* com o dgito na posio d (contando de */
         /* trs para a frente)                    */
      for (i = 0; i < nChaves; ++i) {
            /* Para cada chave  alocado um n que */
            /* ser inserido em alguma das filas   */
         pNo = malloc(sizeof(tNo));
         strcpy(pNo->chave, tabela[i]);
         pNo->proximo = NULL;

            /* Faz pDigito apontar para o dgito */
            /* da corrente coleta de chaves      */
         pDigito = (char *)pNo->chave + tamChave - 1 - d;

            /* Determina a fila na qual a   */
            /* chave corrente ser colocada */
         p = listas[*pDigito - PRIMEIRO_DIGITO];
         q = NULL;

         while(p) { /* O fundo da fila  o final */
                    /* da lista encadeada        */
            q = p;
            p = p->proximo;
         }

         if (!q) /* Esta fila est vazia */
            listas[*pDigito - PRIMEIRO_DIGITO] = pNo;
         else /* A fila j contm pelo menos um elemento */
            q->proximo = pNo;
      }

      i = 0;

         /* Esvazia todas as filas e armazena as chaves */
         /* na tabela comeando-se com a fila associada */
         /* ao menor dgito e terminando-se com a fila  */
         /* associada ao maior dgito                   */
      for (b = BASE - 1; b >= 0; --b) {
         p = listas[b];
         while (p) { /* Esvazia a fila corrente */
            strcpy(tabela[i], p->chave);
            ++i;
            q = p->proximo;
            free(p);
            p = q;
         } /* while */

            /* As filas foram todas esvaziadas.     */
            /* necessrio anular os ponteiros para   */
            /* os incios das filas antes de iniciar */
            /* uma nova rodada do processo.          */
         listas[b] = NULL;

      } /* for */
   } /* for */

   free(listas);
}

/****
 *
 * EstaOrdenada2(): Testa se uma tabela est ordenada em
 *                  ordem crescente (verso para strings)
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenadaDec(tChave tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (strcmp(tabela[i], tabela[i+1]) < 0)
         return 0;

   return 1;
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa funes de ordenao usando RadixSort
 *
 * Parmetros: nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   size_t nElementos, i;
   tChave tabela[] = { "05532", "05886", "04908", "06671",
                       "01500", "03010", "05608", "05596",
                       "05639", "04893", "03774", "05065",
                       "02157", "05208", "05156", "03658",
                       "04895", "02423", "03734", "05620",
                       "02357", "03807", "05607", "04915",
                       "04244", "04495", "03121", "04450",
                       "01361", "05447", "04112", "05109",
                       "06553", "05063", "03939", "05567",
                       "05453", "04695", "03582", "03689" };

   nElementos = sizeof(tabela)/sizeof(tabela[0]);

   printf("\n\t\t   *** Tabela Original ***\n\n");

   for (i = 0; i < nElementos; ++i) {
      printf("%s", (char *) tabela[i]);
      putchar((i + 1) % 8 ? '\t' : '\n');
   }

   RadixSortDecrescente(tabela, nElementos, TAM_CHAVE);

   if (!EstaOrdenadaDec(tabela, nElementos)) {
      printf("\n\n\a>>> Erro: A tabela NAO foi ordenada <<<\n");
//      return 1;
   }

   printf("\n\n\t\t   *** Tabela em Ordem Decrescente ***\n\n");

   for (i = 0; i < nElementos; ++i) {
      printf("%s", (char *) tabela[i]);
      putchar((i + 1) % 8 ? '\t' : '\n');
   }

   putchar('\n');

   RadixSort(tabela, nElementos, TAM_CHAVE);

   if (!EstaOrdenada2(tabela, nElementos)) {
      printf("\n\n\a>>> Erro: A tabela NAO foi ordenada <<<\n");
//      return 1;
   }

   printf("\n\n\t\t   *** Tabela em Ordem Crescente ***\n\n");

   for (i = 0; i < nElementos; ++i) {
      printf("%s", (char *) tabela[i]);
      putchar((i + 1) % 8 ? '\t' : '\n');
   }

   putchar('\n');

   return 0;
}

/****************** Resultado do Programa *****************

                   *** Tabela Original ***

05532   05886   04908   06671   01500   03010   05608   05596
05639   04893   03774   05065   02157   05208   05156   03658
04895   02423   03734   05620   02357   03807   05607   04915
04244   04495   03121   04450   01361   05447   04112   05109
06553   05063   03939   05567   04756   04648   06004   04562
04878   05533   05448   06749   05094   05008   02308   05025
01381   03004   03603   05449   06006   05515   04755   03921
02329   05867   05371   05451   06007   04930   04935   05577
04037   05452   05294   05531   05453   04695   03582   03689


                   *** Tabela Ordenada ***

01361   01381   01500   02157   02308   02329   02357   02423
03004   03010   03121   03582   03603   03658   03689   03734
03774   03807   03921   03939   04037   04112   04244   04450
04495   04562   04648   04695   04755   04756   04878   04893
04895   04908   04915   04930   04935   05008   05025   05063
05065   05094   05109   05156   05208   05294   05371   05447
05448   05449   05451   05452   05453   05515   05531   05532
05533   05567   05577   05596   05607   05608   05620   05639
05867   05886   06004   06006   06007   06553   06671   06749

**********************************************************/

