/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */

/****
 *
 * Troca(): Troca os valores de duas variveis do tipo int
 *
 * Parmetros:
 *      *x, *y (entrada/sada) - variveis que tero seus
 *                               valores trocados
 *
 * Retorno: Nada
 *
 ****/
static void Troca(int *x, int *y)
{
   int aux = *x;

   *x = *y;
   *y = aux;
}

/****
 *
 * BubbleSort(): Ordena uma tabela usando BubbleSort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void BubbleSort(int tabela[], int nElem)
{
   int i, ordenado = 0, cont = 0;

   while (!ordenado){
         /* Supe que a tabela est ordenada */
      ordenado = 1;

      for (i = 0; i < nElem - 1; i++) {
         ++cont;

            /* Compara elementos adjacentes */
         if (tabela[i] > tabela[i + 1]){
               /* Pelo menos um par de elementos */
               /* est fora de ordem             */
            ordenado = 0;

               /* Troca elementos que esto fora de ordem */
            Troca(tabela + i, tabela + i + 1);
         }
      }
   }

   printf("\t>>> BubbleSort: %d comparacoes\n", cont);
}

/****
 *
 * SelectionSort(): Ordena uma tabela usando seleo direta
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void SelectionSort(int tabela[], int nElem)
{
   int i, j, iMenor, aux, cont = 0;

      /* A poro ordenada do array est entre os */
      /* ndices 0 e i - 1, enquanto que a poro */
      /* desordenada est entre i e n - 1         */

      /* Obtm o menor elemento na parte desordenada  */
      /* e troca-o de posio com o primeiro elemento */
      /* da parte desordenada (que  i)               */
   for (i = 0; i < nElem - 1; ++i) {
         /* Supe que o primeiro elemento  */
         /* da parte desordenada  o menor */
      iMenor = i;

//      ++cont;

         /* Verifica se existe um elemento menor */
         /* do 'iMenor' na parte desordenada     */
      for (j = i + 1; j < nElem; ++j)
            /* Se o elemento corrente for menor do   */
            /* que 'iMenor', ele passa a ser o menor */
        if (tabela[j] < tabela[iMenor])
            iMenor = j; /* O novo mnimo passa a ser j */

         ++cont;

         /* Se foi encontrado um elemento menor do que aquele   */
         /* que se encontra na posio i, trocam-se as posies */
         /* desses elementos. Caso contrrio, o elemento na     */
         /* posio i j est em sua posio ordenada.          */
      if (iMenor != i) {
         /* Troca os elementos de posio */
        aux = tabela[i];
        tabela[i] = tabela[iMenor];
        tabela[iMenor] = aux;
      }
   }

   printf("\t>>> SelectionSort: %d comparacoes\n", cont);
}

/****
 *
 * InsertionSort(): Ordena uma tabela usando ordenao por
 *                  insero simples
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void InsertionSort(int tabela[], int  nElem)
{
   int i, j, aux, cont = 0;

      /*                                              */
      /* Inicialmente, o array tabela[]  considerado */
      /* uma tabela ordenada. Ento, a cada iterao, */
      /* os elementos entre 0 e j estaro ordenados.  */
      /*                                              */

      /* Insere tabela[j] na parte ordenada da tabela */
   for (j = 1; j < nElem; ++j) {
      aux = tabela[j];

      ++cont;

         /* Move cada elemento maior do */
         /* que aux uma posio adiante */
         /*      */
      for (i = j-1; i >= 0 && aux < tabela[i]; --i)
         tabela[i + 1] = tabela[i];

      tabela[i + 1] = aux; /* Insere aux na posio corsreta */
   }

   printf("\t>>> InsertionSort: %d comparacoes\n", cont);
}

/****
 *
 * Quick1(): Ordena uma tabela usando o algoritmo QuickSort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser ordenada
 *      inf (entrada) - limite inferior da tabela
 *      sup (entrada) - limite superior da tabela
 *
 * Retorno: Nada
 *
 * Observaes:
 *   1. Por simplicidade, i.e., com o objetivo didtico
 *      de demonstrar o mtodo esta funo  implementada
 *      recursivamente, mas evidentemente uma implementao
 *      no recursiva seria mais eficiente
 *   2. O piv  o primeiro elemento da tabela
 *
 ****/
static int Quick1(int tabela[], int inf, int sup)
{
   int i, j, pivo;
   static int cont = 0;

   if (inf >= sup)
      return cont; /* A tabela j deve estar ordenada */

      /* Nesta implementao, o piv  sempre */
      /* o primeiro elemento da tabela        */
   pivo = inf;

      /* Cria as parties */
   for (i = inf, j = sup; i < j; ) {
      cont = cont + 2;

         /* Enquanto os elementos da tabela forem menores */
         /* do que o piv, avana-se o ndice inferior    */
      while (tabela[i] <= tabela[pivo] && i < sup)
         ++i;

         /* Enquanto os elementos da tabela forem maiores */
         /* do que o piv, recua-se o ndice superior     */
      while (tabela[j] > tabela[pivo])
         --j;

         /* Os elementos nos ndices i e j esto em parties */
         /* erradas e  preciso troc-los de lugar            */
      if (i < j)
         Troca(tabela + i, tabela + j);
   }

      /* Coloca o piv em sua correta posio */
   Troca(tabela + pivo, tabela + j);

      /* Ordena recursivamente a primeira metade da tabela */
   return Quick1(tabela, inf, j - 1);

      /* Ordena recursivamente a segunda metade da tabela */
   return Quick1(tabela, j + 1, sup);
}

/****
 *
 * QuickSort1(): Ordena uma tabela usando quick sort. Esta funo
 *               simplesmente prov uma interface para a funo
 *               Quick1() que realmente faz o servio.
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
int QuickSort1(int tabela[], int nElem)
{
      /* Simplesmente chama Quick() para fazer o servio */
   return Quick1(tabela, 0, nElem - 1);
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero
 *                         exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na
 *                             tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * EstaInvOrdenada(): Testa se uma tabela est ordenada em
 *                    ordem decrescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaInvOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] < tabela[i+1])
         return 0;

   return 1;
}

int main(void)
{
   int  tabela[100];
   int  i, cont, nElementos = 100;

   for (i = 0; i < 100; ++i)
      tabela[i] = i;

   if (EstaOrdenada(tabela, nElementos))
      printf("\n\t*** Tabela Ordenada ***\n\n");

   if (EstaInvOrdenada(tabela, nElementos))
      printf("\n\t*** Tabela Inversamente Ordenada ***\n");

   BubbleSort(tabela, nElementos);
   SelectionSort(tabela, nElementos);
   InsertionSort(tabela, nElementos);
   cont = QuickSort1(tabela, nElementos);

   printf("\t>>> QuickSort1: %d comparacoes\n", cont);

//   printf("\n\n************ Depois da Ordenacao ************\n");
//   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   return 0;
}

/******************* Exemplo de Execuo ******************

        >>> BubbleSort: 99 comparacoes
        >>> SelectionSort: 99 comparacoes
        >>> InsertionSort: 99 comparacoes
        >>> QuickSort1: 2 comparacoes

**********************************************************/
