/****
 *
 * Exemplo: BubbleSort
 *
 * Seo: 11.2.1
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void BubbleSort(int tabela[], int nElem);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * Troca(): Troca os valores de duas variveis do tipo int
 *
 * Parmetros:
 *      *x, *y (entrada/sada) - variveis que tero seus
 *                               valores trocados
 *
 * Retorno: Nada
 *
 ****/
static void Troca(int *x, int *y)
{
   int aux = *x;

   *x = *y;
   *y = aux;
}

/****
 *
 * BubbleSort(): Ordena uma tabela usando BubbleSort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void BubbleSort(int tabela[], int nElem)
{
   int i, ordenado = 0, cont = 0;

   while (!ordenado){
      if (cont >= 3) {
         printf("\n\n*********** Apos 3 Passagens ************");
         return;
      }

         /* Supe que a tabela est ordenada */
      ordenado = 1;

      for (i = 0; i < nElem - 1; i++)
            /* Compara elementos adjacentes */
         if (tabela[i] > tabela[i + 1]){
               /* Pelo menos um par de elementos */
               /* est fora de ordem             */
            ordenado = 0;

               /* Troca elementos que esto fora de ordem */
            Troca(tabela + i, tabela + i + 1);
         }

      ++cont;
   }
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando BubbleSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int  tabela[] = {55, 8, 13, 28, 23, 9, 25, 11, 70, 19};
   int  nElementos = sizeof(tabela)/sizeof(tabela[0]);

   printf("\n\n>>>>>>>>>>>>>>> Bubble Sort <<<<<<<<<<<<<<<\n");

   printf("\n*********** Antes da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);

   BubbleSort(tabela, nElementos);

//   if (!EstaOrdenada(tabela, nElementos)) {
//      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
//      return 1;
//   }

   printf("\n\n*********** Depois da Ordenacao ***********\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>> Bubble Sort <<<<<<<<<<<<<<<

*********** Antes da Ordenacao ************

Numero de elementos: 10
{55, 8, 13, 28, 23, 9, 25, 11, 70, 19}

*********** Depois da Ordenacao ***********

Numero de elementos: 10
{8, 9, 11, 13, 19, 23, 25, 28, 55, 70}

{8, 9, 11, 13, 19, 23, 25, 28, 55, 70}

*********** Apos 3 Passagens ************

{8, 13, 9, 23, 11, 25, 19, 28, 55, 70}

**********************************************************/
