/****
 *
 * Exemplo: InsertionSort
 *
 * Seo: 11.2.3
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void InsertionSort(int tabela[], int nElem);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

void InsertionSort(int tabela[], int  nElem)
{
   int i, j, aux, cont = 0;

      /*                                              */
      /* Inicialmente, o array tabela[]  considerado */
      /* uma tabela ordenada. Ento, a cada iterao, */
      /* os elementos entre 0 e j estaro ordenados.  */
      /*                                              */

      /* Insere tabela[j] na parte ordenada da tabela */
   for (j = 1; j < nElem; ++j) {
      aux = tabela[j];

      if (cont >= 3) {
         printf("\n*********** Apos 3 Passagens ************\n");
         return;
      }

         /* Move cada elemento maior do */
         /* que aux uma posio adiante */
         /*      */
      for (i = j - 1; i >= 0 && aux < tabela[i]; --i)
         tabela[i + 1] = tabela[i];

      tabela[i + 1] = aux; /* Insere aux na posio corsreta */

      ++cont;
   }
}


/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero
 *                         exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na
 *                             tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando InsertionSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int  tabela[] = {55, 8, 13, 28, 23, 9, 25, 11, 70, 19};
   int  nElementos = sizeof(tabela)/sizeof(tabela[0]);

   printf("\n\n>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<\n");

   printf("\n************ Antes da Ordenacao *************\n");
   ApresentaTabela(tabela, nElementos);

   InsertionSort(tabela, nElementos);

//   if (!EstaOrdenada(tabela, nElementos)) {
//      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
//      return 1;
//   }

   printf("\n\n************ Depois da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<

************ Antes da Ordenacao *************

Numero de elementos: 10
{55, 8, 13, 28, 23, 9, 25, 11, 70, 19}

*********** Apos 3 Passagens ************

************ Depois da Ordenacao ************

Numero de elementos: 10
{8, 13, 28, 55, 23, 9, 25, 11, 70, 19}

**********************************************************/
