/****
 *
 * Exemplo: SelectionSort
 *
 * Seo: 11.2.2
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void SelectionSort(int tabela[], int nElem);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * SelectionSort(): Ordena uma tabela usando seleo direta
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void SelectionSort(int tabela[], int nElem)
{
   int i, j, iMenor, aux, cont = 0;

      /* A poro ordenada do array est entre os */
      /* ndices 0 e i - 1, enquanto que a poro */
      /* desordenada est entre i e n - 1         */

      /* Obtm o menor elemento na parte desordenada  */
      /* e troca-o de posio com o primeiro elemento */
      /* da parte desordenada (que  i)               */
   for (i = 0; i < nElem - 1; ++i) {
      if (cont >= 3) {
         printf("\n\n*********** Apos 3 Passagens ************");
         return;
      }

         /* Supe que o primeiro elemento  */
         /* da parte desordenada  o menor */
      iMenor = i;

         /* Verifica se existe um elemento menor */
         /* do 'iMenor' na parte desordenada     */
      for (j = i + 1; j < nElem; ++j)
            /* Se o elemento corrente for menor do   */
            /* que 'iMenor', ele passa a ser o menor */
        if (tabela[j] < tabela[iMenor])
            iMenor = j; /* O novo mnimo passa a ser j */

         /* Se foi encontrado um elemento menor do que aquele   */
         /* que se encontra na posio i, trocam-se as posies */
         /* desses elementos. Caso contrrio, o elemento na     */
         /* posio i j est em sua posio ordenada.          */
      if (iMenor != i) {
         /* Troca os elementos de posio */
        aux = tabela[i];
        tabela[i] = tabela[iMenor];
        tabela[iMenor] = aux;
      }

      ++cont;
   }
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero
 *                         exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na
 *                             tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser
 *                               liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando SelectionSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int  tabela[] = {55, 8, 13, 28, 23, 9, 25, 11, 70, 19};
   int  nElementos = sizeof(tabela)/sizeof(tabela[0]);

   printf("\n\n>>>>>>>>>>>>>>> SelectionSort <<<<<<<<<<<<<<<\n");

   printf("\n************ Antes da Ordenacao *************\n");
   ApresentaTabela(tabela, nElementos);

   SelectionSort(tabela, nElementos);

//   if (!EstaOrdenada(tabela, nElementos)) {
//      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
//      return 1;
//   }

   printf("\n\n************ Depois da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>> SelectionSort <<<<<<<<<<<<<<<

************ Antes da Ordenacao *************

Numero de elementos: 10
{55, 8, 13, 28, 23, 9, 25, 11, 70, 19}

*********** Apos 3 Passagens ************

************ Depois da Ordenacao ************

Numero de elementos: 10
{8, 9, 11, 28, 23, 55, 25, 13, 70, 19}

**********************************************************/
