/****
 *
 * Ttulo: Saltando Linhas de um Arquivo de Texto
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/08/2012
 * ltima modificao: 16/08/2012
 *
 * Descrio: Salta linhas de um arquivo de texto
 *
 * Entrada: v. exemplo de execuo
 *
 * Sada: v. exemplo de execuo
 *
 ****/

#include <stdio.h> /* Entrada e sada */

   /* Nome do arquivo a ser usado como teste */
#define NOME_ARQ    "TesteSaltaLinhas.txt"

   /* Tamanho de um array que armazenar um string */
#define TAM_STRING  20

   /* Tamanho de um array que armazenar um       */
   /* especificador de formato usado com fscanf() */
#define TAM_ESPEC  5

/****
 *
 * SaltaLinhas(): Salta o restante da linha corrente e linhas
 *                vazias subsequentes de um arquivo de texto
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *
 * Retorno: O nmero de linhas saltadas, se no ocorrer
 *          erro. Um valor negativo, em caso contrrio.
 *
 * Observao:
 *     1. Linha vazia  aquela contendo apenas '\n'
 *     2. O arquivo deve ter sido aberto em modo texto que
 *        permite leitura
 *
 ****/
int SaltaLinhas(FILE *stream)
{
   int c, /* Armazena um caractere */
       linhasSaltadas = 0; /* Conta o nmero de linhas saltadas */

      /* O lao encerra quando o final do arquivo for atingido, */
      /* ocorrer erro ou for encontrado um caractere aps a     */
      /* linha corrente que no seja quebra de linha            */
   while (1) {
      c = fgetc(stream); /* L o prximo caractere */

         /* Se ocorreu erro de leitura, retorna */
         /* um valor que indica esse fato       */
      if (ferror(stream)) {
         return -1; /* Ocorreu erro */
      }

         /* Se o final do arquivo foi atingido antes  */
         /* que fosse encontrada uma quebra de linha, */
         /* o nmero de linhas saltadas  zero        */
      if (feof(stream)) { /* O arquivo acabou e no */
         return 0;        /* foi encontrado '\n'    */
      }

         /* Se o final da linha corrente foi atingido, */
         /* saltam-se as linhas vazias subsequentes    */
      if(c == '\n') { /* Fim da linha corrente */

            /* O lao encerra quando o final do arquivo for */
            /* atingido, ocorrer erro ou for encontrado um  */
            /* caractere que  no  '\n'                    */
         while(c == '\n') { /* Salta linhas */
               /* Mais uma quebra de linha encontrada */
            linhasSaltadas++;

            c = fgetc(stream); /* L o prximo caractere */

               /* Se ocorreu erro retorna um */
               /* valor que indica esse fato */
            if (ferror(stream)) {
               return -1;
            }

               /* Se o final do arquivo foi atingido, no */
               /* h mais nada a fazer a no ser retornar */
            if (feof(stream)) {
                  /* Retorna nmero de linhas saltadas */
               return linhasSaltadas;
            }
         }

            /************************************************/
            /* Neste ponto, sabe-se que o ltimo caractere  */
            /* lido  o primeiro caractere da primeira      */
            /* linha que no  vazia. Esse caractere deve   */
            /* ser o primeiro a ser lido na prxima leitura */
            /* do arquivo. Portanto,  preciso devolver     */
            /* esse caractere para o stream.  para isso    */
            /* que serve a funo ungetc().                 */
            /************************************************/

            /* Devolve o ltimo caractere lido para o stream */
         if (ungetc(c, stream) == EOF) {
               /* No foi possvel devolver caractere */
            return -1;
         }

         break; /* Encerra o lao externo */
      }
   }

      /* Retorna o nmero de caracteres '\n' que foram lidos */
   return linhasSaltadas;
}

/****
 *
 * main(): Salta linhas de um arquivo de texto
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 ****/
int main(void)
{
   FILE *stream;
   char  str[TAM_STRING],
         especificador[TAM_ESPEC];
   int   nLinhasSaltadas;

      /* Apresenta o programa */
   printf( "\n\t>>> Este programa cria um arquivo de"
           "\n\t>>> texto com linhas vazias e testa"
           "\n\t>>> uma funcao que salta linhas.\n" );

      /* Cria um arquivo de texto e abre-o para */
      /* leitura e escrita. Se o arquivo j     */
      /* existir, seu contedo ser destrudo.  */
   stream = fopen(NOME_ARQ, "w+");

      /* Se o arquivo no pode ser aberto, */
      /* o programa  encerrado            */
   if (!stream) {
      printf( "\n\t>>> O arquivo \"%s\" nao pode ser "
              "aberto\n", NOME_ARQ );
      return 1;
   }

      /* Coloca o recheio do arquivo */
   fprintf(stream, "Primeira linha\n");
   fprintf(stream, "\n"); /* Uma linha vazia */
   fprintf(stream, "\n"); /* Outra linha vazia */
   fprintf(stream, "\n"); /* Mais uma linha vazia */
   fprintf(stream, "Quinta linha\n"); /* ltima linha */

         /*                                              */
         /* Garante que contedo do buffer  armazenado  */
         /* imediatamente no arquivo. O uso de rewind()  */
         /* adiante torna a chamada de fflush(()         */
         /* desnecessria. Mas, este  um livro didtico */
         /* e  preciso demonstrar o uso de fflush()...  */
         /*                                              */

      /* Tenta descarregar o buffer associado ao arquivo */
   if (fflush(stream)) {
      printf( "\n\t>>> Ocorreu um erro ao tentar descarregar "
              "o buffer do arquivo \"%s\"\n", NOME_ARQ );
   }

      /* Faz o indicador de posio do       */
      /* arquivo voltar ao incio do arquivo */
   rewind(stream);

      /* Cria o especificador que ser usado com fscanf() */
   sprintf(especificador, "%c%d%c", '%', TAM_STRING - 1, 's');

      /* L um string na primeira linha do arquivo */
   fscanf(stream, especificador, str);

      /* Apresenta o string lido no arquivo */
   printf("\n\t>>> Lido no arquivo: \"%s\"\n", str);

      /* Salta o restante da primeira linha e */
      /* as demais linhas vazias do arquivo   */
   nLinhasSaltadas = SaltaLinhas(stream);

      /* Se no houve erro, informa o nmero de linhas que */
      /* foram saltadas. Seno, informa que houve erro.    */
   if (nLinhasSaltadas < 0) {
      printf( "\n\t>>> Ocorreu erro ao tentar saltar "
              "linhas do arquivo \"%s\"", NOME_ARQ);
   } else {
      printf("\n\t>>> Foram saltadas %d linhas", nLinhasSaltadas);
   }

      /* L um string na prxima linha do arquivo */
   fscanf(stream, especificador, str);

      /* Apresenta o string lido no arquivo */
   printf("\n\n\t>>> Lido no arquivo: \"%s\"\n", str);

   return 0;
}

/******************* Exemplo de Execuo ******************

        >>> Este programa cria um arquivo de texto com linhas
        >>> vazias e testa uma funcao que salta linhas

        >>> Lido no arquivo: "Primeira"

        >>> Foram saltadas 4 linhas

        >>> Lido no arquivo: "Quinta"

**********************************************************/
