/****
 *
 * Ttulo: Atualizando Registros de um Arquivo de Texto
 *
 * Data de Criao: 17/05/2012
 * ltima modificao: 18/05/2012
 *
 * Descrio:
 *     1. L cada linha do arquivo especificado pela constante
 *        NOME_ARQUIVO.
 *     2. Converte a linha lida numa estrutura do tipo tAluno.
 *     3. Atualiza a estrutura acrescentando 1 ao campo n2.
 *     5. Escreve a estrutura modificada no arquivo especificado
 *        usando o mesmo formato de linha do arquivo original.
 *
 * Entrada: Nada
 *
 * Sada: O arquivo atualizado
 *
 ****/

/******************* Arquivo de Entrada ********************

(As duas primeiras linhas a seguir no existem no arquivo)

Nome            Matr   N1    N2

Henrique VIII   1029   9.5   9.0
Catarina Aragon 1014   5.5   6.5
Ana Bolena      1012   7.8   8.0
Joana Seymour   1017   7.7   8.7
Ana de Cleves   1022   4.5   6.0
Catarina Howard 1340   6.0   7.7
Catarina Parr   1440   4.0   6.0

************************ Includes *************************/

#include <stdio.h>  /* Entrada e sada          */
#include <string.h> /* Processamento de strings */
#include <stdlib.h> /* Funo strtod()          */

/***************** Constantes Simblicas ******************/

#define MAX_NOME  20 /* Nmero mximo de caracteres num nome */
#define TAM_MATR   4 /* Nmero de dgitos numa matrcula */
#define MAX_LINHA 50 /* Tamanho mximo de uma linha no arquivo */
#define NOME_ARQUIVO "Tudor.txt" /* Arquivo original */
   /* Arquivo atualizado */
#define NOME_ARQUIVO_ATUAL "TudorAtual.txt"

/****************** Definies de Tipos *******************/

typedef  struct {
            char    nome[MAX_NOME + 1];
            char    matr[TAM_MATR + 1];
            double  n1, n2;
         } tAluno;

/***************** Definies de Funes ******************/

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa estrutura
 *                    do tipo tAluno.
 *
 * Parmetros:
 *     pAluno (sada) - ponteiro para a estrutura resultante da
 *                      converso
 *     linha (entrada/sada) - linha que ser convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado da
 *          converso
 *
 * Observaes:
 *     1. Para simplificar, esta funo assume que o parmetro
 *        'linha' realmente  um string no formato das linhas do
 *        arquivo. Portanto, os valores retornados por strtok()
 *        no so testados como deveriam.
 *     2. O parmetro 'linha'  alterado por strtok().
 *
 ****/
tAluno *LinhaEmRegistro(tAluno *pAluno, char *linha)
{
   char *str; /* Apontar para cada token da linha */

      /* Obtm o nome e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = strtok(linha, "\t\n");
   strcpy(pAluno->nome, str);

      /* Obtm a matrcula e acrescenta-a */
      /* ao respectivo campo da estrutura */
   str = strtok(NULL, "\t\n");
   strcpy(pAluno->matr, str);

      /* Obtm a 1a. nota, converte-a em double e acrescenta o */
      /* valor convertido ao campo respectivo da estrutura     */
   str = strtok(NULL, "\t\n"); /* Obtm a 1a. nota e... */
   pAluno->n1 = strtod(str, NULL); /* converte em double */

      /* Idem para a 2a. nota */
   str = strtok(NULL, "\t\n"); /* Obtm a 2a. nota e... */
   pAluno->n2 = strtod(str, NULL); /* converte em double */

   return pAluno;
}

/****
 *
 * AtualizaArquivo(): L cada linha do arquivo de entrada,
 *                    converte-a numa estrutura do tipo tAluno,
 *                    atualiza a estrutura e escreve a estrutura
 *                    modificada no arquivo de sada usando o
 *                    mesmo formato do arquivo de entrada
 *
 * Parmetros:
 *     streamSaida (entrada) - stream associado ao arquivo que
 *                             conter a atualizao
 *     streamEntrada (entrada) - stream associado ao arquivo que
 *                               ser lido
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de texto
 *        que permite leitura
 *     2. O stream de sada deve estar aberto em modo "w"
 *
 ****/
int AtualizaArquivo(FILE *streamSaida, FILE *streamEntrada)
{
   tAluno umAluno; /* Armazenar dados de uma estrutura */
   char   linha[MAX_LINHA + 1]; /* Armazena uma linha do */
                                /* arquivo de entrada    */

      /* Garante que a leitura comea no primeiro byte */
   rewind(streamEntrada);

      /* L cada linha do arquivo de entrada, armazena os      */
      /* dados numa estrutura, atualiza a estrutura e escreve- */
      /* a no arquivo de sada no formato de linha do arquivo  */
      /* original. O lao encerra quando houver tentativa de   */
      /* leitura alm do final do arquivo de entrada, ou       */
      /* ocorrer erro de leitura ou escrita.                   */
   while (1) {
         /* L uma linha no arquivo de entrada */
      fgets(linha, MAX_LINHA + 1, streamEntrada);

         /* Verifica se ocorreu erro ou tentativa */
         /* de leitura alm do final do arquivo   */
      if ( feof(streamEntrada) || ferror(streamEntrada) )
         break;

         /* Converte a linha lida numa */
         /* estrutura do tipo tAluno   */
      LinhaEmRegistro(&umAluno, linha);

         /* Atualiza o campo n2 da estrutura */
      umAluno.n2 = umAluno.n2 + 1;

         /* Escreve a estrutura no arquivo */
         /* de sada usando fprintf()      */
      fprintf(streamSaida, "%s\t%s\t%3.1f\t%3.1f\n", umAluno.nome,
               umAluno.matr, umAluno.n1, umAluno.n2);

         /* Encerra o lao se ocorreu algum erro de escrita */
      if (ferror(streamSaida)) {
         break;
      }
   }

      /* Se ocorreu algum erro de processamento, */
      /* o valor retornado ser 1                */
   return ferror(streamEntrada) || ferror(streamSaida);
}

/****
 *
 * main():
 *   1. L cada linha do arquivo cujo nome  especificado pela
 *      constante NOME_ARQUIVO.
 *   2. Converte cada a linha lida numa estrutura do tipo tAluno.
 *   3. Atualiza a estrutura acrescentando 1.0 ao campo n2.
 *   4. Escreve a estrutura modificada no arquivo especificado
 *      pela constante NOME_ARQUIVO_ATUAL usando o mesmo formato
 *      de linha do arquivo original.
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 ****/
int main(void)
{
   FILE *streamE, /* Associado ao arquivo de entrada */
        *streamS; /* Associado ao arquivo de sada   */

      /* Tenta abrir arquivo de entrada */
      /* para leitura em modo texto     */
   streamE = fopen(NOME_ARQUIVO, "r");

      /* Se o arquivo de entrada no pode     */
      /* ser aberto, nada mais pode ser feito */
   if (!streamE) {
      printf( "\nArquivo \"%s\" nao pode ser aberto\n",
              NOME_ARQUIVO );
      return 1; /* Arquivo de entrada no foi aberto */
   }

      /* Tenta abrir arquivo de sada */
      /* para escrita em modo texto   */
   streamS = fopen(NOME_ARQUIVO_ATUAL, "w");

      /* Se o arquivo de sada no pode ser */
      /* aberto, nada mais pode ser feito   */
   if (!streamS) {
         /* Fecha arquivo de entrada antes de partir */
      fclose(streamE);

      printf( "\nArquivo \"%s\" nao pode ser aberto\n",
              NOME_ARQUIVO_ATUAL );

      return 1; /* Arquivo de sada no foi aberto */
   }

      /* Cria o arquivo atualizado */
   if (AtualizaArquivo(streamS, streamE)) {
      printf( "\n\t>>> Ocorreu um erro na atualizacao do"
              "\n\t>>> arquivo \"%s\"\n", NOME_ARQUIVO_ATUAL );
      return 1; /* Ocorreu algum erro durante atualizao */
   }

   printf( "\n\t>>> Atualizacao do arquivo \"%s\" foi"
           "\n\t>>> escrita no arquivo \"%s\"\n",
           NOME_ARQUIVO, NOME_ARQUIVO_ATUAL );

   return 0; /* Tudo ocorreu bem */
}

/******************* Exemplo de Execuo ******************

        >>> Atualizacao do arquivo "Tudor.txt" foi
        >>> escrita no arquivo "TudorAtual.txt"

**********************************************************/
