/****
 *
 * Ttulo: Convertendo um Arquivo de Registros de Texto para Binrio
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 21/05/2012
 * ltima modificao: 21/05/2012
 *
 * Descrio: L cada linha de um arquivo de texto, converte-a numa
 *            estrutura do tipo tAluno e armazena-a num arquivo
 *            binrio
 *
 * Entrada: Arquivo cujo nome  especificado pela constante
 *          NOME_ARQUIVO_TEXTO
 *
 * Sada: Arquivo cujo nome  especificado pela constante
 *        NOME_ARQUIVO_BIN
 *
 ****/

/******************* Arquivo de Entrada ********************

(As duas primeiras linhas a seguir no existem no arquivo)

Nome            Matr   N1   N2

Henrique VIII   1029   9.5   9.0
Catarina Aragon 1014   5.5   6.5
Ana Bolena      1012   7.8   8.0
Joana Seymour   1017   7.7   8.7
Ana de Cleves   1022   4.5   6.0
Catarina Howard 1340   6.0   7.7
Catarina Parr   1440   4.0   6.0

***********************************************************/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de Entrada e Sada */
#include <string.h> /* Processamento de strings   */
#include <stdlib.h> /* Funo strtod()            */

/***************** Constantes Simblicas ******************/

#define MAX_NOME  20 /* Nmero mximo de caracteres num nome */
#define TAM_MATR   4 /* Nmero de dgitos numa matrcula */
#define MAX_LINHA 50 /* Tamanho mximo de uma linha no arquivo */
#define NOME_ARQUIVO_TEXTO "Tudor.txt" /* Arquivo original */
#define NOME_ARQUIVO_BIN "Tudor.bin" /* Arquivo atualizado */

/****************** Definies de Tipos *******************/

typedef  struct {
            char    nome[MAX_NOME + 1];
            char    matr[TAM_MATR + 1];
            double  n1, n2;
         } tAluno;

/***************** Definies de Funes ******************/

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa
 *                    estrutura do tipo tAluno.
 *
 * Parmetros: aluno (sada) - ponteiro para a estrutura
 *                             resultante da converso
 *             linha (entrada/sada) - linha que ser
 *                                     convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado
 *          da converso
 *
 * Observaes:
 *     1. Para facilitar o entendimento, esta funo assume
 *        que o parmetro 'linha' realmente  um string no
 *        formato das linhas do arquivo. Portanto, os
 *        valores retornados por strtok() no so testados
 *        como deveriam.
 *
 *     2. O parmetro 'linha'  alterado por strtok().
 *
 ****/
tAluno *LinhaEmRegistro(tAluno *aluno, char *linha)
{
   char *str; /* Apontar para cada token da linha */

      /* Obtm o nome e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = strtok(linha, "\t\n");
   strcpy(aluno->nome, str);

      /* Obtm a matrcula e acrescenta-a */
      /* ao respectivo campo da estrutura */
   str = strtok(NULL, "\t\n");
   strcpy(aluno->matr, str);

      /* Obtm a 1a. nota, converte-a em double e */
      /* acrescenta o valor convertido ao campo   */
      /* respectivo da estrutura                  */
   str = strtok(NULL, "\t\n"); /* Obtm a 1a. nota e... */
   aluno->n1 = strtod(str, NULL); /* converte em double */

      /* Idem para a 2a. nota */
   str = strtok(NULL, "\t\n"); /* Obtm a 2a. nota e... */
   aluno->n2 = strtod(str, NULL); /* converte em double */

   return aluno;
}

/****
 *
 * CriaArquivoBin(): L cada linha de um arquivo no formato
 *                   especificado, converte-a numa estrutura
 *                   do tipo tAluno e armazena a estrutura
 *                   num arquivo binrio.
 *
 * Parmetros:
 *     streamTexto (entrada) - stream associado ao arquivo que
 *                             ser lido
 *     streamBin (entrada) - stream associado ao arquivo que ser
 *                           escrito
 *
 * Retorno: Nmero de estruturas escritas no arquivo
 *          binrio ou um valor negativo, se ocorrer erro
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de
 *        texto que permite leitura
 *     2. O stream de sada deve estar aberto em modo "wb"
 *
 ****/
int CriaArquivoBin(FILE *streamTexto, FILE  *streamBin)
{
   char    linha[MAX_LINHA + 1]; /* Armazenar cada linha lida */
   tAluno  umAluno; /* Dados de uma linha      */
                    /* convertida em estrutura */
   int     nRegistros = 0; /* Nmero de estruturas escritas */
                           /* no arquivo binrio            */

      /* Garante leitura a partir do incio do arquivo */
   rewind(streamTexto);

      /* L cada linha do arquivo, cria um registro do tipo     */
      /* tAluno e armazena-o no arquivo binrio. O lao encerra */
      /* quando h tentativa de leitura alm do final do        */
      /* arquivo de entrada, ou ocorre erro de leitura/escrita  */
   while (1) {
         /* L uma linha no arquivo de entrada */
      fgets(linha, MAX_LINHA + 1, streamTexto);

         /* Verifica se o final do arquivo foi atingido */
      if (feof(streamTexto))
         break;

         /* Verifica se ocorreu erro de leitura */
      if (ferror(streamTexto))
         return -1; /* Operao mal sucedida */

         /* Converte a linha lida em estrutura */
      LinhaEmRegistro(&umAluno, linha);

         /* Escreve a estrutura resultante da */
         /* converso no arquivo binrio      */
      fwrite(&umAluno, sizeof(umAluno), 1, streamBin);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamBin))  /* Ocorreu */
         return -1; /* Operao mal sucedida */

      ++nRegistros; /* Mais um registro lido */
   }

   return nRegistros; /* No ocorreu nenhum erro */
}

/****
 *
 * main(): Converte um arquivo de texto em arquivo binrio
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 ****/
int main(void)
{
   int   nRegistros; /* Nmero de registros nos arquivos */
   FILE *streamTexto, /* Associado ao arquivo de texto */
        *streamBin;   /* Associado ao arquivo binrio */

      /* Tenta abrir arquivo para leitura em modo texto */
   streamTexto = fopen(NOME_ARQUIVO_TEXTO, "r");

      /* Se o arquivo no pode ser aberto, */
      /* nada mais pode ser feito          */
   if (!streamTexto) {
      printf( "O arquivo %s nao pode ser aberto\n",
              NOME_ARQUIVO_TEXTO );
      return 1; /* Operao mal sucedida */
   }

      /* Tenta criar arquivo binrio. Se um arquivo com esse   */
      /* nome existir no diretrio corrente ele ser destrudo */
   streamBin = fopen(NOME_ARQUIVO_BIN, "wb");

      /* Se o arquivo no pode ser aberto, */
      /* nada mais pode ser feito.         */
   if (!streamBin) {
      printf("Arquivo %s nao pode ser aberto\n",NOME_ARQUIVO_BIN);

         /* Antes de retornar, fecha arquivo aberto */
      fclose(streamTexto);

      return 1; /* Operao mal sucedida */
   }

      /* L cada linha do arquivo de texto,     */
      /* converte-a em estrutura do tipo tAluno */
      /* e armazena-a no arquivo binrio.       */
   nRegistros = CriaArquivoBin(streamTexto, streamBin);

      /* Verifica se ocorreu erro na criao do arquivo */
      /* binrio. Se for o caso, informa o usurio,     */
      /* fecha os arquivos e encerra o programa.        */
   if (nRegistros < 0) {
      printf( "\n\t>>> Ocorreu erro na criacao do arquivo"
              "\n\t>>> \"%s\"\n", NOME_ARQUIVO_BIN );

         /* Fecha os arquivos antes de encerrar */
      fclose(streamTexto);
      fclose(streamBin);

      return 1; /* Ocorreu erro */
   }

      /* Apresenta o resultado da converso */
   printf( "\n\t>>> Foram armazenados %d registros no arquivo"
           "\n\t>>> \"%s\"\n", nRegistros, NOME_ARQUIVO_BIN );

      /* Fecha os arquivos */
   fclose(streamTexto);
   fclose(streamBin);

   return 0; /* Tudo ocorreu bem */
}

/******************* Exemplo de Execuo *******************

        >>> Foram armazenados 7 registros no arquivo
        >>> "Tudor.bin"

***********************************************************/
