/****
 *
 * Exemplo: Listas Encadeadas com Saltos
 *
 * Seo: 3.6
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 19/05/2015
 * ltima modificao: 25/05/2015
 *
 * Entrada: V. exemplo de execuo
 *
 * Sada: V. exemplo de execuo
 *
 ****/

/************************ Includes ************************/

#include <stdio.h>    /* Entrada e sada  */
#include <stdlib.h>   /* Miscelnea       */
#include <string.h>   /* Funes strXXX() */
#include <math.h>     /* Operaes reais  */
#include <time.h>     /* Funo time()    */
#include "ASSEGURA.h" /* Macro ASSEGURA   */
#include "Lista.h"    /* Interface deste mdulo */

/****************** Definies de Tipos *******************/

typedef enum {CARA, COROA} tCaraCoroa;

/********************* Funes Locais *********************/

/****
 *
 * CaraOuCoroa(): Simula o lanamento de uma moeda
 *
 * Parmetros: Nenhum
 *
 * Retorno: CARA se o nmero aleatrio gerado for par;
 *          COROA se o nmero aleatrio gerado for mpar
 *
 * Observaes: Na primeira chamada desta funo, ela
 *              alimenta o gerador de nmeros aleatrios
 *
 ****/
static tCaraCoroa CaraOuCoroa(void)
{
   static int primeiraChamada = 1;

   if (primeiraChamada) {
         /* Inicia o gerador de nmeros aleatrios */
      srand(time(NULL));

         /* A prxima chamada no ser mais a primeira */
      primeiraChamada = 0;
   }

      /* CARA corresponde a um nmero par e COROA corresponde */
      /* a um nmero mpar, mas poderia ser o contrrio       */
   return rand()%2 ? CARA : COROA;
}

/****
 *
 * ComprimentoPorNivel(): Calcula o nmero de ns de uma lista
 *                        com saltos num determinado nvel
 *
 * Parmetros:
 *     lista (entrada) - lista com saltos cujo comprimento
 *                       ser calculado
 *     nivel (entrada) - o referido nvel
 *
 * Retorno: O nmero de ns da lista no nvel especificado
 *
 ****/
static int ComprimentoPorNivel(tListaComSaltos lista, int nivel)
{
   int  tamanho = 0; /* Armazena o referido nmero de ns */
   tNoListaComSalto *ptrNo = lista.cabeca;

      /* Acessa cada n da lista no nvel especificado */
      /* e conta quantos ns so acessados             */
   while (ptrNo->proximo[nivel] != lista.cabeca) {
      ++tamanho; /* Mais um n foi encontrado */

         /* Passa para o prximo n neste nvel */
      ptrNo = ptrNo->proximo[nivel];
   }

   return tamanho;
}

/******************* Funes de Depurao *****************/

/****
 *
 * ApresentaComprimentosPorNivel(): Apresenta o nmero de ns
 *                                  de uma lista com saltos em
 *                                  cada nvel
 *
 * Parmetros:
 *     lista (entrada) - a referida lista
 *
 * Retorno: Nada
 *
 ****/
void ApresentaComprimentosPorNivel(tListaComSaltos lista)
{
   int i,
       total = ComprimentoPorNivel(lista, 0);

   printf("\n>>> Numero total de nos: %d <<<\n", total);

   printf( "\nNivel\tComprimento\tEsperado"
           "\n=====\t===========\t========\n");

      /* Exibe na tela nmero de ns por nvel da lista */
   for (i = 0; i < lista.nivelMax; ++i) {
      printf( "\n %2d\t  %6d\t %6d", i,
              ComprimentoPorNivel(lista, i),
             (int) lround(total/pow(2, i)) );
   }

   putchar('\n'); /* Embelezamento */
}

/****
 *
 * ApresentaRelacoesEntreNiveis(): Apresenta relaes entre
 *                                 nmeros de ns de uma lista
 *                                 com saltos em nveis
 *                                 consecutivos
 *
 * Parmetros:
 *     lista (entrada) - a referida lista
 *
 * Retorno: Nada
 *
 ****/
void ApresentaRelacoesEntreNiveis(tListaComSaltos lista)
{
   int i, nos1, nos2;

   printf( "\n\n    *** Relacoes entre Numeros de Nos ***"
           "\n        =============================\n");

      /* Exibe na tela nmero de ns por nvel da lista */
   for (i = 0; i < lista.nivelMax; ++i) {
      nos1 = ComprimentoPorNivel(lista, i);
      nos2 = ComprimentoPorNivel(lista, i + 1);

      printf( "\n>>> Nos no nivel %2d / Nos no nivel %2d:  ",
              i, i + 1);

      if (nos2) {
         printf("%6.4f", (double) nos1/nos2);
      } else {
         printf("------");
      }
   }

   putchar('\n'); /* Embelezamento */
}

/****
 *
 * ApresentaNiveis(): Apresenta o nmero mximo de nveis e o
 *                    nmero efetivo de nveis de uma lista
 *                    com saltos
 *
 * Parmetros:
 *     lista (entrada) - a referida lista
 *
 * Retorno: Nada
 *
 ****/
void ApresentaNiveis(tListaComSaltos lista)
{
   printf( "\n\t>>> Numero maximo de niveis na lista:   %d",
           lista.nivelMax + 1 );
   printf( "\n\t>>> Numero de niveis efetivos na lista: %d\n",
           lista.nivel + 1 );

   putchar('\n'); /* Embelezamento */
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaListaComSaltos(): Inicia uma lista com saltos
 *
 * Parmetros:
 *     *lista (sada) - a lista que ser iniciada
 *      maxRegs (entrada) - nmero mximo de registros
 *                          esperados na lista
 *
 * Retorno: Nada
 *
 ****/
void IniciaListaComSaltos(tListaComSaltos *lista, int maxRegs)
{
   int i, nMax;

      /* Determina o nvel mximo de um n da lista */
   nMax = lista->nivelMax = (int) floor(log2(maxRegs) )+ 1;

      /* Aloca a cabea da lista */
   lista->cabeca = malloc(sizeof(tNoListaComSalto));

      /* Se a alocao no foi possvel, aborta o programa */
   ASSEGURA(lista->cabeca, "Nao foi possivel alocar cabeca");

      /* Aloca o array de ponteiros da cabea */
   lista->cabeca->proximo = malloc((nMax + 1)*sizeof(tNoListaComSalto *));

      /* Se a alocao no foi possvel, aborta o programa */
   ASSEGURA(lista->cabeca->proximo, "Nao foi possivel alocar cabeca");

      /* A lista  circular com cabea, de modo que, inicialmente, */
      /* cada ponteiro proximo[i] aponta para a cabea da lista    */
   for (i = 0; i <= lista->nivelMax; i++)
      lista->cabeca->proximo[i] = lista->cabeca;

   lista->nivel = 0; /* O nvel corrente da lista  0 */
}

/****
 *
 * EstaVaziaListaComSaltos(): Verifica se uma lista com saltos
 *                            est vazia
 *
 * Parmetros:
 *     lista (entrada) - a lista que ser verificada
 *
 * Retorno:
 *    1, quando a lista est vazia;
 *    0, em caso contrrio
 *
 ****/
int EstaVaziaListaComSaltos(tListaComSaltos lista)
{
      /* Todos os ns da lista aparecem no nvel 0. Logo, */
      /* se o ponteiro para o primeiro n desse nvel     */
      /* aponta para a cabea, a lista est vazia.        */
   return lista.cabeca->proximo[0] == lista.cabeca;
}

/****
 *
 * ComprimentoListaComSalto(): Calcula o nmero de ns de uma
 *                             lista com saltos
 *
 * Parmetros:
 *     lista (entrada) - lista cujo comprimento ser calculado
 *
 * Retorno: O nmero de ns da referida lista
 *
 ****/
int ComprimentoListaComSalto(tListaComSaltos lista)
{
   int  tamanho = 0; /* Armazena o nmero de ns da lista */
   tNoListaComSalto *ptrNo = lista.cabeca; /* Apontar para cada */
                              /* n da lista        */

      /* Acessa cada n no primeiro nvel da lista */
      /* e conta quantos ns so acessados         */
   while (ptrNo->proximo[0] != lista.cabeca) {
      ++tamanho; /* Mais um n foi encontrado */
      ptrNo = ptrNo->proximo[0]; /* Passa para o prximo n */
   }

   return tamanho;
}

/****
 *
 * BuscaListaComSaltos(): Retorna o endereo do contedo do
 *                        n que possui uma determinada chave
 *                        numa lista com saltos
 *
 * Parmetros:
 *      lista (entrada) - lista que ser pesquisada
 *      chave (entrada) - chave de busca
 *
 * Retorno: Valor associado  chave, se ela for encontrada;
 *          caso contrrio, -1.
 *
 ****/
int BuscaListaComSaltos(tListaComSaltos lista, tCEP chave)
{
   int               i; /* Representar um nvel da lista  */
   tNoListaComSalto *ptrNo = lista.cabeca; /* Ponteiro utilizado para */
                                           /* visitar ns da lista    */

      /* A busca comea sempre no nvel mais alto */
      /* da lista e termina sempre no nvel 0     */
   for (i = lista.nivel; i >= 0; i--) {
         /* A busca prossegue no mesmo nvel at que a sentinela ou uma  */
         /* chave maior do que ou igual a chave de busca seja encontrada */
      while (ptrNo->proximo[i] != lista.cabeca &&
             strcmp(ptrNo->proximo[i]->conteudo.chave, chave) < 0)
            /* Passa para o prximo n no mesmo nvel */
         ptrNo = ptrNo->proximo[i];
   }

   /*****************************************************/
   /* Neste ponto o prximo n no nvel 0 adiante de    */
   /* ptrNo  a cabea da lista ou  o primeiro n cuja */
   /* chave  maior do que ou igual  chave de busca    */
   /*****************************************************/

      /* Faz ptrNo apontar para a cabea da lista ou para o primeiro */
      /* n cuja chave  maior do que ou igual a chave de busca      */
   ptrNo = ptrNo->proximo[0];

      /* Verifica se a chave foi encontrada */
   if ( ptrNo != lista.cabeca && !strcmp(ptrNo->conteudo.chave, chave) )
      return ptrNo->conteudo.valor; /* Chave encontrada */

   return -1; /* A chave no foi encontrada */
}

/****
 *
 * InsereListaComSaltos(): Insere um novo n numa lista com saltos
 *
 * Parmetros:
 *    *lista (entrada/sada) - lista na qual ser feita a insero
 *     conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Endereo do novo n ou NULL se a chave do novo
 *          elemento j se encontra na lista
 *
 * Observao: A chave  considerada primria
 *
 ****/
tNoListaComSalto *InsereListaComSaltos(tListaComSaltos *lista, tCEP_Ind conteudo)
{
   int                i, /* Representar um nvel da lista */
                      novoNivel; /* Nvel do novo n */
   tNoListaComSalto **enderecos, /* Apontar para um array */
                                 /* de ponteiros para ns  */
                     *novoNo, /* Ponteiro para o n a ser criado */
                     *ptrNo; /* Ponteiro utilizado para visitar ns da lista */

      /* Aloca um array auxiliar de ponteiros para ns.  */
      /* Esses ponteiros apontaro para ns que tero    */
      /* ponteiros 'proximo' apontando para o novo n.   */
   enderecos = calloc(lista->nivelMax + 1, sizeof(tNoListaComSalto *));

      /* Se o array no foi alocado aborta o programa */
   ASSEGURA(enderecos, "Nao foi possivel alocar array");

         /**********************************/
         /* Encontra a posio de insero */
         /**********************************/

   ptrNo = lista->cabeca; /* Faz ptrNo apontar para a cabea da lista */

      /* A busca comea sempre no nvel mais alto e termina no nvel 0 */
   for (i = lista->nivel; i >= 0; i--) {
         /* A busca prossegue no mesmo nvel at que a sentinela ou uma  */
         /* chave maior do que ou igual a chave de busca seja encontrada */
      while ( ptrNo->proximo[i] != lista->cabeca &&
              strcmp(ptrNo->proximo[i]->conteudo.chave, conteudo.chave ) < 0)
         ptrNo = ptrNo->proximo[i]; /* Passa para o prximo n no mesmo nvel */

         /* O array enderecos[] armazena o endereo do */
         /* ltimo n visitado no nvel i antes que    */
         /* se atinja o local da possvel insero     */
      enderecos[i] = ptrNo;
   }

      /* Verifica se a chave foi encontrada. Aqui, a chave */
      /*  considerada primria e no ser duplicada.      */
   if ( ptrNo->proximo[0] != lista->cabeca &&
        !strcmp( ptrNo->proximo[0]->conteudo.chave, conteudo.chave ) ) {

      /* A chave foi encontrada */

      free(enderecos); /* Este array no  mais necessrio */

      return NULL; /* Retorna NULL indicando que no houve insero */
   }

      /* Determina o nvel do novo n por meio do lanamento de uma moeda;  */
      /* i.e., enquanto o resultado do lanamento for CARA, o nvel do novo */
      /* n vai aumentando (se preferir COROA, o resultado ser o mesmo)    */
   for (novoNivel = 0; CaraOuCoroa() == CARA &&
                       novoNivel < lista->nivelMax; novoNivel++)
      ; /* Vazio */

   novoNo = malloc(sizeof(tNoListaComSalto));  /* Cria um n novo */

      /* Se a alocao no ocorreu, aborta o programa */
   ASSEGURA(novoNo, "Impossivel criar um no'");

      /* Cria o array ponteiros do novo n */
   novoNo->proximo = malloc((novoNivel + 1)*sizeof(tNoListaComSalto *));

      /* Se a alocao no ocorreu, aborta o programa */
   ASSEGURA(novoNo->proximo, "Impossivel criar um no'");

   novoNo->conteudo = conteudo; /* Armazena o contedo do novo n */

      /* Se o nvel do novo n for maior do que o nvel   */
      /* corrente da lista, os ponteiros 'proximo' dos    */
      /* nveis excedentes do novo n apontaro para a    */
      /* cabea, j que os demais ns tm nveis menores  */
      /* e o novo n no dever apontar para nenhum deles */
   if (novoNivel > lista->nivel) {
      for (i = lista->nivel + 1; i <= novoNivel; i++)
            /* O array 'enderecos' armazena os endereos dos  */
            /* ns que apontaro para o novo n. Como o nvel */
            /* desse n  maior do que o nvel corrente, o    */
            /* nico n com nvel maior do que o nvel        */
            /* corrente que apontar para ele  a cabea.     */
         enderecos[i] = lista->cabeca;

         /* O nvel da lista passa a ser igual ao nvel do novo n */
      lista->nivel = novoNivel;
   }

      /* Ajusta os ponteiros do novo n e dos ns que o antecedem */
   for (i = 0; i <= novoNivel; i++) {
         /* Faz o novo n apontar para os ns para  */
         /* os quais os antecessores dele apontavam */
      novoNo->proximo[i] = enderecos[i]->proximo[i];

         /* Faz os antecessores do novo n apontarem para ele */
      enderecos[i]->proximo[i] = novoNo;
   }

   free(enderecos); /* O array auxiliar no  mais necessrio */

   return novoNo; /* Retorna o endereo do novo n */
}

/****
 *
 * RemoveListaComSaltos(): Remove de uma lista com saltos um
 *                         n que contm um valor especificado
 *
 * Parmetros:
 *     *lista (entrada/sada) - lista na qual ser feita a remoo
 *      chave (entrada) - chave do n a ser removido
 *
 * Retorno: 0, se a remoo for bem sucedida, ou 1, se o n a
 *          ser removido no for encontrado.
 *
 ****/
int RemoveListaComSaltos(tListaComSaltos *lista, tCEP chave)
{
   int                i; /* Representar um nvel da lista */
   tNoListaComSalto **enderecos, /* Apontar para um array */
                                 /* de ponteiro para ns   */
                     *ptrNo; /* Ponteiro utilizado para visitar ns da lista */

      /* Aloca um array auxiliar de ponteiros */
   enderecos = calloc(lista->nivelMax + 1, sizeof(tNoListaComSalto *));

      /* Se o array no foi alocado aborta o programa */
   ASSEGURA(enderecos, "Nao foi possivel alocar array");

         /*************************************/
         /* Procura o elemento a ser removido */
         /*************************************/

   ptrNo = lista->cabeca; /* Faz ptrNo apontar para a cabea da lista */

      /* A busca comea sempre no nvel mais alto e termina no nvel 0 */
   for (i = lista->nivel; i >= 0; i--) {
         /* A busca prossegue no mesmo nvel at que a sentinela ou uma  */
         /* chave maior do que ou igual a chave de busca seja encontrada */
      while ( ptrNo->proximo[i] != lista->cabeca &&
              strcmp( ptrNo->proximo[i]->conteudo.chave, chave ) < 0 )
         ptrNo = ptrNo->proximo[i]; /* Passa para o prximo n no mesmo nvel */

         /* O array enderecos[] armazena o endereo do ltimo n visitado */
         /* no nvel i antes  que se encontre o n a ser removido         */
      enderecos[i] = ptrNo;
   }

      /* Se o elemento no foi encontrado,  ele no pode ser removido */
   if ( ptrNo->proximo[0] == lista->cabeca ||
        strcmp( ptrNo->proximo[0]->conteudo.chave, chave ) ) {
      return 1; /* Elemento no foi encontrado */
   }

      /* Neste ponto, ptrNo aponta para o n antecessor  */
      /* daquele que ser removido. Agora, faz-se ptrNo  */
      /* apontar para esse ltimo n. Esta instruo no */
      /*  essencial, mas facilita a escrita das demais. */
   ptrNo = ptrNo->proximo[0];

      /* Faz os antecessores do n removido apontarem para  */
      /* o sucessor dele. A atualizao comea no nvel 0 e */
      /* encerra quando o nvel da lista for atingido ou    */
      /* quando for encontrado um n que no aponta para o  */
      /* n a ser removido.                                 */
   for (i = 0; i <= lista->nivel && enderecos[i]->proximo[i] == ptrNo; i++)
      enderecos[i]->proximo[i] = ptrNo->proximo[i]; /* Efetua o desvio */

      /* Libera o espao ocupado pelo n removido e pelo array auxiliar */
   free(ptrNo);
   free (enderecos);

      /*  possvel que o n removido tenha o mesmo nvel da lista e que ele */
      /* seja o nico n com esse nvel.  Logo, se esse for o caso, ajusta-  */
      /* se o nvel da  lista de modo que no exista nenhum ponteiro que     */
      /* emane da cabea e termine na prpria cabea.                        */
   while (lista->nivel > 0 &&
          lista->cabeca->proximo[lista->nivel] == lista->cabeca)
      lista->nivel--;

   return 0;
}

/****
 *
 * ProximoListaComSaltos(): Encontra o prximo n de uma lista
 *                          com saltos
 *
 * Parmetros:
 *     lista (entrada) - a referida lista com saltos
 *
 * Retorno: Endereo do contedo do prximo n, se ele existir.
 *          NULL, se no existir prximo n.
 *
 ****/
tCEP_Ind *ProximoListaComSaltos(tListaComSaltos lista)
{
      /* A varivel a seguir indica qual  o prximo n */
      /* a ser visitado. Quando ela  NULL, o prximo   */
      /* n a ser visitado  o primeiro da lista.       */
   static tNoListaComSalto *proximoNo = NULL;

      /* Verifica se a lista est vazia */
   if (lista.cabeca->proximo[0] != lista.cabeca) {
         /* Se 'proximoNo' for NULL, o prximo n ser o   */
         /* primeiro da lista. Caso contrrio, 'proximoNo' */
         /* aponta para o n que foi acessado na ltima    */
         /* chamada desta funo.                          */
      if (!proximoNo) {
            /* Obtm o primeiro da lista */
         proximoNo = lista.cabeca->proximo[0];
      } else {
            /* Obtm o prximo da lista */
         proximoNo = proximoNo->proximo[0];

            /* Se 'proximoNo' estiver apontando para a  */
            /* cabea da lista,  porque a lista acabou */
         if (proximoNo == lista.cabeca) {
            proximoNo = NULL; /* A lista acabou */
         }
      }

         /* Se houver prximo n, retorna seu */
         /* contedo. Seno, retorna NULL.    */
      return proximoNo ? &proximoNo->conteudo : NULL;
   } else {
      return NULL; /* A lista est vazia */
   }
}

/****
 *
 * DestroiListaComSaltos(): Libera o espao ocupado pelos ns de
 *                          uma lista com saltos, tornando-a vazia
 *
 * Parmetros:
 *     lista (entrada/sada) - lista com saltos que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiListaComSaltos(tListaComSaltos lista)
{
   tNoListaComSalto *p, /* Aponta para o prximo n a ser liberado */
       *q; /* Ponteiro auxiliar */

       /* Verifica se a lista est vazia */
   if (lista.cabeca->proximo[0] == lista.cabeca) {
         /* Libera array de ponteiros da cabea */
         /* e depois libera a prpria cabea    */
      free(lista.cabeca->proximo);
      free(lista.cabeca);

      return; /* Lista vazia no precisa ser destruda */
   }

      /* Faz p e q apontarem para o incio da lista */
   q = p = lista.cabeca->proximo[0];

      /* Visita cada n da lista liberando-o */
   do {
         /* Faz q apontar para o prximo n antes */
         /* que o n corrente seja destrudo      */
      q = q->proximo[0];

         /* Libera o array de ponteiros do n      */
         /* corrente e depois libera o prprio n  */
      free(p->proximo);
      free(p);

         /* Faz p apontar para o prximo n,   */
         /* cujo endereo est armazenado em q */
      p = q;
   } while (p != lista.cabeca);

      /* Libera array de ponteiros da cabea */
      /* e depois libera a prpria cabea    */
   free(lista.cabeca->proximo);
   free(lista.cabeca); /* Libera a cabea */
}
