/****
 *
 * Exemplo: Piso de Chave
 *
 * Seo: 3.7.3
 *
 * Descrio:
 *    -> A tabela de busca  implementada como tabela indexada
 *       sem ordenao contendo as chaves e os ndices dos
 *       registros
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Arquivo de dados: CEPs.bin
 *
 * Observao: Crie o arquivo de dados no diretrio do programa
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* Funo strtod()          */
#include <time.h>     /* Medio de tempo         */
#include "Tabela.h"    /* Interface do tipo tabela  */
#include "Leitura.h"  /* Leitura via teclado      */
#include "Registros.h" /* Interface do mdulo Registros */
#include "ASSEGURA.h"  /* Macro ASSEGURA()        */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo de dados */
#define NOME_ARQUIVO_BIN "CEPs.bin"

   /* Nome do arquivo contendo as chaves da tabela de busca */
#define NOME_ARQUIVO_CHAVES "ListaDeChaves.txt"

/***************** Definies de Funes ******************/

/****
 *
 * EscreveListaEmArquivo(): Escreve em arquivo, em forma de
 *                          tabela, os elementos de uma tabela
 *                          armazenada num array
 *
 * Parmetros:
 *     tabela (entrada) - a tabela
 *     tamanho (entrada) - tamanho da tabela
 *
 * Retorno: Nada
 *
 ****/
void EscreveListaEmArquivo(tTabelaIdx tabela, int tamanho)
{
   tCEP_Ind   elemento;
   FILE       *stream;

      /* Tenta abrir o arquivo para escrita em modo texto */
   stream = fopen(NOME_ARQUIVO_CHAVES, "w");

      /* Se arquivo no foi aberto, aborta o programa */
   ASSEGURA(stream, "Impossivel criar arquivo para escrita de chaves");

   printf("\nEscrevendo tabela em arquivo...");

      /* Cabealho de apresentao */
   fprintf( stream,
            "   Chave   \tIndice\n"
            "   =====   \t======\n");

      /* Apresenta cada estrutura numa linha separada */
   for (int i = 0; i < tamanho; ++i) {
      elemento = ObtemElementoIdx(tabela, i);
      fprintf( stream, "%s\t%d\n", elemento.chave, i );
   }

   fclose(stream);

   printf("\n*** Resultado escrito no arquivo %s *** \n", NOME_ARQUIVO_CHAVES);
}

/****
 *
 * ExibeLista(): Apresenta na tela ou em arquivo
 *               os elementos da tabela de busca
 *
 * Parmetros:
 *     tabela (entrada) - a tabela
 *
 * Retorno: Nada
 *
 ****/
void ExibeLista(tTabelaIdx tabela)
{
   tCEP_Ind elemento;
   int       tamanho = ComprimentoIdx(tabela),
             op;

      /* Se a tabela estiver vazia, no h o que exibir */
   if (!tamanho) {
      printf("\n*** Lista Vazia *** \n");
      return;
   }

      /* Se a tabela for considerada grande demais, oferece   */
      /* ao usurio a opo de exbir apenas alguns elementos */
   if (tamanho > 1000) {
      printf( "\nA tabela possui %d elementos. "
              "\nTem certeza que quer exibi-la inteira (s/n)? ",
              tamanho );
      op = LeOpcao("snSN");

      if (op == 'n' || op == 'N') {
         printf("\nQuantos elementos deseja exibir? ");
         tamanho = LeInteiroEntre(1, tamanho);
      } else {
            /* No faz sentido exibir uma */
            /* tabela to grande na tela  */
         EscreveListaEmArquivo(tabela, tamanho);

         return; /* Servio concludo */
      }
   }

      /* Cabealho de apresentao na tela */
   printf( "\n   Chave   \tIndice"
           "\n   =====   \t======\n");

      /* Apresenta cada estrutura numa linha separada */
   for (int i = 0; i < tamanho; ++i) {
      elemento = ObtemElementoIdx(tabela, i);
      printf("%s\t%d\n", elemento.chave, i);
   }

   putchar('\n'); /* Embelezamento apenas */
}

/****
 *
 * ConstroiTabela(): L arquivo de dados (binrio) e armazena a
 *                   chave de cada registro numa tabela indexada
 *
 * Parmetros:
 *      arqBin (entrada) - nome do arquivo binrio
 *      tabela (sada) - tabela que conter as chaves dos
 *                       registros lidos
 *
 * Retorno: Stream associado ao arquivo aberto para
 *          leitura e escrita
 *
 ****/
FILE *ConstroiTabelaIdx( const char *arqBin, tTabelaIdx tabela )
{
   FILE      *stream;
   tCEP_Ind  umElemento;
   tRegistroCEP  umRegistro;
   int        indiceReg = 0;

      /* Tenta abrir o arquivo binrio para leitura e escrita */
   stream = fopen(arqBin, "rb+");

      /* Se o arquivo no pode ser aberto, aborta o programa */
   ASSEGURA(stream, "Arquivo nao pode ser aberto");

      /* L cada registro do arquivo e acrescenta */
      /* sua chave (CEP) na tabela                 */
   while (1) {
         /* L um registro no arquivo */
      fread( &umRegistro, sizeof(umRegistro), 1, stream);

         /* Verifica se ocorreu erro ou o */
         /* final do arquivo foi atingido */
      if (ferror(stream) || feof(stream)) {
         break;
      }

         /* A chave no  armazenada como string no registro. */
         /* Portanto, no se pode usar strcpy().              */
      strncpy(umElemento.chave, umRegistro.CEP, TAM_CEP);

         /* Transforma a chave num string. Isso no  */
         /* essencial, mas facilita o processamnto.   */
      umElemento.chave[TAM_CEP] = '\0';

         /* Acrescenta o ndice ao elemento e ao */
         /* mesmo tempo incrementa esse ndice   */
      umElemento.valor = indiceReg++;

         /* Acrescenta a chave na tabela. O valor associado */
         /*  chave  sua posio (ndice) na tabela.       */
      AcrescentaElementoIdx(tabela, &umElemento);
   }

      /* Se ocorreu erro de leitura no arquivo, fecha */
      /* o arquivo e aborta o programa                */
   if (ferror(stream)) {
      fclose(stream);
      ASSEGURA(0, "Erro de leitura de arquivo ");
   }

   return stream;
}

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
static void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\nEscolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t(%d) %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * AtualizaArquivoBin(): Cria um novo arquivo binrio contendo
 *                       cada registro de outro arquivo binrio
 *                       cuja chave no faa parte de uma tabela
 *                       de registros que devem ser removidos
 *
 * Parmetros:
 *      nomeArquivo (entrada) - nome do arquivo binrio original
 *      removidos (entrada) - tabela contendo as estruturas que
 *                            NO sero escritas no novo arquivo
 *
 * Retorno: Nenhum
 *
 ****/
void AtualizaArquivoBin( const char *nomeArquivo,
                         tTabelaIdx removidos )
{
   char      ar[L_tmpnam]; /* Nome de um arquivo temporrio */
   FILE     *stream, /* Stream associado ao arquivo original */
            *streamTmp; /* Stream associado ao temporrio */
   tRegistroCEP umRegistro; /* Um registro do arquivo */
   tCEP    umaChave;
   int       ocorreuErro = 0; /* Indicar se ocorreu erro */

      /* Tenta abrir o arquivo original para leitura */
   ASSEGURA( stream = fopen(nomeArquivo, "rb"),
             "Nao foi possivel atualizar arquivo" );

      /* Obtm o nome de um arquivo temporrio */
   tmpnam(ar);

      /* Tenta abrir o arquivo temporrio */
      /* para escrita em modo binrio     */
   ASSEGURA( streamTmp = fopen(ar, "wb"),
             "Nao foi possivel criar arquivo temporario" );

      /* O lao encerra quando no for */
      /* mais possvel ler um registro */
   while (1) {
         /* L um registro no arquivo de entrada */
      fread(&umRegistro, sizeof(umRegistro), 1, stream);

         /* Testa se o final do arquivo de entrada  */
         /* foi atingido ou ocorreu erro de leitura */
      if (feof(stream) || ferror(stream)) {
         break; /* Processamento encerrado */
      }

         /* Cria uma chave (string). O campo CEP */
         /* do registro no  string.            */
      strncpy(umaChave, umRegistro.CEP, TAM_CEP);
      umaChave[TAM_CEP] = '\0';

         /* Se a chave do registro lido no faz parte da  */
         /* tabela de removidos, escreve-o no novo arquivo */
      if (BuscaSequencialIdx(removidos, umaChave) < 0) {
            /* Escreve o registro lido no arquivo temporrio */
         fwrite(&umRegistro, sizeof(umRegistro), 1, streamTmp);

            /* Verifica se ocorreu erro de escrita */
         if (ferror(streamTmp)) {
            break;
         }
      }
   }

         /* Verifica se ocorreu erro de leitura ou escrita */
   if ( ferror(stream) || ferror(streamTmp) ) {
      ocorreuErro = 1;
   }

      /* Fecha os arquivos */
   fclose(stream);
   fclose(streamTmp);

   ASSEGURA(!ocorreuErro, "Erro na atualizacao do arquivo");

   ASSEGURA( !remove(nomeArquivo),
             "Impossivel remover o arquivo antigo" );

      /* Tenta rebatizar o arquivo temporrio */
      /* com o nome do arquivo original       */
   ocorreuErro = rename(ar, nomeArquivo);

   if (ocorreuErro) {
      printf( "\n\t>>> O resultado da atualizacao se encontra"
              "no arquivo \"%s\". Nao foi possivel "
              "rebatiza-lo.\n", ar );
      ASSEGURA( 0, "Erro em batismo de arquivo" );
   }
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa as operaes sobre listas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tTabelaIdx  tabela, /* Tabela de busca */
               removidos; /* Chaves dos registros a ser  */
                          /* removidos quando o arquivo  */
                          /* de registros for atualizado */
   tRegistroCEP umRegistro,    /* Armazenam temporariamente */
                outroRegistro; /* um registro               */
   tCEP_Ind   elemento, /* Um elemento da tabela */
              *pElemento; /* Pnteiro para um elemento da tabela */
   const char *opcoes[] = { "Acrescenta um CEP",
                            "Remove um CEP",
                            "Consulta dados de um CEP",
                            "Piso de um CEP",
                            "Altera dados de um CEP",
                            "Exibe a tabela de CEPs na tela",
                            "Encerra o programa"
                          };
   int         indice, /* ndice de um elemento */
                       /* da tabela ou registro  */
               op, /* Opo escolhida pelo usurio */
               nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]);
   char        umCEP[TAM_CEP + 1]; /* Uma chave de busca */
   FILE        *stream; /* Stream associado ao  */
                        /* arquivo de registros */
   int          primeiroByte; /* Indica a posio do */
                              /* primeiro byte de um */
                              /* registro no arquivo */

   MedidaDeTempo();
   printf("\nCriando tabela...");

      /* Inicia a tabela que armazena a tabela de busca */
   tabela = CriaTabelaIdx();

      /* Inicia a tabela que armazena os ndices */
      /* dos registros removidos                */
   removidos = CriaTabelaIdx();

      /* L o contedo do arquivo de */
      /* dados e o armazena na tabela */
   stream = ConstroiTabelaIdx(NOME_ARQUIVO_BIN, tabela);

   printf("\n...tabela criada com sucesso\n");
   MedidaDeTempo();

      /* O lao a seguir encerra quando o usurio */
      /* escolher a opo de encerramento         */
   while (1) {
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("1234567");

         /* Verifica se o usurio quer encerrar o programa */
      if (op == '7') { /* Encerra o programa */
         fclose(stream);

         DestroiTabelaIdx(tabela);

         break; /* Sada do lao */
      }

         /* Processa as demais opes */
      switch (op) {
         case '1': /* Acrescenta um registro de CEP */
               /* L dados do novo CEP */
            LeRegistro(&umRegistro);

               /* O valor do campo 'numero' corresponde  */
               /* posio do registro no arquivo de CEPs  */
               /* (isso  coisa dos Correios)             */
            umRegistro.numero = ComprimentoIdx(tabela) + 1;

               /* Cria a chave a ser inserida na tabela de */
               /* busca. Na tabela, a chabe  um string    */
               /* mas no registro CEP no  string         */
            strncpy(elemento.chave, umRegistro.CEP, TAM_CEP);

               /* Torna a chave um string */
            elemento.chave[TAM_CEP] = '\0';

               /* Acrescenta o ndice do registro ao elemento */
            elemento.valor = umRegistro.numero;

               /* Acrescemta novo elemento  tabela de busca */
            AcrescentaElementoIdx(tabela, &elemento);

               /* Move apontador de posio para o final */
               /* do arquivo. Se isso no for possvel,  */
               /* aborta o programa.                     */
            ASSEGURA( !fseek(stream, 0, SEEK_END),
                      "Erro de posicionamento em arquivo");

               /* Escreve o novo registro no final do arquivo */
            fwrite(&umRegistro, sizeof(umRegistro), 1, stream);

               /* Se ocorreu erro de escrita, aborta */
            ASSEGURA(!ferror(stream),
                     "Erro de escrita em arquivo ");

            printf("\n>>> Acrescimo bem sucedido\n");

            break;

         case '2': /* Remove um umCEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o CEP na tabela de busca */
            indice = BuscaSequencialIdx(tabela, umCEP);

               /* Se a chave for encontrada, remove-a da     */
               /* tabela de busca e acrescenta-se um novo    */
               /* elemento na tabela de removidos. A remoo */
               /* do registro do arquivo s ser efetuada    */
               /* ao final do programa.                      */
            if (indice >= 0) {
                  /* Remove chave da tabela de busca */
               (void) RemoveElementoIdx(tabela, indice);

                  /* Cria um novo elemento e acrescenta-o   */
                  /*  tabela de removidos. O campo 'valor' */
                  /* no  importanate aqui.                */
               strcpy(elemento.chave, umCEP);
               AcrescentaElementoIdx(removidos, &elemento);

               printf("\n>>> Remocao bem sucedida\n");
            } else { /* O CEP no se encontra na tabela */
               printf("\n>>> CEP nao encontrado\n");
            }

            break;

         case '3': /* Consulta dados de um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o CEP na tabela de busca */
            indice = BuscaSequencialIdx(tabela, umCEP);

               /* Se a chave for encontrada, utiliza-se */
               /* seu ndice para recuperar o registro  */
               /* correspondente no arquivo             */
            if (indice < 0) {
               printf("\n>>> CEP nao foi encontrado\n");
            } else {
                  /* Obtm o elemento da tabela que */
                  /* contm o ndice do registro    */
               elemento = ObtemElementoIdx(tabela, indice);

                  /* Utilizando o ndice do registro, */
                  /* calcula-se a posio no arquivo  */
                  /* do primeiro byte do registro     */
               primeiroByte = elemento.valor*sizeof(tRegistroCEP);

                  /* Move apontador de posio para o     */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
               ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                         "Erro de posicionamento em arquivo" );

                  /* L o registro no arquivo */
               fread(&umRegistro, sizeof(umRegistro), 1, stream);

                  /* Se ocorreu erro de leitura aborta */
               ASSEGURA(!ferror(stream), "Erro de leitura");

                  /* Apresenta o registro ao usurio */
               ExibeRegistro(&umRegistro);

               putchar('\n'); /* Embelezamento */
            }

            break;

         case '4': /* Encontra o piso de um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o piso na tabela de busca */
            pElemento = BuscaPisoIdx(tabela, umCEP);

               /* Se a chave for encontrada, utiliza-se */
               /* seu ndice para recuperar o registro  */
               /* correspondente no arquivo             */
            if (!pElemento) {
               printf("\n>>> Piso nao foi encontrado\n");
            } else {
                  /* Utilizando o ndice do registro, */
                  /* calcula-se a posio no arquivo  */
                  /* do primeiro byte do registro     */
               primeiroByte = pElemento->valor*sizeof(tRegistroCEP);

                  /* Move apontador de posio para o     */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
               ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                         "Erro de posicionamento em arquivo" );

                  /* L o registro no arquivo */
               fread(&umRegistro, sizeof(umRegistro), 1, stream);

                  /* Se ocorreu erro de leitura aborta */
               ASSEGURA(!ferror(stream), "Erro de leitura");

                  /* Apresenta o registro ao usurio */
               ExibeRegistro(&umRegistro);

               putchar('\n'); /* Embelezamento */
            }

            break;

         case '5': /* Altera dados de um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o CEP na tabela de busca */
            indice = BuscaSequencialIdx(tabela, umCEP);

            if (indice < 0) { /* Chave NO encontrada */
               printf("\n>>> CEP nao foi encontrado\n");
            } else {  /* Chave encontrada */
                  /* Obtm o elemento contendo a chave e o */
                  /* ndice do registro na tabela de busca */
               elemento = ObtemElementoIdx(tabela, indice);

                  /* Utilizando o ndice do registro, */
                  /* calcula-se a posio no arquivo  */
                  /* do primeiro byte do registro     */
               primeiroByte = elemento.valor*sizeof(tRegistroCEP);

                  /* Move o apontador de posio para o   */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
               ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                         "Erro de posicionamento em arquivo" );

                  /* L o registro no arquivo */
               fread(&umRegistro, sizeof(umRegistro), 1, stream);

                  /* Se ocorreu erro de leitura aborta */
               ASSEGURA(!ferror(stream), "Erro de leitura");

                  /* Obtm as alteraes introduzidas */
                  /* pelo usurio                     */
               SubstituiRegistro(&outroRegistro, &umRegistro);

               /***********************************************/
               /*                                             */
               /* >> Apresenta os dados novos e os antigos << */
               /* >> e solicita confirmao do usurio     << */
               /*                                             */
               /***********************************************/

               printf("\n\n\t***** Dados antigos *****\n");
               ExibeRegistro(&umRegistro);

               printf("\n\n\t***** Dados novos *****\n");
               ExibeRegistro(&outroRegistro);

               printf( "\n\n\t>>> Deseja substituir os dados antigos (s/n)? " );
               op = LeOpcao("snSN");

               if (op == 's' || op == 'S') {
                     /* Constri um novo elemento para */
                     /* substituir o antigo            */
                  strncpy(elemento.chave, outroRegistro.CEP,
                          TAM_CEP);
                  elemento.chave[TAM_CEP] = '\0';

                     /* Altera o valor do elemento na tabela */
                  AlteraElementoIdx(tabela, indice, &elemento);

                  /* Move apontador de posio para o     */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
                  ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                            "Erro de posicionamento em arquivo" );

                  /* Sobrescreve o registro antigo */
                  fwrite(&outroRegistro, sizeof(umRegistro), 1,
                         stream);

                     /* Se ocorreu erro de escrita aborta */
                  ASSEGURA(!ferror(stream), "Erro de escrita");

                     /* Garante que o novo registro estar */
                     /* no arquivo (e no no buffer) na    */
                     /* prxima consulta                   */
                  fflush(stream);

                  printf("\n>>> Alteracao bem sucedida\n");
               }
            }

            break;

         case '6': /* Apresenta a tabela na tela */
            ExibeLista(tabela);
            break;

         default: /* O programa no deve chegar at aqui */
            printf("\nEste programa contem um erro\n");
            return 1;
      }
   }

      /* Atualiza arquivo binrio se for necessrio */
   if (!EstaVaziaIdx(removidos)) {
      MedidaDeTempo();
      printf("\nAtualizando arquivo...");

         /* Atualiza arquivo binrio */
      AtualizaArquivoBin(NOME_ARQUIVO_BIN, removidos);

      printf("\n... Atualizacao bem sucedida");
      MedidaDeTempo();
   }

         /* Despede-se do usurio */
   printf( "\n\t>>> Obrigado por usar este programa.\n");

   return 0;
}

