/****
 *
 * Exemplo: Busca de Intervalo
 *
 * Seo: 3.7.4
 *
 * Descrio:
 *    -> A tabela de busca  implementada como lista ENCADEADA
 *       em que cada elemento contm uma a chave e a posio
 *       do respectivo registro no arquivo
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Tipo de busca: sequencial e busca de intervalo
 *
 *    -> Arquivo de dados: CEPs.bin
 *
 * Observao: Crie o arquivo de dados no diretrio do programa
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* Funo strtod()          */
#include <time.h>     /* Medio de tempo         */
#include "Tabela.h"   /* Interface do tipo lista  */
#include "Leitura.h"  /* Leitura via teclado      */
#include "Registros.h" /* Interface do mdulo Registros */
#include "ASSEGURA.h"  /* Macro ASSEGURA()        */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo de dados */
#define NOME_ARQUIVO_BIN "CEPs.bin"

   /* Nome do arquivo contendo as chaves da tabela de busca */
#define NOME_ARQUIVO_CHAVES "ListaDeChaves.txt"

   /* Nmero de registros exibidos na tela a cada vez */
#define MAX_REGISTROS_NA_TELA  4

/***************** Definies de Funes ******************/

/****
 *
 * EscreveListaEmArquivo(): Escreve em arquivo, em forma de
 *                          tabela, os elementos de uma lista
 *                          armazenada num array
 *
 * Parmetros:
 *     lista (entrada) - a lista
 *     tamanho (entrada) - tamanho da lista
 *
 * Retorno: Nada
 *
 ****/
void EscreveListaEmArquivo(tListaSE lista, int tamanho)
{
   tCEP_Ind *ptrElemento;
   FILE      *stream;

      /* Tenta abrir o arquivo para escrita em modo texto */
   stream = fopen(NOME_ARQUIVO_CHAVES, "w");

      /* Se arquivo no foi aberto, aborta o programa */
   ASSEGURA(stream,
            "Impossivel criar arquivo para escrita de chaves");

   printf("\nEscrevendo lista em arquivo...");

      /* Cabealho de apresentao */
   fprintf( stream,
            "   Chave   \tIndice\n"
            "   =====   \t======\n");

      /* Apresenta cada estrutura numa linha separada */
   while ((ptrElemento = ProximoListaSE(lista))) {
      fprintf( stream, "%s\t%d\n", ptrElemento->chave,
               ptrElemento->valor );
   }

   fclose(stream);

   printf( "\n*** Resultado escrito no arquivo %s *** \n",
           NOME_ARQUIVO_CHAVES );
}

/****
 *
 * ExibeLista(): Apresenta na tela ou em arquivo
 *               os elementos da tabela de busca
 *
 * Parmetros:
 *     lista (entrada) - a tabela
 *
 * Retorno: Nada
 *
 ****/
void ExibeLista(tListaSE lista)
{
   tCEP_Ind *ptrElemento;
   int        tamanho = ComprimentoListaSE(lista),
              op,
              i = 1;

      /* Se a lista estiver vazia, no h o que exibir */
   if (!tamanho) {
      printf("\n*** Lista Vazia *** \n");
      return;
   }

      /* Se a lista for considerada grande demais, oferece   */
      /* ao usurio a opo de exbir apenas alguns elementos */
   if (tamanho > 1000) {
      printf( "\nA lista possui %d elementos. "
              "\nTem certeza que quer exibi-la inteira (s/n)? ",
              tamanho );
      op = LeOpcao("snSN");

      if (op == 'n' || op == 'N') {
         printf("\nQuantos elementos deseja exibir? ");
         tamanho = LeInteiroEntre(1, tamanho);
      } else {
            /* No faz sentido exibir uma */
            /* tabela to grande na tela  */
         EscreveListaEmArquivo(lista, tamanho);

         return; /* Servio concludo */
      }
   }

      /* Cabealho de apresentao na tela */
   printf( "\n   Chave   \tIndice"
           "\n   =====   \t======\n");

      /* Apresenta cada estrutura numa linha separada */
      /* Apresenta cada estrutura numa linha separada */
   while ( (ptrElemento = ProximoListaSE(lista)) &&
           i++ <= tamanho ) {
      printf("%s\t%d\n", ptrElemento->chave, ptrElemento->valor);
   }

   putchar('\n'); /* Embelezamento apenas */
}

/****
 *
 * ConstroiTabela(): L arquivo de binrio e armazena as chaves e
 *              os ndices dos registro numa lista encadeada
 *
 * Parmetros:
 *      arqBin (entrada) - nome do arquivo binrio
 *     *lista (sada) - lista que conter as chaves e os
 *                      ndices dos registros lidos
 *
 * Retorno: Stream associado ao arquivo aberto para
 *          leitura e escrita
 *
 ****/
FILE *ConstroiTabela( const char *arqBin, tListaSE *lista )
{
   FILE      *stream;
   tCEP_Ind  umElemento;
   tRegistroCEP  umRegistro;
   int        indiceReg = 0;

      /* Tenta abrir o arquivo binrio para leitura e escrita */
   stream = fopen(arqBin, "rb+");

      /* Se o arquivo no pode ser aberto, aborta o programa */
   ASSEGURA(stream, "Arquivo nao pode ser aberto");

      /* L cada registro do arquivo e acrescenta */
      /* sua chave (CEP) e seu ndice na lista    */
   while (1) {
         /* L um registro no arquivo */
      fread( &umRegistro, sizeof(umRegistro), 1, stream);

         /* Verifica se ocorreu erro ou o */
         /* final do arquivo foi atingido */
      if (ferror(stream) || feof(stream)) {
         break;
      }

         /* A chave no  armazenada como string no registro. */
         /* Portanto, no se pode usar strcpy().              */
      strncpy(umElemento.chave, umRegistro.CEP, TAM_CEP);

         /* Transforma a chave num string. Isso no  */
         /* essencial, mas facilita o processamnto.   */
      umElemento.chave[TAM_CEP] = '\0';

         /* Armazena no novo elemento da tabela o */
         /* ndice do registro no arquivo e, ao   */
         /* mesmo tempo, incrementa esse ndice   */
      umElemento.valor = indiceReg++;

         /* Acrescenta o novo elemento  tabela */
      InsereNoLSE(lista, &umElemento);
   }

      /* Se ocorreu erro de leitura no arquivo, fecha */
      /* o arquivo e aborta o programa                */
   if (ferror(stream)) {
      fclose(stream);
      ASSEGURA(0, "Erro de leitura de arquivo ");
   }

   return stream;
}

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
static void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\nEscolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t(%d) %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * AtualizaArquivoBin(): Cria um novo arquivo binrio contendo
 *                       cada registro de outro arquivo binrio
 *                       cuja chave no faa parte de uma lista
 *                       de registros que devem ser removidos
 *
 * Parmetros:
 *      nomeArquivo (entrada) - nome do arquivo binrio original
 *      removidos (entrada) - lista contendo as estruturas que
 *                            NO sero escritas no novo arquivo
 *
 * Retorno: Nenhum
 *
 ****/
void AtualizaArquivoBin( const char *nomeArquivo,
                         tListaSE removidos )
{
   char      ar[L_tmpnam]; /* Nome de um arquivo temporrio */
   FILE     *stream, /* Stream associado ao arquivo original */
            *streamTmp; /* Stream associado ao temporrio */
   tRegistroCEP umRegistro; /* Um registro do arquivo */
   tCEP    umaChave;
   int       ocorreuErro = 0; /* Indicar se ocorreu erro */

      /* Tenta abrir o arquivo original para leitura */
   ASSEGURA( stream = fopen(nomeArquivo, "rb"),
             "Nao foi possivel atualizar arquivo" );

      /* Obtm o nome de um arquivo temporrio */
   tmpnam(ar);

      /* Tenta abrir o arquivo temporrio */
      /* para escrita em modo binrio     */
   ASSEGURA(streamTmp = fopen(ar, "wb"),
            "Nao foi possivel abrir arquivo temporario");

      /* O lao encerra quando no for */
      /* mais possvel ler um registro */
   while (1) {
         /* L um registro no arquivo de entrada */
      fread(&umRegistro, sizeof(umRegistro), 1, stream);

         /* Testa se final do arquivo de entrada foi */
         /* atingido ou ocorreu erro de leitura      */
      if (feof(stream) || ferror(stream)) {
         break; /* Processamento encerrado */
      }

         /* Cria uma chave (string). O campo CEP */
         /* do registro no  string.            */
      strncpy(umaChave, umRegistro.CEP, TAM_CEP);
      umaChave[TAM_CEP] = '\0';

         /* Se a chave do registro lido no faz parte da  */
         /* lista de removidos, escreve-o no novo arquivo */
      if (BuscaSequencialLSE(removidos, umaChave)) {
            /* Escreve o registro lido no arquivo de sada */
         fwrite(&umRegistro, sizeof(umRegistro), 1, streamTmp);

            /* Verifica se ocorreu erro de escrita */
         if (ferror(streamTmp)) {
            break;
         }
      }
   }

      /* Verifica se ocorreu erro de leitura ou escrita */
   if ( ferror(stream) || ferror(streamTmp) ) {
      ocorreuErro = 1;
   }

      /* Fecha os arquivos */
   fclose(stream);
   fclose(streamTmp);

      /* Se ocorreu erro de leitura ou */
      /* escrita, aborta o programa    */
   ASSEGURA(!ocorreuErro, "Erro na atualizacao do arquivo");

      /* Tenta apagar o arquivo original. Se isso */
      /* no for possvel, aborta o programa      */
   ASSEGURA( !remove(nomeArquivo),
             "Erro na atualizacao do arquivo" );

      /* Tenta rebatizar o arquivo temporrio */
      /* com o nome do arquivo original       */
   ocorreuErro = rename(ar, nomeArquivo);

   if (ocorreuErro) {
      printf( "\n\t>>> O resultado da atualizacao se encontra"
              "no arquivo \"%s\". Nao foi possivel "
              "rebatiza-lo.\n", ar );
      ASSEGURA( 0, "Erro em batismo de arquivo" );
   }
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/****
 *
 * ExibeEncontrados(): Apresenta lista de registros encontrados
 *                     numa busca
 *
 * Parmetros:
 *      lista (entrada) - lista na qual cada elemento contm
 *                        uma chave e o ndice do respectivo
 *                       registro no arquivo
 *      stream (entrada) - stream associado ao arquivo que
 *                         armazena os registros
 *
 * Retorno: Nmero de registros exibidos na tela
 *
 * Observao: O arquivo deve estar aberto em modo binrio
 *             que permita leitura
 *
 ****/
int ExibeEncontrados(tListaSE lista, FILE *stream)
{
   tCEP_Ind *ptrElemento; /* Apontar para o contedo */
                           /* de um n da lista        */
   int        ordem = 1, /* Nmero de ordem de um registro */
              restantes, /* Nmero de registros que */
                         /* faltam ser apresentados */
              nRegistros = ComprimentoListaSE(lista),
              op; /* Opo do usurio */

      /* Se o nmero de elementos da lista for zero, */
      /* esta funo no deveria ter sido chamada    */
   ASSEGURA(nRegistros, "Nao ha' registros para apresentar");

      /* Garante que a lista ser escrita */
      /* a partir do seu incio           */
   while (ProximoListaSE(lista)) {
      ; /* Vazio */
   }

      /* Apresenta um cabealho diferente se */
      /* a lista tiver apenas um elemento    */
   if (nRegistros == 1) {
      printf("\n************ Registro encontrado ************\n");
   } else {
      printf("\n************ Registro No. %d ************\n",
             ordem++);
   }

      /* Obtm o primeiro elemento da lista e exibe-o na tela */
   ptrElemento = ProximoListaSE(lista);
   ApresentaUmRegistro(ptrElemento->valor, stream);

      /* Obtm os demais elementos da lista e exibe-os */
   while ((ptrElemento = ProximoListaSE(lista))) {
      printf("\n\n************ Registro No. %d ************\n",
             ordem);
      ApresentaUmRegistro(ptrElemento->valor, stream);

         /* Calcula o nmero de registros */
         /* que faltam ser apresentados   */
      restantes = nRegistros - ordem;

         /* Se  restam registros a ser exibidos e o nmero  */
         /* mximo de registros apresentados por tela foi   */
         /* atingido, solicita confirmao do usurio antes */
         /* de continuar exibindo registros                 */
      if ((restantes > 0) && !(ordem%MAX_REGISTROS_NA_TELA)) {
         printf( "\n\n>>> Restam ainda %d registros.\n",
                 restantes );
         printf("\n>>> Digite 'c' para continuar ou 'e'"
                "\n>>> para encerrar a apresentacao: ");
         op = LeOpcao("cCeE");

            /* Se o usurio estiver satisfeito, retorna */
            /* o nmero de registros que foram exibidos */
         if (op == 'e' || op == 'E') {
            return ordem;
         }
      }

      ++ordem;
   }

   putchar('\n'); /* Embelezamento */

      /* Todos os registros foram exibidos */
   return nRegistros;
}

/********************* Funo main() **********************/

/****
 *
 * main(): Implementa operaes sobre tabela implementada como
 *         lista encadeada NO ordenada contendo os registros
 *         do tipo tCEP e com busca com movimentao para o incio
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tListaSE    lista, /* Tabela de busca */
               removidos, /* Chaves dos registros a ser  */
                          /* removidos quando o arquivo  */
                          /* de registros for atualizado */
               encontrados; /* Chaves encontradas numa */
                            /* busca de intervalo      */
   tRegistroCEP   umRegistro,    /* Armazenam temporariamente */
               outroRegistro; /* um registro               */
   tCEP_Ind   elemento, /* Um elemento da lista */
               *ptrElemento; /* Ponteiro para um  */
                             /* elemento da lista */
   const char *opcoes[] = { "Acrescenta um CEP",
                            "Remove um CEP",
                            "Consulta dados de um CEP",
                            "Busca de intervalo",
                            "Altera dados de um CEP",
                            "Apresenta a lista de CEPs",
                            "Encerra o programa"
                          };
   int         op, /* Opo escolhida pelo usurio */
               nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]);
   char        umCEP[TAM_CEP + 1],  /* Uma chave de busca   */
               umCEP2[TAM_CEP + 1]; /* Outra chave de busca */
   FILE        *stream; /* Stream associado ao  */
                        /* arquivo de registros */
   int          primeiroByte, /* Indica a posio do */
                              /* primeiro byte de um */
                              /* registro no arquivo */
                atualizarArq = 0; /* Indica se o arquivo */
                                  /* de registros ser   */
                                  /* atualizado          */

   MedidaDeTempo();
   printf("\nCriando lista...");

      /* Inicia a lista que armazena a tabela de busca */
   IniciaListaSE(&lista);

      /* Inicia a lista que armazena os ndices */
      /* dos registros removidos                */
   IniciaListaSE(&removidos);

      /* L o contedo do arquivo de */
      /* dados e o armazena na lista */
   stream = ConstroiTabela(NOME_ARQUIVO_BIN, &lista);

   printf("\n...lista criada com sucesso\n");
   MedidaDeTempo();

      /* O lao a seguir encerra quando o usurio */
      /* escolher a opo de encerramento         */
   while (1) {
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("1234567");

         /* Verifica se o usurio quer encerrar o programa */
      if (op == '7') { /* Encerra o programa */
         fclose(stream);

            /* A tabela de busca no  mais necessria */
         DestroiListaSE(&lista);

         break; /* Sada do lao */
      }

         /* Processa as demais opes */
      switch (op) {
         case '1': /* Acrescenta um registro de CEP */
               /* L dados do novo CEP */
            LeRegistro(&umRegistro);

               /* O valor do campo 'numero' corresponde  */
               /* posio do registro no arquivo de CEPs  */
               /* (isso  coisa dos Correios)             */
            umRegistro.numero = ComprimentoListaSE(lista) + 1;

               /* Cria a chave a ser inserida na tabela de */
               /* busca. Na tabela, a chave  um string    */
               /* mas no registro CEP no  string         */
            strncpy(elemento.chave, umRegistro.CEP, TAM_CEP);

               /* Torna a chave um string */
            elemento.chave[TAM_CEP] = '\0';

               /* Acrescenta o ndice do registro ao elemento */
            elemento.valor = umRegistro.numero;

               /* Insere novo elemento  tabela de busca */
            InsereNoLSE(&lista, &elemento);

               /* Move apontador de posio para o final */
               /* do arquivo. Se isso no for possvel,  */
               /* aborta o programa.                     */
            ASSEGURA( !fseek(stream, 0, SEEK_END),
                      "Erro de posicionamento em arquivo");

               /* Escreve o novo registro no final do arquivo */
            fwrite(&umRegistro, sizeof(umRegistro), 1, stream);

               /* Se ocorreu erro de escrita, aborta */
            ASSEGURA(!ferror(stream),
                     "Erro de escrita em arquivo ");

            printf("\n>>> Acrescimo bem sucedido\n");

            break;

         case '2': /* Remove um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Se a chave for encontrada, remove-a da    */
               /* tabela de busca e acrescenta-se um novo   */
               /* elemento na lista de removidos. A remoo */
               /* do registro do arquivo s ser efetuada   */
               /* ao final do programa.                     */
            if (!RemoveNoLSE(&lista, umCEP)) {
                  /* Cria um novo elemento e acrescenta-o */
                  /*  lista de removidos                 */
               strcpy(elemento.chave, umCEP);
               InsereNoLSE(&removidos, &elemento);

                  /* Arquivo de registros precisar */
                  /* ser atualizado                 */
               atualizarArq = 1;

               printf("\n>>> Remocao bem sucedida\n");
            } else {
               printf("\n>>> CEP nao encontrado\n");
            }

            break;

         case '3': /* Consulta dados de um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o CEP na tabela de busca */
            ptrElemento = BuscaSequencialLSE(lista, umCEP);

               /* Se a chave for encontrada, utiliza-se o */
               /* ndice associado a ela para recuperar   */
               /* o registro correspondente no arquivo    */
            if (!ptrElemento) {
               printf("\n>>> CEP nao foi encontrado\n");
            } else {
                  /* Utilizando o ndice do registro, */
                  /* calcula-se a posio no arquivo  */
                  /* do primeiro byte do registro     */
               primeiroByte = ptrElemento->valor *
                              sizeof(tRegistroCEP);

                  /* Move o apontador de posio para o   */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
               ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                         "Erro de posicionamento em arquivo" );

                  /* L o registro no arquivo */
               fread( &umRegistro, sizeof(umRegistro), 1,
                      stream );

                  /* Se ocorreu erro de leitura aborta */
               ASSEGURA(!ferror(stream), "Erro de leitura");

                  /* Apresenta o registro ao usurio */
               ExibeRegistro(&umRegistro);

               putchar('\n'); /* Embelezamento */
            }

            break;

         case '4': /* Busca de intervalo  */
               /* L o CEP que ser o limite inferior */
            LeMatricula("primeiro CEP", umCEP, TAM_CEP + 1);

               /* L o CEP que ser o limite superior */
            LeMatricula("segundo CEP", umCEP2, TAM_CEP + 1);

            /* Obtm uma lista de chaves que */
            /* casam com a chave de busca    */
         encontrados = BuscaIntervaloLSE(lista, umCEP, umCEP2);

            /* Se a chave for encontrada, apresenta  */
            /* os registros cujas chaves casam com a */
            /* chave de busca                        */
         if (EstaVaziaLSE(encontrados)) {
            printf("\n>>> Nenhuma chave foi encontrada\n");
         } else {
            printf("\n>>> Foram encontrados %d registros <<<\n",
                   ComprimentoListaSE(encontrados));
            (void) ExibeEncontrados(encontrados, stream);

               /* A lista de chaves encontradas no  */
               /* mais necessria e deve ser liberada */
            DestroiListaSE(&encontrados);
         }

            break;

         case '5': /* Altera dados de um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o CEP na tabela de busca */
            ptrElemento = BuscaSequencialLSE(lista,umCEP);

            if (!ptrElemento) { /* Chave NO encontrada */
               printf("\n>>> CEP nao foi encontrado\n");
            } else {
                  /* Utilizando o ndice do registro, */
                  /* calcula-se a posio no arquivo  */
                  /* do primeiro byte do registro     */
               primeiroByte = ptrElemento->valor *
                              sizeof(tRegistroCEP);

                  /* Move o apontador de posio para o   */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
               ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                         "Erro de posicionamento em arquivo" );

                  /* L o registro no arquivo */
               fread(&umRegistro, sizeof(umRegistro), 1, stream);

                  /* Se ocorreu erro de leitura aborta */
               ASSEGURA(!ferror(stream), "Erro de leitura");

                  /* Obtm as alteraes introduzidas */
                  /* pelo usurio                     */
               SubstituiRegistro(&outroRegistro, &umRegistro);

               /***********************************************/
               /*                                             */
               /* >> Apresenta os dados novos e os antigos << */
               /* >> e solicita confirmao do usurio     << */
               /*                                             */
               /***********************************************/

               printf("\n\n\t***** Dados antigos *****\n");
               ExibeRegistro(&umRegistro);

               printf("\n\n\t***** Dados novos *****\n");
               ExibeRegistro(&outroRegistro);

               printf( "\n\n\t>>> Deseja substituir os dados "
                       "antigos (s/n)? " );
               op = LeOpcao("snSN");

               if (op == 's' || op == 'S') {
                     /* Constri um novo elemento para       */
                     /* substituir o antigo. O campo         */
                     /* 'valor' no precisa ser alterado */
                  strncpy( elemento.chave, outroRegistro.CEP,
                           TAM_CEP );
                  elemento.chave[TAM_CEP] = '\0';

                     /* Altera o respectivo elemento */
                     /* na tabela de busca           */
                  AlteraConteudoLSE( lista, elemento.chave,
                                     &elemento );

                     /* Move apontador de posio para o     */
                     /* primeiro byte do registro. Se isso   */
                     /* no for possvel, aborta o programa. */
                  ASSEGURA( !fseek( stream, primeiroByte,
                                    SEEK_SET ),
                            "Erro de posicionamento em arquivo" );

                     /* Sobrescreve o registro antigo */
                  fwrite( &outroRegistro, sizeof(umRegistro), 1,
                          stream );

                     /* Se ocorreu erro de escrita aborta */
                  ASSEGURA(!ferror(stream), "Erro de escrita");

                     /* Garante que o novo registro estar */
                     /* no arquivo (e no no buffer) na    */
                     /* prxima consulta                   */
                  fflush(stream);

                  printf("\n>>> Alteracao bem sucedida\n");
               }
            }

            break;

         case '6': /* Apresenta a lista na tela */
            ExibeLista(lista);
            break;

         default: /* O programa no deve chegar at aqui */
            printf("\nEste programa contem um erro\n");
            return 1;
      }
   }

      /* Atualiza arquivo binrio se for necessrio */
   if (atualizarArq) {
      MedidaDeTempo();
      printf("\nAtualizando arquivo...");

         /* Atualiza arquivo binrio */
      AtualizaArquivoBin(NOME_ARQUIVO_BIN, removidos);

      printf("\n... Atualizacao bem sucedida");
      MedidaDeTempo();
   }

      /* Despede-se do usurio */
   printf( "\n\t>>> Obrigado por usar este programa.\n");

   return 0;
}

