/****
 *
 * Ttulo: Convertendo um Arquivo de Registros de Texto para Binrio
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 21/05/2012
 * ltima modificao: 21/05/2012
 *
 * Descrio: L cada linha de um arquivo de texto, converte-a numa
 *            estrutura do tipo tAluno e armazena-a num arquivo
 *            binrio
 *
 * Entrada: Arquivo cujo nome  especificado pela constante
 *          NOME_ARQUIVO_TEXTO
 *
 * Sada: Arquivo cujo nome  especificado pela constante
 *        NOME_ARQUIVO_BIN
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de Entrada e Sada   */
#include <string.h> /* Processamento de strings     */
#include <stdlib.h> /* strtod() e alocao dinmica */
#include "Registros.h" /* Interface deste mdulo    */
#include "Leitura.h"  /* Leitura via teclado        */
#include "ASSEGURA.h" /* Macro ASSEGURA             */

/********************* Funes Locais *********************/

/****
 *
 * ExibeArrayDeChars(): Exibe na tela um array de elementos
 *                      do tipo char que no  string
 *
 * Parmetros:
 *      *nome (entrada) - string que ser exibido antes do array
 *       ar (entrada) - endereo do array
 *       tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 ****/
static void ExibeArrayDeChars(const char *nome, const char *ar,
                              int tam)
{
   putchar('\n');

   if (*nome) {
      printf("%s: ", nome);
   }

   for (int i = 0; i < tam && *ar; ++i) {
      putchar(ar[i]);
   }
}

/****
 *
 * LeIntNaoNegativo(): L um string e tenta convert-lo
 *                     em nmero inteiro
 *
 * Parmetros: Nenhum
 *
 * Retorno: O valor lido e convertido se a converso
 *          for possvel; caso contrrio, -1
 *
 * Observao: Diferentemente do que ocorre com as funes
 *             do mdulo 'leitura', esta funo no insiste
 *             com o usurio
 *
 ****/
static int LeIntNaoNegativo(void)
{
   char  ar[MAX_NOME + 1];
   char *final;
   int   numero, resto;

      /* MAX_NOME  uma pssima escolha neste caso, mas  */
      /* este programa j tem constante simblica demais */
   resto = LeString(ar, MAX_NOME + 1);

      /* Se o usurio digitou apenas [ENTER] ou digitou */
      /* demais, retorna -1 informando que o usurio    */
      /* no quis digitar o valor esperado              */
   if (!*ar || resto) {
      return -1;
   }

      /* Tenta efetuar a converso em long int */
   numero = (int)strtol(ar, &final, 10);

      /* Se o string no foi totalmente convertido, */
      /* retorna -1 informando que o usurio no    */
      /* quis digitar o valor esperado              */
   if (*final) {
      return -1;
   }

      /* Se o nmero convertido for negativo, quem chamou */
      /* esta funo dever saber como lidar com o caso   */
   return numero;
}

/********************* Funes Globais ********************/

/****
 *
 * ExibeRegistro(): Exibe na tela um registro do tipo tRegistroCEP
 *
 * Parmetros: *registro (entrada) - registro que ser exibido
 *
 * Retorno: Nada
 *
 ****/
void ExibeRegistro(const tRegistroCEP *registro)
{
   printf("\nNumero: %d", registro->numero);

   ExibeArrayDeChars("UF", registro->UF, TAM_UF);

   printf("\nNumero de localidade: %d",
          registro->localidadeNumero);

   ExibeArrayDeChars("Nome abreviado", registro->nomeAbr,
                     MAX_NOME);

   ExibeArrayDeChars("Nome", registro->nome, MAX_NOME);

   printf("\nBairro inicio: %d", registro->bairroInicio);

   printf("\nBairro fim: %d", registro->bairroFim);

   ExibeArrayDeChars("CEP", registro->CEP, TAM_CEP);

   ExibeArrayDeChars("Complemento", registro->complemento,
                     MAX_COMP);

   ExibeArrayDeChars("Tipo de logradouro",
                     registro->tipoLogradouro, MAX_TIPO_LOG);

   printf("\nStatus de logradouro: %c",
          registro->statusTipoLogradouro);

   ExibeArrayDeChars("Nome sem acento:",
                     registro->nomeSemAcento, MAX_NOME);

   ExibeArrayDeChars("Chave DNE", registro->chaveDNE, TAM_DNE);
}

/****
 *
 * LeRegistro(): L os dados de um CEP
 *
 * Parmetros:
 *     *regCEP (sada) - estrutura que receber os
 *                       dados introduzidos pelo usurio
 *      cep (entrada) - se no for NULL, o
 * Retorno: O parmetro
 *
 ****/
tRegistroCEP *LeRegistro(tRegistroCEP *regCEP, const char *cep)
{
   char ar[MAX_COMP + 1];

   printf("\n\t>>> UF: ");
   LeString(ar, TAM_UF + 1);
   strncpy(regCEP->UF, ar, TAM_UF);

   printf("\n\t>>> Numero de localidade: ");
   regCEP->localidadeNumero = LeNatural();

   printf("\n\t>>> Nome abreviado: ");
   LeString(ar, MAX_NOME + 1);
   strncpy(regCEP->nomeAbr, ar, MAX_NOME);

   printf("\n\t>>> Nome: ");
   LeString(ar, MAX_NOME + 1);
   strncpy(regCEP->nome, ar, MAX_NOME);

   printf("\n\t>>> Bairro inicio: ");
   regCEP->bairroInicio = LeNatural();

   printf("\n\t>>> Bairro fim: ");
   regCEP->bairroFim = LeNatural();

   if (!cep) {
      printf("\n\t>>> CEP: ");
      LeString(ar, TAM_CEP + 1);
      strncpy(regCEP->CEP, ar, TAM_CEP);
   } else {
      strncpy(regCEP->CEP, cep, TAM_CEP);
   }

   printf("\n\t>>> Complemento: ");
   LeString(ar, MAX_COMP + 1);
   strncpy(regCEP->complemento, ar, MAX_COMP);

   printf("\n\t>>> Tipo de logradouro: ");
   LeString(ar, MAX_TIPO_LOG + 1);
   strncpy(regCEP->tipoLogradouro, ar, MAX_TIPO_LOG);

   printf("\n\t>>> Status tipo de logradouro: ");
   regCEP->statusTipoLogradouro = LeOpcao("SsNn");

   printf("\n\t>>> Nome sem acento: ");
   LeString(ar, MAX_NOME + 1);
   strncpy(regCEP->nomeSemAcento, ar, MAX_NOME);

   printf("\n\t>>> Chave DNE: ");
   LeString(ar, TAM_DNE + 1);
   strncpy(regCEP->chaveDNE, ar, TAM_DNE);

   return regCEP;
}

/****
 *
 * SubstituiRegistro(): L os dados de um registro que, se forem
 *                      vlidos, substituiro valores antigos
 *
 * Parmetros:
 *     *novo (sada) - estrutura que armazenar os novos
 *                     dados introduzidos pelo usurio
 *     *antigo (entrada) - estrutura contendo os dados antigos
 *
 * Retorno: O primeiro parmetro
 *
 * Observao: Esta funo no est rigorosamente bem
 *             implementada pois no faz parte do tema
 *             central deste exemplo
 *
 ****/
tRegistroCEP *SubstituiRegistro( tRegistroCEP *novo,
                              const tRegistroCEP *antigo )
{
   char ar[MAX_COMP + 1];
   int  num;

   printf("\n>>> ATENCAO: Dados invalidos nao substituem "
          "dados antigos <<<\n");

      /* L o valor de UF. Qualquer string que no seja  */
      /* vazio  considerado vlido, mas no deveria ser */
      /* assim (v. Observao). O mesmo vale para os     */
      /* demais strings.                                 */
   printf("\n\t>>> UF: ");
   LeString(ar, TAM_UF + 1);
   strncpy(novo->UF, *ar ? ar : antigo->UF, TAM_UF);

   printf("\n\t>>> Numero de localidade: ");
   num = LeIntNaoNegativo();
   novo->localidadeNumero = num >= 0 ?
                            num :
                            antigo->localidadeNumero;

   printf("\n\t>>> Nome abreviado: ");
   LeString(ar, MAX_NOME + 1);
   strncpy(novo->nomeAbr, *ar ? ar : antigo->nomeAbr, MAX_NOME);

   printf("\n\t>>> Nome: ");
   LeString(ar, MAX_NOME + 1);
   strncpy(novo->nome, *ar ? ar : antigo->nome, MAX_NOME);

   printf("\n\t>>> Bairro inicio: ");
   num = LeIntNaoNegativo();
   novo->bairroInicio = num >= 0 ?
                            num :
                            antigo->bairroInicio;

   printf("\n\t>>> Bairro fim: ");
   num = LeIntNaoNegativo();
   novo->bairroFim = num >= 0 ?
                            num :
                            antigo->bairroFim;

   printf("\n\t>>> CEP: ");
   LeString(ar, TAM_CEP + 1);
   strncpy(novo->CEP, *ar ? ar : antigo->CEP, TAM_CEP);

   printf("\n\t>>> Complemento: ");
   LeString(ar, MAX_COMP + 1);
   strncpy(novo->complemento, *ar ? ar : antigo->complemento,
           MAX_COMP);

   printf("\n\t>>> Tipo de logradouro: ");
   LeString(ar, MAX_TIPO_LOG + 1);
   strncpy(novo->tipoLogradouro,
           *ar ? ar : antigo->tipoLogradouro,
           MAX_TIPO_LOG);

   printf("\n\t>>> Status tipo de logradouro: ");
   num = LeCaractere();

   if (num == 's' || num == 'S') {
      novo->statusTipoLogradouro = 'S';
   } else if (num == 'n' || num == 'N') {
      novo->statusTipoLogradouro = 'N';
   } else {
      novo->statusTipoLogradouro = antigo->statusTipoLogradouro;
   }

   printf("\n\t>>> Nome sem acento: ");
   LeString(ar, MAX_NOME + 1);
   strncpy(novo->nomeSemAcento,
           *ar ? ar : antigo->nomeSemAcento, MAX_NOME);

   printf("\n\t>>> Chave DNE: ");
   LeString(ar, TAM_DNE + 1);
   strncpy(novo->chaveDNE, *ar ? ar : antigo->chaveDNE,
           TAM_DNE);

   return novo;
}

/****
 *
 * TamanhoDeArquivo(): Calcula o tamanho de um arquivo lendo cada
 *                     byte que o constitui e contando quantos
 *                     bytes so lidos
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *
 * Retorno: -1, se ocorrer algum erro.
 *          O tamanho do arquivo, em caso contrrio.
 *
 * Observaes:
 *     1. O arquivo deve estar aberto em modo binrio que permite
 *        leitura
 *     2. A posio atual do indicador de posio do arquivo 
 *        preservada
 *
 ****/
int TamanhoDeArquivo(FILE *stream)
{
   int posicaoAtual, tamanho = 0;

      /* Guarda a posio atual do indicador de posio */
      /* do arquivo para que ele possa ser restaurado   */
   posicaoAtual = ftell(stream);

      /* Verifica se ftell() foi bem sucedida */
   if (posicaoAtual < 0) {
      return -1; /* A chamada de ftell() falhou */
   }

      /* Assegura leitura a partir do incio do arquivo */
   if ( fseek(stream, 0, SEEK_SET) ) {
         /* O movimento do indicador de posio no foi   */
         /* possvel. Neste caso, no precisa restaur-lo */
      return -1;
   }

      /* O lao encerra quando o final do arquivo for */
      /* atingido ou se ocorrer erro de leitura       */
   while (1){
         /* L e descarta o prximo byte */
      fgetc(stream);

         /* Checa se o arquivo j foi inteiramente lido */
      if (feof(stream)) {
         break; /* Encerra lao */
      }

         /* Verifica se ocorreu erro de leitura */
      if (ferror(stream)) {
            /* Restaura a posio original do indicador */
            /* de posio do arquivo  antes de retornar */
         fseek(stream, posicaoAtual, SEEK_SET);
         return -1; /* Ocorreu erro de leitura */
      }

      ++tamanho; /* Mais um byte foi lido */
   }

      /* Restaura a posio original do indicador de posio */
   if ( fseek(stream, posicaoAtual, SEEK_SET) ) {
      return -1; /* O movimento no foi possvel */
   }

   return tamanho;
}

/****
 *
 * NumeroDeRegistros(): Calcula o nmero de registros (i.e.,
 *                      parties de tamanho especificado) de um
 *                      arquivo binrio
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *    tamRegistro (entrada) - tamanho de cada partio do arquivo
 *
 * Retorno: O nmero de parties do arquivo, se no ocorrer
 *          nenhum erro, ou -1, em caso contrrio.
 *
 * Observao: O arquivo deve estar aberto em modo binrio que
 *             permite leitura
 *
 ****/
int NumeroDeRegistros( FILE *stream, int tamRegistro )
{
   int    nBytes = 0; /* Nmero de bytes do arquivo */

      /* Calcula o nmero de bytes do arquivo */
   nBytes = TamanhoDeArquivo(stream);

      /* Verifica se ocorreu erro na contagem de bytes */
   if (nBytes < 0) {
      return -1; /* Erro na contagem de bytes */
   }

      /* O nmero de registros  o nmero de bytes do arquivo */
      /* dividido pelo nmero de bytes de cada registro       */
   return nBytes/tamRegistro;
}
