/****
 *
 * Exemplo: Exemplos de Anlise Amortizada
 *
 * Seo: 5.2.5
 *
 * Descrio:
 *    -> Uma estrutura de dados fictcia, tabela de destruio,
 *        implementada como lista simplesmente encadeada
 *
 *    -> As nicas operaes suportadas so
 *       1. Acrscimo, que acrescenta um elemento no incio da tabela.
 *       2. Destruio, que destri a tabela (i.e., libera o
 *          espao ocupado pela tabela).
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* Funes de alocao dinmica e */
                      /* gerao de nmeros aleatrios  */
#include <time.h>     /* Funo time() para alimentar o */
                      /* gerador de nmeros aleatrios  */

/************************ Macros **************************/

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

#define N_ELEMENTOS  1000
#define MAX_ELEMENTO 1000

/****************** Definies de Tipos *******************/

   /* Tipo de n e tipo de ponteiro para n de uma lista encadeada */
typedef struct rotNoLSE {
           int              conteudo;
           struct rotNoLSE *proximo;
        } tNoListaSE, *tListaSE;

/************************* Aluses ************************/

extern void IniciaListaSE(tListaSE *lista);
extern void InsereListaSE(tListaSE *lista, int conteudo);
extern int EstaVaziaListaSE(tListaSE lista);
extern void DestroiListaSE(tListaSE *lista);
extern void ApresentaMenu(void);

/***************** Definies de Funes ******************/

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'.
 *
 ****/
static int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeCaractere(): L um caractere via teclado e remove os
 *                caracteres remanescentes no buffer
 *
 * Parmetros: Nenhum
 *
 * Retorno: O caractere lido
 *
 * Observao: Esta funo deixa o buffer limpo
 *
 ****/
int LeCaractere(void)
{
   int c, /* Armazenar cada caractere lido */
       nResto = 0; /* Nmero de caracteres */
                   /* excedentes           */

   /* Volta para c se ocorrer erro de leitura */
inicio:
   c = getchar(); /* L o caractere digitado */

      /* Verifica se getchar() retornou EOF */
   if (c == EOF) {
      printf( "\a\n\t>>>Erro de leitura. "
              "Tente novamente\n\t> " );
      goto inicio; /* No faz mal algum */
   }

      /* Pelo menos '\n' se encontra no buffer */
   nResto = LimpaBuffer();

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf( "\t>>> %d caracteres foram descartados\n",
              nResto );
   }

   return c; /* Retorna o caractere lido */
}

/****
 *
 * LeInteiro(): L um nmero inteiro via teclado e deixa o buffer intacto
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero inteiro lido
 *
 ****/
int LeInteiro(void)
{
   int num, /* O nmero lido */
       teste, /* Valor retornado por scanf() */
       nResto = 0; /* Nmero de caracteres excedentes */

   /* Desvia para c se o valor for invlido */
inicio:
      /* Tenta ler um valor vlido */
   teste = scanf("%d", &num);

      /* Se no ocorreu erro de leitura ou de final */
      /* de arquivo, h caracteres remanescentes    */
      /* que precisam ser removidos                 */
   if (teste != EOF) {
      nResto = LimpaBuffer();
   }

      /* Enquanto o valor retornado por scanf()  */
      /* no indicar que um valor vlido foi     */
      /* lido continua tentando obter esse valor */
   while(teste != 1) {
      printf("\a\n\t>>> O valor digitado e' invalido. "
                      "Tente novamente\n\t> " );
      goto inicio; /* No causa dano algum */
   }

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf("\t>>> %d caracteres foram descartados\n", nResto);
   }

      /* O valor retornado certamente  vlido */
   return num;
}

/****
 *
 * LeOpcao(): L, via teclado, um caractere que deve fazer
 *            parte de um string para ser considerado vlido
 *
 * Parmetros:
 *      opcoes (entrada) - string contendo os caracteres
 *                         vlidos
 *
 * Retorno: Um caractere vlido
 *
 * Observaes:
 *      1. Esta funo deixa o buffer virgem novamente
 *      2. Esta funo  denominada 'LeOpcao' porque o
 *         caractere lido e validado  restrito a um
 *         conjunto de caracteres (opes) representado
 *         pelo string recebido como parmetro
 *
 ****/
int LeOpcao(const char *opcoes)
{
   int op; /* Opo (caractere) escolhida pelo usurio */

      /* Enquanto o usurio no escolher uma opo */
      /* vlida, o lao a seguir no encerra       */
   while (1) {
         /* L o caractere digitado */
      op = LeCaractere();

         /* Verifica se o caractere  vlido */
      if (strchr(opcoes, op)) {
         break; /*  */
      } else { /* No  */
         printf( "\a\n\t>>> Opcao incorreta. "
                "Tente novamente\n\t> " );
      }
   }

      /* Certamente, a opo retornada  vlida */
   return op;
}

/****
 *
 * IniciaListaSE(): Inicia uma lista simplesmente encadeada
 *
 * Parmetros: *lista (sada) - a lista que ser iniciada
 *
 * Retorno: Nada
 *
 * Observao: Esta funo independe do tipo de contedo
 *             efetivo armazenado em cada n da lista
 *
 ****/
void IniciaListaSE(tListaSE *lista)
{
   *lista = NULL;
}

/****
 *
 * EstaVaziaListaSE(): Verifica se uma lista simplesmente
 *                 encadeada est vazia
 *
 * Parmetros:
 *     lista (entrada) - a lista que ser verificada
 *
 * Retorno: 1, quando a lista est vazia, ou 0, em caso contrrio
 *
 ****/
int EstaVaziaListaSE(tListaSE lista)
{
   return lista == NULL;
}

/****
 *
 * InsereListaSE(): Insere um novo n no incio de uma
 *                lista simplesmente encadeada
 *
 * Parmetros:
 *    *lista (entrada/sada) - ponteiro para a lista na qual
 *                             ser feita a insero
 *     conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Nada
 *
 * Observao: Como a lista no obedece nenhuma ordem, pode-se
 *             inserir um n em qualquer posio da lista.
 *             Escolheu-se insero no incio da lista porque
 *              mais simples e eficiente.
 *
 ****/
void InsereListaSE(tListaSE *lista, int conteudo)
{
   tNoListaSE *ptrNovoNo; /* Apontar para o novo n alocado */

   ASSEGURA(ptrNovoNo = malloc(sizeof(tNoListaSE)), /* Tenta alocar */
            "Nao foi possivel alocar no'");  /* um novo n   */

   ptrNovoNo->conteudo = conteudo; /* Preenche contedo do n */

      /* Faz o novo n apontar para o primeiro n da lista */
   ptrNovoNo->proximo = *lista;

   *lista = ptrNovoNo; /* Torna o novo n o primeiro da lista */
}

/****
 *
 * DestroiListaSE(): Libera o espao ocupado pelos ns de
 *                   uma lista encadeada, tornando-a vazia
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro para a
 *                             lista simplesmente encadeada que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiListaSE(tListaSE *lista)
{
   tListaSE p; /* Aponta para o prximo n a ser liberado */

   if (!*lista)
      return; /* Lista vazia no precisa ser destruda */

   p = *lista; /* Faz p apontar para o incio da lista */

   do { /* Visita cada n da lista liberando-o */
      *lista = (*lista)->proximo;

      free(p); /* Libera o espao do n corrente */

      p = *lista; /* Faz p apontar para o prximo n */
   } while (p);
}

/****
 *
 * ApresentaMenu(): Apresenta o menu do programa
 *
 * Parmetros: No h
 *
 * Retorno: Nada
 *
 ****/
void ApresentaMenu(void)
{
   printf( "\n\t>>>>> Opcoes <<<<<\n"
           "\n\t(1) Acrescenta um elemento"
           "\n\t(2) Destroi a lista"
           "\n\t(3) Encerra o programa"
           "\n\n\t>>> Escolha sua opcao: " );
}

/****
 *
 * main(): Testa as operaes sobre listas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tListaSE lista;
   int      i, valor,op;

   IniciaListaSE(&lista);

      /* Ser criada uma lista inicial com 1000 elementos. */
      /* Os contedos destes elementos sero nmeros       */
      /* inteiros entre 0 e 1000 gerados aleatoriamente    */

      /* Alimenta o gerador de nmeros aleatrios */
   srand((int) time(NULL));

      /* Cria a lista inicial */
   for (i = 1; i <= N_ELEMENTOS; ++i) {
      valor = (int) rand()%MAX_ELEMENTO + 1;
      InsereListaSE(&lista, valor);
   }

   while (1) {
      ApresentaMenu();

      op = LeOpcao("123");

      if (op == '3') { /* Encerra o programa */
         printf("\nBye!\n");
         break; /* Sada do lao */
      }

      switch (op) {
         case '1': /* Acrescenta um elemento */
            printf("\n>>> Digite o valor a ser acrescentado: ");
            valor = LeInteiro();

            InsereListaSE(&lista, valor);
            printf("\n>>> Acrescimo efetuado\n");

            break;

         case '2': /* Destroi a lista */
            printf("\n>>> A lista sera' destruida <<<");
            DestroiListaSE(&lista);

            if (EstaVaziaListaSE(lista))
               printf("\n>>> Remocao bem sucedida\n");
            else
               printf("\n>>> Valor nao encontrado\n");

            break;

         default: /* O programa no deve chegar at aqui */
            printf("\nEste programa contem um erro\n");
            return 1;
      }
   }

   return 0;
}

/******************* Exemplo de Execuo ******************



**********************************************************/
