/****
 *
 * CriaArquivoBin.c: Programa que cria o arquivo binrio "CEPs.bin"
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* exit() e strtod()        */
#include <time.h>     /* Medio de tempo         */

/***************** Constantes Simblicas ******************/

#define MAX_NOME  40 /* Nmero mximo de caracteres num nome */
#define TAM_UF     2 /* Nmero de dgitos num UF */
#define TAM_CEP    8 /* Nmero de dgitos num CEP */
#define MAX_COMP  60 /* Nmero mximo de caracteres */
                     /* num complemento             */
#define MAX_TIPO_LOG 10 /* Nmero mximo de caracteres */
                        /* num tipo de logradouro      */
#define TAM_DNE   16 /* Nmero de dgitos numa chave DNE */

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

#define SEPARADORES "\t;\n" /* Possveis separadores de campos */

   /* Nome dos arquivos binrio */
#define NOME_ARQUIVO_BIN "CEPs.bin"

   /* Altere o string a seguir para incluir o diretrio */
   /* em que se encontra o arquivo "CEPs.txt"           */
#define NOME_ARQUIVO_TXT "I:\\Dados\\CEPs\\CEPs.txt"

   /* Se o arquivo "CEPs.txt" encontra-se no diretrio */
   /* corrente, comente a linha anterior e remova o    */
   /* comentrio da linha a seguir                     */
//#define NOME_ARQUIVO_TXT "CEPs.txt"

   /* Mltiplos de bytes */
#define KIBIBYTE 1024             /* 1 Kibibyte */
#define MEBIBYTE (1024*1024)      /* 1 Mebibyte */
#define GIBIBYTE (1024*1024*1024) /* 1 Gibibyte */

/****************** Definies de Macros ******************/

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(cond, msg) if (!(cond)) {\
                               fprintf(stderr, "\n%s\n", msg);\
                               exit(1); \
                            }

/****************** Definies de Tipos *******************/

   /* Tipo de uma estrutura que representa um registro de CEP */
typedef struct {
           int  numero;
           char UF[TAM_UF];
           int  localidadeNumero;
           char nomeAbr[MAX_NOME];
           char nome[MAX_NOME];
           int  bairroInicio;
           int  bairroFim;
           char CEP[TAM_CEP];
           char complemento[MAX_COMP];
           char tipoLogradouro[MAX_TIPO_LOG];
           char statusLogradouro;
           char nomeSemAcento[MAX_NOME];
           char chaveDNE[TAM_DNE];
         } tRegistroCEP;

/***************** Definies de Funes ******************/

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
static char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor  */
         /* que i poderia assumir deveria ser tamanho - 1.  */
         /* Mas, como ao final, o caractere '\0' dever ser */
         /* inserido, limita-se o valor de i a tamanho - 2. */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n')
         break; /* Leitura encerrada */

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontado pelo parmetro */
      /* 'tam', se ele no for NULL               */
   if (tam)
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****
*
* ObtemTokens(): Separa um string em partes (tokens)
*
* Parmetros:
*     str (entrada/sada) - o string que ser separado
*                           ou NULL para continuar uma
*                           separao anterior
*     sep (entrada) - string contendo os separadores de tokens
*
* Retorno: Endereo do prximo token do string, se este
*          existir, ou NULL, em caso contrrio
*
* Observao: Esta funo NO  funcionalmente equivalente 
*             funo strtok() da biblioteca padro de C
*
****/
static char *ObtemTokens(char *str, char const *sep)
{
   static char *proximoToken; /* Aponta para o prximo */
                              /* token, se ele existir */
   char        *s, /* Apontar para o string no qual a */
                   /* busca pelo token ser efetuada   */
               *inicio = NULL; /* Guardar o incio */
                               /* do token corrente */

      /* Se 'str' no for um ponteiro nulo, o prximo  */
      /* token ser obtido a partir do incio de 'str' */
   if (str)
      proximoToken = str;

      /* Se 'proximoToken' for um ponteiro nulo, o */
      /* string ora explorado no tem mais tokens  */
   if (!proximoToken)
      return NULL; /* No h mais token nesse string */

      /* Obtm o endereo do primeiro separador */
      /* encontrado em 'proximoToken'           */
   s = strpbrk(proximoToken, sep);

      /* Verifica se foi encontrado algum separador */
      /* no string ora sendo explorado              */
   if (s) {
         /* Termina o token corrente na posio em */
         /* que se encontra o separador encontrado */
      *s = '\0';

         /* Guarda o incio do token corrente */
      inicio = proximoToken;

         /* O prximo token comear no primeiro */
         /* caractere aps o separador           */
      proximoToken = ++s;
   } else
      if (*proximoToken) {
            /* No foi encontrado nenhum separador, mas o string  */
            /* corrente no  vazio e seu endereo ser retornado */
         inicio = proximoToken;

         /* No foi encontrado nenhum separador, de modo que */
         /* no haver mais nenhum token na prxima chamada  */
      proximoToken = NULL;
   }

   return inicio;
}

/****
 *
 * CopiaNStr(): Copia um determinado nmero de caracteres de um
 *              string para um array de caracteres
 *
 * Parmetros:
 *     dest (sada) - o array que receber os caracteres
 *     origem (entrada) - string doador de caracteres
 *     n  (entrada) - nmero mximo de caracteres que sero copiados
 *
 * Retorno: O primeiro parmetro
 *
 * Observao: Esta funo  similar  strncpy(). A diferena 
 *             que, quando o segundo parmetro  NULL, um string
 *             vazio  copiado para o primeiro parmetro e o
 *             programa no  abortado.
 *
 ****/
char *CopiaNStr(char *dest, const char *origem, size_t n)
{
   return origem && *origem ? strncpy(dest, origem, n)
                            : strcpy(dest, "");
}

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa estrutura
 *                    do tipo tRegistroCEP
 *
 * Parmetros: regCEP (sada) - ponteiro para a estrutura
 *                             resultante da converso
 *             linha (entrada/sada) - linha que ser convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado
 *          da converso
 *
 * Observaes:
 *     1. Para facilitar o entendimento, esta funo assume
 *        que o parmetro 'linha' realmente  um string no
 *        formato das linhas do arquivo. Portanto, os
 *        valores retornados por ObtemTokens() no so testados
 *        como deveriam.
 *
 *     2. O parmetro 'linha'  alterado por ObtemTokens().
 *
 *     3. A funo ObtemTokens()  usada em lugar de strtok()
 *        devido  presena de campos vazios no arquivo de
 *        texto, o que faz com que dois separadores apaream
 *        juntos. Nesse caso, a funo strtok() consideraria
 *        os dois separadores como sendo apenas um.
 *
 ****/
static tRegistroCEP *LinhaEmRegistro(tRegistroCEP *regCEP, char *linha)
{
   char *str; /* Apontar para cada token da linha */
   char *final; /* Final de uma converso numrica */

      /* Obtm o nmero e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(linha, SEPARADORES);
   ASSEGURA(str, "Token nulo: numero");
   regCEP->numero = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm a UF e acrescenta-o ao  */
      /* respectivo campo da estrutura */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->UF, str, TAM_UF);

      /* Obtm o nmero de localidade e acrescenta-o */
      /* ao respectivo campo da estrutura            */
   str = ObtemTokens(NULL, SEPARADORES);
   ASSEGURA(str, "Token nulo: localidadeNumero");
   regCEP->localidadeNumero = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->localidadeNumero < 0,
            "Erro de conversao numerica");

      /* Obtm o nome abreviado e acrescenta-o */
      /* ao respectivo campo da estrutura      */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->nomeAbr, str, MAX_NOME);

      /* Obtm o nome e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->nome, str, MAX_NOME);

      /* Obtm o incio do bairro e acrescenta-o */
      /* ao respectivo campo da estrutura        */
   str = ObtemTokens(NULL, SEPARADORES);
   ASSEGURA(str, "Token nulo: bairroInicio");
   regCEP->bairroInicio = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->bairroInicio < 0,
            "Erro de conversao numerica");

      /* Obtm o final do bairro e acrescenta-o */
      /* ao respectivo campo da estrutura       */
   str = ObtemTokens(NULL, SEPARADORES);
   ASSEGURA(str, "Token nulo: bairroFim");
   regCEP->bairroFim = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->bairroFim < 0,
            "Erro de conversao numerica");

      /* Obtm o CEP e acrescenta-o ao */
      /* respectivo campo da estrutura */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->CEP, str, TAM_CEP);

      /* Obtm o complemento e acrescenta-o */
      /* ao respectivo campo da estrutura   */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->complemento, str, MAX_COMP);

      /* Obtm o tipo de logradouro e acrescenta-o */
      /* ao respectivo campo da estrutura          */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->tipoLogradouro, str, MAX_TIPO_LOG);

      /* Obtm o status de logradouro e acrescenta-o */
      /* ao respectivo campo da estrutura            */
   str = ObtemTokens(NULL, SEPARADORES);
   regCEP->statusLogradouro = *str;

      /* Obtm o nome sem acento e acrescenta-o */
      /* ao respectivo campo da estrutura       */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->nomeSemAcento, str, MAX_NOME);

      /* Obtm a chave DNE e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, SEPARADORES);
   CopiaNStr(regCEP->chaveDNE, str, TAM_DNE);

   return regCEP;
}

/****
 *
 * CriaArquivoBin(): L cada linha de um arquivo no formato
 *                   especificado, converte-a numa estrutura do
 *                   tipo tRegistroCEP e armazena a estrutura num arquivo
 *                   binrio
 *
 * Parmetros:
 *     streamTexto (entrada) - stream associado ao arquivo que ser
 *                             lido
 *     streamBin (entrada) - stream associado ao arquivo que ser
 *                           escrito
 *     linhasVazias (sada) - linhas do arquivo de entrada que no
 *                            so levadas em considerao
 *
 * Retorno: Nmero de estruturas escritas no arquivo binrio ou um
 *          valor negativo, se ocorrer erro
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de texto
 *        que permite leitura
 *     2. O stream de sada deve estar aberto em modo "wb"
 *
 ****/
int CriaArquivoBin( FILE *streamTexto, FILE *streamBin,
                    int *linhasVazias )
{
   char  *linha; /* Apontar para cada linha lida */
   tRegistroCEP   umCEP; /* Dados de uma linha      */
                 /* convertida em estrutura */
   int    nRegistros = 0; /* Nmero de estruturas escritas */
                          /* no arquivo binrio            */

   *linhasVazias = 0; /* Ainda no foi lida nenhuma linha vazia */

      /* Garante leitura a partir do incio do arquivo */
   rewind(streamTexto);

      /* L cada linha do arquivo, cria um registro do tipo tRegistroCEP */
      /* e armazena-o no arquivo binrio. O lao encerra quando  */
      /* h tentativa de leitura alm do final do arquivo de     */
      /* entrada, ou ocorre erro de leitura/escrita.             */
   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(NULL, streamTexto);

         /* Verifica se o final do arquivo foi atingido */
      if (feof(streamTexto))
         break;

         /* Verifica se ocorreu erro de leitura */
      if (ferror(streamTexto))
         return -1; /* Operao mal sucedida */

         /* Verifica se a linha lida  vazia */
      if (!linha || !*linha || *linha == '\n') {
         ++*linhasVazias; /* Foi lida uma linha vazia */

         free(linha); /* A linha no contm dados relevantes */

         continue; /* Salta o restante do corpo do lao */
      }

         /* Converte a linha lida em estrutura */
      LinhaEmRegistro(&umCEP, linha);

         /* Escreve a estrutura resultante da */
         /* converso no arquivo binrio      */
      fwrite(&umCEP, sizeof(umCEP), 1, streamBin);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamBin))
         return -1; /* Operao mal sucedida */

      ++nRegistros; /* Mais um registro lido */

      free(linha); /* Libera o espao ocupado pela linha corrente */
   }

   return nRegistros; /* No ocorreu nenhum erro */
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas chamadas
 *                  consecutivas
 *
 * Parmetros:
 *     info (entrada) - informao que ser apresentada na tela
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e., na
 *             primeira chamada ela marca o incio da contagem de
 *             tempo e na segunda chamada ela apresenta o intervalo
 *             de tempo decorrido desde a ltima chamada
 *
 ****/
void MedidaDeTempo(const char *info)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

   if (info) {
      printf("\n>>> %s\n", info);
      fflush(stdout);
   }

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/****
 *
 * EscreveNumeroBytes(): Escreve um determinado nmero de bytes
 *                       (valor) num stream de texto
 *
 * Parmetros:
 *     nBytes (entrada) - o valor a ser escrito
 *     stream (entrada) - o stream de texto
 *
 * Retorno: Nenhum
 *
 ****/
void EscreveNumeroBytes(int nBytes, FILE *stream)
{
   fprintf(stream, " %d bytes", nBytes);

   if (nBytes > GIBIBYTE)
      fprintf(stream, " (%.2f GiB)", (double) nBytes/GIBIBYTE);
   else if (nBytes > MEBIBYTE)
      fprintf(stream, " (%.2f MiB)", (double) nBytes/MEBIBYTE);
   else if (nBytes > KIBIBYTE)
      fprintf(stream, " (%.2f KiB)", (double) nBytes/KIBIBYTE);

   fprintf(stream, "\n");
}

/****
 *
 * main(): Executa a funo CriaArquivoBin()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int   nRegistros, /* Nmero de registros nos arquivos  */
         tamArquivo, /* Tamanho do arquivo binrio        */
         vazias;     /* Linhas que no sero consideradas */
   FILE *streamTexto, /* Associado ao arquivo de texto */
        *streamBin;   /* Associado ao arquivo binrio */

   printf("\n>>> Tamanho de um registro (tRegistroCEP): %d\n", sizeof(tRegistroCEP));

      /* Tenta abrir arquivo para leitura em modo texto */
   streamTexto = fopen(NOME_ARQUIVO_TXT, "r");

      /* Se o arquivo no pode ser aberto, */
      /* nada mais pode ser feito          */
   if (!streamTexto) {
      printf( "O arquivo %s nao pode ser aberto\n",
              NOME_ARQUIVO_TXT );
      return 1; /* Operao mal sucedida */
   }

      /* Tenta criar arquivo binrio. Se um arquivo com esse   */
      /* nome existir no diretrio corrente ele ser destrudo */
   streamBin = fopen(NOME_ARQUIVO_BIN, "wb");

      /* Se o arquivo no pode ser aberto, */
      /* nada mais pode ser feito.         */
   if (!streamBin) {
      printf("Arquivo %s nao pode ser aberto\n",NOME_ARQUIVO_BIN);

         /* Antes de retornar, fecha arquivo aberto */
      fclose(streamTexto);

      return 1; /* Operao mal sucedida */
   }

   MedidaDeTempo("Criando arquivo...");

      /* L cada linha do arquivo de texto,   */
      /* converte-a em estrutura do tipo tRegistroCEP */
      /* e armazena-a no arquivo binrio.     */
   nRegistros = CriaArquivoBin(streamTexto, streamBin, &vazias);

      /* Verifica se ocorreu erro na criao do arquivo */
      /* binrio. Se for o caso, informa o usurio,     */
      /* fecha os arquivos e encerra o programa.        */
   if (nRegistros < 0) {
      printf( "\n\t>>> Ocorreu erro na criacao do arquivo"
              "\n\t>>> \"%s\"\n", NOME_ARQUIVO_BIN );

         /* Fecha os arquivos antes de encerrar */
      fclose(streamTexto);
      fclose(streamBin);

      return 1; /* Ocorreu erro */
   }

   MedidaDeTempo("...Arquivo criado com sucesso");

   tamArquivo = nRegistros*sizeof(tRegistroCEP);

      /* Apresenta o resultado da converso */
   printf( "\n>>> Foram armazenados %d registros no arquivo "
           "\"%s\"\n", nRegistros, NOME_ARQUIVO_BIN );

   printf( "\n>>> Tamanho do arquivo \"%s\":", NOME_ARQUIVO_BIN );
   EscreveNumeroBytes(tamArquivo, stdout);

   printf("\n>>> Linhas vazias: %d\n", vazias);

      /* Fecha os arquivos */
   fclose(streamTexto);
   fclose(streamBin);

   return 0; /* Tudo ocorreu bem */
}

/****************** Resultado do Programa *****************

>>> Tamanho de um registro (tRegistroCEP): 236

>>> Criando arquivo...

>>> ...Operacao encerrada. Arquivo criado com sucesso

        *** Tempo gasto na operacao:  5.00 segundos ***

>>> Foram armazenados 673580 registros no arquivo "CEPs.bin"

>>> Tamanho do arquivo "CEPs.bin": 158964880 bytes (151.60 MiB)

>>> Linhas vazias: 0

**********************************************************/

