/****
 *
 * Ttulo: Lista Esttica Ordenada
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2014
 * ltima modificao: 16/04/2014
 *
 * Descrio: Determina se um nmero inteiro positivo  primo
 *
 * Entrada: Valores inteiros positivos ou zero para encerrar o programa
 *
 * Sada: Uma mensagem indicando se o nmero  primo ou no
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Entrada e sada         */
#include <string.h>   /* Funes strXXX()        */
#include <stdlib.h>   /* malloc()                */
#include "ASSEGURA.h" /* Macro ASSEGURA()        */
#include "Lista.h"    /* Interface do tipo lista */

/***************** Definies de Funes ******************/

/****
 *
 * IniciaLista(): Inicia uma lista
 *
 * Parmetros:
 *      lista (sada) - ponteiro para a estrutura que armazena
 *                      a lista a ser iniciada
 *
 * Retorno: Endereo da lista iniciada
 *
 ****/
tListaPtr IniciaLista(tListaPtr lista)
{
      /* Inicialmente a lista est vazia */
   lista->nElementos = 0;

      /* Inicialmente o tamanho do array  zero */
   lista->tamanhoArray = 0;

      /* Aloca previamente um array */
   lista->elementos = calloc(ACRESCIMO, sizeof(tElemento));

      /* Garante que o array foi realmente alocado */
   ASSEGURA( lista->elementos, "Nao foi possivel alocar "
             "array para conter elementos da lista" );

      /* Atualiza a varivel que armazena o tamanho do array */
   lista->tamanhoArray += ACRESCIMO;

   return lista;
}

/****
 *
 * DestroiLista(): Libera o espao ocupado pela lista
 *
 * Parmetros: lista (entrada) - a lista
 *
 * Retorno: Nada
 *
 ****/
void DestroiLista(tListaPtr lista)
{
      /* Libera o array que contm os elementos */
   free(lista->elementos);

   lista->tamanhoArray = 0;
}

/****
 *
 * Comprimento(): Calcula o comprimento de uma lista
 *
 * Parmetros:
 *     *lista (entrada) - a lista cujo comprimento
 *                        ser calculado
 *
 * Retorno: o nmero de elementos na lista
 *
 ****/
int Comprimento(tListaPtr lista)
{
   return lista->nElementos; /* No h o que calcular */
}

/****
 *
 * ObtemElemento(): Obtm o elemento de uma lista que se
 *                  encontra numa dada posio
 *
 * Parmetros:
 *     *lista (entrada) - a lista
 *      indice (entrada) - o elemento
 *
 * Retorno: o nmero de elementos na lista
 *
 ****/
tElemento ObtemElemento(tListaPtr lista, int indice)
{
   ASSEGURA( indice >= 0 && indice < lista->nElementos,
             "Elemento inexistente" );

   return lista->elementos[indice];
}

/****
 *
 * EstaVazia(): Verifica se uma lista sem ordenao est vazia
 *
 * Parmetros: *lista (entrada) - a lista que ser testada
 *
 * Retorno: 1, se a lista estiver vazia; 0, em caso contrario
 *
 ****/
int EstaVazia(tListaPtr lista)
{
   return lista->nElementos == 0;
}

/****
 *
 * AcrescentaElemento(): Acrescenta um novo elemento ao final
 *                       de uma lista sem ordenao
 *
 * Parmetros:
 *      lista (sada) - lista na qual ser feita o acrscimo
 *      elemento (entrada) - valor do elemento que ser
 *                           acrescentado
 *
 * Retorno: Nada
 *
 ****/
void AcrescentaElemento(tListaPtr lista, const tElemento *elemento)
{
   tElemento *novoArray; /* Ponteiro para o array */
                         /* redimensionado        */

      /* Se o array estiver repleto, tenta redimension-lo */
   if (lista->nElementos >= lista->tamanhoArray) {
         /* Calcula o novo tamanho do array */
      lista->tamanhoArray += ACRESCIMO;

         /* Tenta redimensionar o array */
      novoArray = realloc(lista->elementos,
                          lista->tamanhoArray*sizeof(tElemento));

         /* Checa se houve realocao de memria */
      ASSEGURA(novoArray, "Nao houve redimensionamento");

         /* O ponteiro que apontava para o incio */
         /* do array pode no ser mais vlido.    */
         /* Portanto,  preciso atualiz-lo.      */
      lista->elementos = novoArray;
   }

      /* Acrescenta o novo elemento ao final da lista */
   lista->elementos[lista->nElementos] = *elemento;

   ++lista->nElementos; /* O tamanho da lista aumentou */
}
