/****
 *
 * Arquivo DispersaoEstatica.h
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 26/08/2017
 *
 * Descrio: Iterface das funes que implementam operaes
 *            de busca, insero e remoo usando disperso
 *            esttica em memria secundria
 *
 ****/

#ifndef _DispersaoEstatica_H_ /* Previne incluso mltipla */
#define _DispersaoEstatica_H_

#include "Registros.h" /* Tipos tRegistroMEC e tChave */

/************ Definies de Tipos e Constantes ************/

   /* Posio nula (invlida) em arquivo */
#define POSICAO_NULA -1

   /* Tipo usado para classificar um coletor */
typedef enum {PRIMARIO, EXCEDENTE} tTipoDeColetorDEst;

#define TB 4096 /* Tamanho do bloco lido ou escrito  */
#define TT (int) sizeof(tTipoDeColetorDEst) /* Tamanho de tTipoDeColetorDEst */
#define TR (int) sizeof(tRegistroMEC) /* Tamanho de um registro */
#define TI (int) sizeof(int) /* Tamanho de um campo do tipo int */

   /* Clculo do nmero mximo de registros num coletor */
#define M (((TB - 2*TI - TT)/TR) - 1)

   /* Tipo de uma varivel que armazena um coletor */
typedef struct {
           tTipoDeColetorDEst tipo; /* Tipo de coletor */
           tRegistroMEC       registros[M]; /* Array de registros  */
           int                nRegistros; /* Nmero de registros num coletor */
           int                proximo; /* Prximo coletor excedente */
        } tColetorDEst;

/************************* Aluses ************************/

extern int IniciaTabDEst(FILE *streamCol, int nRegistros, int maxRegs);
extern tRegistroMEC *BuscaDEst( FILE *stream, int nColetores,
                                tChave chave, tRegistroMEC *reg );
extern void InsereDEst(int nColetores, const tRegistroMEC *registro, FILE *stream);
extern int RemoveDEst(int nColetores, tChave chave, FILE *stream);
extern void SubstituiDEst(FILE *stream, int nColetores, const tRegistroMEC *reg);
extern int NRegistrosTabDEst( FILE *stream );
extern int NColetoresDEst(FILE *stream);
extern int NColetoresVaziosDEst(FILE *stream);
extern int NColetoresRepletosDEst(FILE *stream);
extern int NColetoresOcupadosDEst(FILE *stream);
extern int NColetoresExcedentesDEst(int nColetores, FILE *stream);
extern int MaiorCadeiaExcedentesDEst(int nColetores, FILE *stream);

#endif
