/****
 *
 * Arquivo DispersaoExt.h
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 26/08/2017
 *
 * Descrio: Iterface das funes que implementam operaes
 *            de busca, insero e remoo usando disperso
 *            extensvel em memria secundria
 *
 ****/

#ifndef _DispersaoExt_H_ /* Previne incluso mltipla */
#define _DispersaoExt_H_

#include "Registros.h" /* Tipos tRegistro e tChave */

/************ Definies de Tipos e Constantes ************/

   /* Posio nula (invlida) em arquivo */
#define POSICAO_NULA -1

#define TB 4096 /* Tamanho do bloco lido ou escrito  */
#define TR (int) sizeof(tRegistro) /* Tamanho de um registro */
#define TI (int) sizeof(int) /* Tamanho de um campo do tipo int */

   /* Clculo do nmero mximo de registros num coletor */
//#define M    ((TB - (2*TI))/TR)
#define M 3

   /* Tipo de uma varivel que armazena um coletor */
typedef struct {
           tRegistro registros[M]; /* Array de registros  */
           int       nRegistros;   /* Nmero de registros */
                                   /* num coletor         */
           int       profLocal;    /* Profundidade local  */
        } tColetor;

   /* Tipo de uma varivel que representa uma  */
   /* tabela de busca com disperso extensvel */
typedef struct {
           int *dir;        /* Diretrio            */
           int  profGlobal; /* Profundidade global  */
           int  tamDir;     /* Tamanho do diretrio */
           int  nReg;       /* Nmero de registros  */
        } tTabelaDExt;

/************************* Aluses ************************/

extern int Dispersao(const char *s);
extern void LeColetor(FILE *stream, int pos, tColetor *coletor);
extern void EscreveColetor( FILE *stream, int pos,
                            const tColetor *coletor );
extern void IniciaTabelaExt(tTabelaDExt *tabela, FILE *stream);
extern tRegistro *BuscaExt( FILE *stream, const tTabelaDExt *tab,
                            tChave chave, tRegistro *reg );
extern void InsereExt( tTabelaDExt *tab, const tRegistro *registro,
                       FILE *stream );
extern int RemoveExt(tTabelaDExt *tab, tChave chave, FILE *stream);
extern void SubstituiExt( FILE *stream, const tTabelaDExt *tab,
                          const tRegistro *reg );
extern int NRegistrosTabExt( const tTabelaDExt *tabela,
                             FILE *stream );
extern int NColetores(FILE *stream);
extern int NColetoresVazios(FILE *stream);
extern int NColetoresRepletos(FILE *stream);
extern int NColetoresOcupados(FILE *stream);
extern int ProfLocalMin(FILE *stream);
extern int ProfLocalMax(FILE *stream);

#endif
