/****
 *
 * Arquivo DispersaoExt.h
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 26/08/2017
 *
 * Descrio: Iterface das funes que implementam operaes
 *            de busca, insero e remoo usando disperso
 *            extensvel em memria secundria
 *
 ****/

#ifndef _DispersaoExt_H_ /* Previne incluso mltipla */
#define _DispersaoExt_H_

#include "Registros.h" /* Tipos tRegistroMEC e tChave */

/************ Definies de Tipos e Constantes ************/

   /* Posio nula (invlida) em arquivo */
#define POSICAO_NULA -1

#define TB 4096 /* Tamanho do bloco lido ou escrito  */
#define TR (int) sizeof(tRegistroMEC) /* Tamanho de um registro   */
                                      /* armazenado nos coletores */
#define TI (int) sizeof(int) /* Tamanho de um campo do tipo int */

   /* Clculo do nmero mximo de registros num coletor */
#define M (((TB - (2*TI))/TR) - 1)

   /* Tipo de uma varivel que armazena um coletor */
typedef struct {
           tRegistroMEC registros[M]; /* Array de registros  */
           int          nRegistros;   /* Nmero de registros num coletor */
           int          profLocal;    /* Profundidade local  */
        } tColetorDExt;

   /* Tipo de uma varivel que representa uma  */
   /* tabela de busca com disperso extensvel */
typedef struct {
           long *dir;        /* Diretrio            */
           int   profGlobal; /* Profundidade global  */
           int   tamDir;     /* Tamanho do diretrio */
           int   nReg;       /* Nmero de registros  */
        } tTabelaDExt;

/************************* Aluses ************************/

extern void LeColetorDExt(FILE *stream, int pos, tColetorDExt *coletor);
extern void EscreveColetorDExt( FILE *stream, int pos,
                                const tColetorDExt *coletor );
extern void IniciaTabelaDExt(tTabelaDExt *tabela, FILE *stream);
extern tRegistroMEC *BuscaDExt( FILE *stream, const tTabelaDExt *tab,
                                tChave chave, tRegistroMEC *reg );
extern void InsereDExt(tTabelaDExt *tab, const tRegistroMEC *item, FILE *stream);
extern int RemoveDExt(tTabelaDExt *tab, tChave chave, FILE *stream);
extern int NRegistrosDExt(const tTabelaDExt *tabela, FILE *stream);
extern int NColetoresDExt(FILE *stream);
extern int NColetoresVaziosDExt(FILE *stream);
extern int NColetoresRepletosDExt(FILE *stream);
extern int NColetoresOcupadosDExt(FILE *stream);
extern int ProfLocalMinDExt(FILE *stream);
extern int ProfLocalMaxDExt(FILE *stream);

#endif
