/****
 *
 * Exemplo: Casamento de Strings por Fora Bruta
 *
 * Seo: 9.2.3
 *
 ****/

#include <stdio.h>
#include <string.h>

/****
 *
 * CasamentoFB(): Verifica se um padro ocorre num texto
 *               usando fora bruta (verso jnior)
 *
 * Parmetros:
 *      t (entrada) - string que representa o texto
 *      p (entrada) - string que representa o padro
 *
 * Retorno: A posio do padro no texto, se ele for
 *          encontrado; -1, em caso contrrio.
 *
 ****/
int CasamentoFB(const char *t, const char *p)
{
   int n = strlen(t), /* Comprimento do texto  */
       m = strlen(p), /* Comprimento do padro */
       i,  /* ndice de um caractere do texto  */
       j;  /* ndice de um caractere do padro */

      /* Enquanto os ndices i e j referem-se a caracteres    */
      /* coincidentes eles so incrementados. Se eles se      */
      /* referirem a caracteres que no casam, enquanto isso  */
      /* ocorrer, j  reiniciado com o ndice inicial do      */
      /* padro e i  associado ao prximo caractere do texto */
      /* a ser comparado.                                     */
   for (i = 0, j = 0; i < n && j < m; ++i, ++j) {
         /* Enquanto os caracteres ora sendo comparados */
         /* no casarem, associa j ao ndice inicial do */
         /* padro e i ao ndice do prximo caractere a */
         /* ser comparado no texto                      */
      while (t[i] != p[j]) {
         i = i - j + 1;
         j = 0;
      }
   }

      /* Se o ndice j for igual ao tamanho do padro, todos */
      /* os caracteres do padro casaram com caracteres do   */
      /* texto. Caso contrrio, o lao for encerrou porque   */
      /* todos os caracteres do texto foram examinados       */
      /* (i.e., i == n) sem que o padro fosse encontrado.   */
   if (j == m) {
         /* O padro foi encontrado. Retorna-se */
         /* sua posio no texto.               */
      return i - m;
   } else {
      return -1; /* Padro no foi encontrado no texto */
   }
}

/****
 *
 * CasamentoFB2(): Verifica se um padro ocorre num texto
 *                usando fora bruta (verso snior)
 *
 * Parmetros:
 *      t (entrada) - string que representa o texto
 *      p (entrada) - string que representa o padro
 *
 * Retorno: A posio do padro no texto, se ele for
 *          encontrado; -1, em caso contrrio.
 *
 * Observao: Esta funo  ligeiramente mais eficiente
 *             do que a funo CasamentoFB(), pois ela no
 *             calcula os tamanhos dos strings. Mas, do
 *             ponto de vista de anlise assinttica, as
 *             duas funes apresentam o mesmo custo.
 *
 ****/
int CasamentoFB2(const char *t, const char *p)
{
   const char *st, /* Aponta para um caractere no texto  */
              *sp, /* Aponta para um caractere no padro */
              *inicio = t; /* Guarda o incio do texto */

      /* Enquanto o texto no for exaurido ou o padro no for */
      /* encontrado no texto, continua tentando encontr-lo    */
   for (; *t; t++) {
         /* Se os caracteres correspondentes no texto e no     */
         /* padro casarem continua incrementando os ponteiros */
      if (*t == *p) {
         for (st = t, sp = p; *st && *st == *sp; ++st, ++sp)
            ; /* Instruo vazia */

            /* Se o ponteiro sp estiver apontando para  */
            /* o caractere terminal do padro, o padro */
            /* foi encontrado no texto                  */
         if (!*sp)
               /* Retorna a posio do padro no texto */
            return t - inicio;
      }
   }

      /* Se o padro tivesse sido encontrado, */
      /* j teria ocorrido retorno            */
   return -1;
}

/****
 *
 * main(): Testa as funes CasamentoFB() e CasamentoFB2()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   char *texto = "abacaabaccabacabaabb";
   char *padrao = "abacab";
   int   posicao;

   printf("\n\t>>>> Usando CasamentoFB Junior <<<\n");

   posicao = CasamentoFB(texto, padrao);

   if (posicao >= 0) {
      printf( "\n\"%s\" encontrado em \"%s\" na posicao %d\n",
              padrao, texto, posicao );
   } else {
      printf( "\n\"%s\" nao foi encontrado em \"%s\"\n",
              padrao, texto );
   }

   printf("\n\t>>>> Usando CasamentoFB Senior <<<\n");

   posicao = CasamentoFB2(texto, padrao);

   if (posicao >= 0) {
      printf( "\n\"%s\" encontrado em \"%s\" na posicao %d\n",
              padrao, texto, posicao );
   } else {
      printf( "\n\"%s\" nao foi encontrado em \"%s\"\n",
              padrao, texto );
   }

   return 0;
}

/****************** Resultado do Programa *****************

        >>>> Usando CasamentoFB Junior <<<

"abacab" encontrado em "abacaabaccabacabaabb" na posicao 10

        >>>> Usando CasamentoFB Senior <<<

"abacab" encontrado em "abacaabaccabacabaabb" na posicao 10

**********************************************************/
