/****
 *
 * Ttulo: Quantas Vezes Machado Fala em Amor?
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2016
 * ltima modificao: 16/04/2016
 *
 * Descrio: Utiliza uma trie para indexar as palavras da obra
 *            de Machado de Assis, de tal modo que o programa
 *            permita consultas
 *
 * Entrada: Itens (strings) e opes escolhidas pelo usurio
 *
 * Sada: A lista criada pelo usurio
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* Alocao dinmica de memria */
#include <ctype.h>    /* Classificao de caracteres  */
#include "Lista.h"    /* Interface do tipo lista */
#include "Trie.h"     /* Interface do tipo trie   */
#include "leitura.h"  /* Funes de leitura      */
#include "ASSEGURA.h" /* Macro ASSEGURA         */

/***************** Constantes Simblicas ******************/

#define NOME_ARQUIVO "Machado.txt"
//#define NOME_ARQUIVO "Machadinho.txt"

#define SEPARADORES " '[](){}:,!.-?\";/\\"

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

/***************** Definies de Funes ******************/

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor  */
         /* que i poderia assumir deveria ser tamanho - 1.  */
         /* Mas, como ao final, o caractere '\0' dever ser */
         /* inserido, limita-se o valor de i a tamanho - 2. */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontado pelo parmetro */
      /* 'tam', se ele no for NULL               */
   if (tam) {
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;
   }

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****
 *
 * ExibeLista(): Exibe na tela do primeiro ao ltimo elemento
 *               de uma lista simplesmente encadeada
 *
 * Parmetros:
 *     *lista (entrada) - a lista
 *
 * Retorno: Nada
 *
 ****/
void ExibeLista(tListaSE lista)
{
   tConteudo *pConteudo = NULL;

   while ((pConteudo = ProximoListaSE(lista))) {
      printf("Linha: %d\tPosicao: %d\n",
             pConteudo->linha, pConteudo->pos);
   }
}

/****
 *
 * ValidaPalavra(): Remove de um string smbolos diacrticos e
 *                  converte maisculas em minsculas. Se aps
 *                  essa operao no sobrarem apenas letras de
 *                  'a' a 'z', o token  rejeitado como palavra
 *
 * Parmetros:
 *     *pPalavra (entrada/sada) - o referido string
 *
 * Retorno: 1, se o string for considerado uam palavra volida;
 *          0, em caso contrrio.
 *
 ****/
int ValidaPalavra(char *pPalavra)
{
   char *p = pPalavra;

      /* Remove smbolos diacrticos */
   while (*p) {
      switch (*p) {
         case '':
         case '':
         case '':
         case '':
         case '':
         case '':
         case '':
         case '':
            *p = 'a';
            break;

         case '':
         case '':
         case '':
         case '':
            *p = 'e';
            break;

         case '':
         case '':
            *p = 'i';
            break;

         case '':
         case '':
         case '':
         case '':
         case '':
         case '':
            *p = 'o';
            break;

         case '':
         case '':
            *p = 'u';
            break;

         case '':
         case '':
            *p = 'c';
            break;
      }

      ++p;
   }

   p = pPalavra;

      /* Garante que todos os caracteres so */
      /* letras minsculas de 'a' a 'z'      */
   while (*p) {
      *p = tolower(*p);

         /* Se o caractere corrente no for letra, */
         /* o string no  considerado palavra     */
      if (!islower(*p)) {
         return 0;
      }

      ++p;
   }

   return 1; /* O string passou no teste */
}

/****
 *
 * InsereLinhaEmTrieMachado(): Insere as palavras de um string numa trie
 *
 * Parmetros:
 *    raiz (entrada) - raiz da trie
 *    linha (entrada/sada) - string que contm as palavras
 *    nLinha (entrada) - ndice associado ao string que contm
 *                       as palavras
 *
 * Retorno: Nada
 *
 ****/
void InsereLinhaEmTrieMachado(tTrieMachado raiz, char *linha, int nLinha)
{
   tConteudoMachado conteudo;
   int              posPalavra = 0;
   char            *pPalavra;

      /* Obtm o primeiro token */
   pPalavra = strtok(linha, SEPARADORES);

      /* Verifica se foi obtido algum token no vazio */
   if (!pPalavra || !*pPalavra)
      return; /* No h palavras */

      /* O valor do campo 'linha' ser o mesmo */
      /* para todas as palavras desta linha    */
   conteudo.linha = nLinha;

      /* A funo ValidaPalavra() remove smbolos diacrticos e converte */
      /* maisculas em minsculas para simplificar o problema. Se, aps  */
      /* tudo isso, no sobrarem apenas letras de 'a' a 'z', o token    */
      /* rejeitado como palavra.                                         */
   if (ValidaPalavra(pPalavra)) {
         /* Se desejar indexao de palavras a partir de */
         /* zero, troque incremento prefixo por sufixo   */
      conteudo.pos = ++posPalavra;

      InsereTrieMachado(raiz, pPalavra, &conteudo);
   }

      /* Obtm as demais palavras da linha e insere-as na trie */
   while (1) {
         /* Obtm o prximo token */
       pPalavra = strtok(NULL, SEPARADORES);

         /* Se no houver mais tokens, encerra o lao */
       if (!pPalavra)
          break;

         /* Insere apenas palavras vlidas na trie */
       if (pPalavra && ValidaPalavra(pPalavra)) {
          conteudo.pos = ++posPalavra;

          InsereTrieMachado(raiz, pPalavra, &conteudo);
       }
   }
}

/****
 *
 * ConstroiTrieMachado(): Cria uma trie com o contedo de um arquivo de texto
 * Parmetros:
 *    nomeArq (entrada) - nome do arquivo de texto
 *
 * Retorno: O endereo da trie criada
 *
 ****/
tTrieMachado ConstroiTrieMachado(const char *nomeArq)
{
   FILE        *stream; /* Stream associado ao arquivo */
                        /* no qual ocorrer a leitura  */
   char        *linha; /* Apontar para cada linha lida */
   int          nLinha = 0;
   tTrieMachado trie = IniciaTrieMachado();

      /* Abre arquivo em formato de texto para leitura */
   stream = fopen(nomeArq, "r");

      /* Se o arquivo no foi aberto, nada mais pode ser feito */
   ASSEGURA(stream, "Arquivo no foi aberto");

   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(NULL, stream);

         /* Verifica se o final do arquivo foi atingido */
      if (feof(stream))
         break;

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA(!ferror(stream), "Erro de leitura");

      ++nLinha; /* Mais uma linha lida */

         /* Insere as palavras da linha lida na trie */
      InsereLinhaEmTrieMachado(trie, linha, nLinha);

      free(linha);
   }

   fclose(stream); /* Processamento terminado. Fecha o arquivo. */

   free(linha);

   return trie;
}

/****
 *
 * ApresentaMenu(): Apresenta o menu do programa
 *
 * Parmetros: No h
 *
 * Retorno: Nada
 *
 ****/
void ApresentaMenu(void)
{
   printf( "\n\t>>>>> Opcoes <<<<<\n"
           "\n\t(1) Acrescenta palavra"
           "\n\t(2) Remove palavra"
           "\n\t(3) Busca palavra"
           "\n\t(4) Numero de palavras"
           "\n\t(5) Encerra o programa"
           "\n\n\t>>> Escolha sua opcao: " );
}

/****
 *
 * main(): Testa as funes de busca, insero e remoo em tries
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   tListaSE         lista;
   tTrieMachado     raiz;
   char            *umaPalavra;
   tConteudoMachado conteudo = {0, 0};
   int              op,
                    nOcorrencias;

   raiz = ConstroiTrieMachado(NOME_ARQUIVO); /* Cria a trie */

   while (1) {
      ApresentaMenu();

      op = LeOpcao("12345");

      if (op == '5') { /* Encerra o programa */
         printf("\nBye!\n");
         break; /* Sada do lao */
      }

      switch (op) {
         case '1': /* Acrescenta uma palavra na trie */
            printf("\n>>> Digite a palavra a ser acrescentada: ");
            umaPalavra = LeLinhaIlimitada(NULL, stdin);

            if (umaPalavra && ValidaPalavra(umaPalavra) )
               InsereTrieMachado(raiz, umaPalavra, &conteudo);

            free(umaPalavra);
            printf("\n>>> Acrescimo efetuado\n");

            break;

         case '2': /* Remove uma palavra da trie */
            printf("\n>>> Digite a palavra a ser removida: ");
            umaPalavra = LeLinhaIlimitada(NULL, stdin);

            RemoveTrieMachado(&raiz, umaPalavra);

            break;

         case '3': /* Verifica se uma palavra faz parte da trie */
            printf("\n>>> Digite a palavra a ser procurada: ");
            umaPalavra = LeLinhaIlimitada(NULL, stdin);

            lista = BuscaTrieMachado(raiz, umaPalavra);

            if (!lista) {
               printf("\n>>> \"%s\" nao foi encontrado\n", umaPalavra);
            } else {
               nOcorrencias = ComprimentoListaSE(lista);

               printf( "\n>>> \"%s\" foi encontrada %d vezes. "
                       "\n>>> Deseja exibir as ocorrencias" (s/n)? ",
                       umaPalavra, nOcorrencias );

               op = LeOpcao("sSnN");

               if (op == 's' || op == 'S')
                  ExibeLista(lista);
            }

            free(umaPalavra);

            break;

         case '4': /* Numero de palavras */
            printf( "\n\n>>> Numero de chaves na trie: %d\n",
                    NumeroDeChavesTrieMachado(raiz) );

            break;

         default: /* O programa no deve chegar at aqui */
            printf("\nEste programa contem um erro\n");
            return 1;
      }
   }

   return 0;
}

/******************* Exemplo de Execuo ******************



**********************************************************/
