/****
 *
 * Ttulo: Maior Subsequncia Comum (MSC) a Dois Strings
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 05/04/2016
 * ltima modificao: 05/04/2016
 *
 * Descrio: Encontra o maior prefixo comum (MSC) a um conjunto
 *            de strings usando trie
 *
 * Entrada: No h
 *
 * Sada: Mensagens apresentando os resultados das operaes
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/****************** Definies de Macros ******************/

   /* Determina o maior valor dentre dois nmeros */
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

/***************** Definies de Funes ******************/

/****
 *
 * ExibeMSC(): Exibe na tela o comprimento do MSC de dois
 *             strings e o primeiro MSC encontrado
 *
 * Parmetros:
 *      X, Y (entrada) - os strings
 *      m, n (entrada) - os tamanhos dos strings
 *
 * Retorno: Nenhum
 *
 ****/
void ExibeMSC( const char *X, const char *Y, int m, int n )
{
   int    i, j,
          indice,
          M[m + 1][n + 1]; /* Vlido nos padres C99 e C11 */
   char  *MSC;

   /*                                            */
   /* O primeiro segmento da funo  dedicado   */
   /* ao clculo do comprimento do MSC           */
   /*                                            */

      /* Constri a tabela M[m + 1][n + 1] de modo */
      /* ascendente. M[i][j] contm o comprimento  */
      /* de MSC de X[0..i - 1] e Y[0..j - 1]       */
   for (i = 0; i <= m; ++i) {
      for (j = 0; j <= n; ++j) {
         if (i == 0 || j == 0)
            M[i][j] = 0;
         else if (X[i - 1] == Y[j - 1])
            M[i][j] = M[i - 1][j - 1] + 1;
         else
            M[i][j] = MAX(M[i - 1][j], M[i][j - 1]);
      }
   }

      /* Exibe na tela o comprimento de MSC */
   printf( "\n>>> Comprimento de MSC de %s e %s: %d\n",
           X, Y, M[m][n] );

   /*                                        */
   /* O segundo segmento a seguir  dedicado */
   /*  apresentao do MSC                  */
   /*                                        */

      /* Inicia 'indice' com o ltimo valor da tabela */
   indice = M[m][n];

      /* Cria um array de caracteres para armazenar o MSC */
   MSC = calloc(indice + 1, sizeof(char));

      /* Inicia com o ltimo elemento da matriz e prossegue */
      /* armazenando os caracteres no array MSC[]           */
   i = m;
   j = n;

   while (i > 0 && j > 0) {
         /* Se os caracteres corrente em X e Y so os     */
         /* mesmos, ento esse caractere faz parte do MSC */
      if (X[i - 1] == Y[j - 1]) {
            /* Acrescenta o caractere corrente ao resultado */
         MSC[indice - 1] = X[i - 1];

            /* Atualiza os valores de i, j e indice */
         --i;
         --j;
         --indice;
      } else if (M[i - 1][j] > M[i][j - 1]) {
            /* Os caracteres correntes em X e Y no so os */
            /* mesmos. Prossegue em direo do valor maior */
         --i;
      } else {
            /* Idem */
         --j;
      }
   }

      /* Exibe na tela o MSC */
   printf("\n>>> MSC de %s e %s: %s\n", X, Y, MSC);

   free(MSC); /* Libera o espao ocupado pelo array */
}

/****
 *
 * main(): Testa a funo ExibeMSC()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main()
{
//   char *X = "AGGTAB",
//        *Y = "GXTXAYB";
   char *X = "CTGACA",
        *Y = "ACGCTAC";
   int   m, n;

   m = strlen(X);
   n = strlen(Y);

   ExibeMSC(X, Y, m, n);

   return 0;
}

/****************** Resultado do Programa *****************

>>> Comprimento de MSC de CTGACA e ACGCTAC: 4

>>> MSC de CTGACA e ACGCTAC: CGCA

**********************************************************/
