/****
 *
 * Ttulo: Distncia de Edio
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2016
 * ltima modificao: 16/04/2016
 *
 * Descrio: Encontra a distncia de edio de dois strings
 *
 * Entrada: Strings
 *
 * Sada: Distncias de edio
 *
 ****/

#include <stdio.h>
#include <string.h>
#include <time.h>   /* Processamento de strings */

#define TAMANHO_MAX 80
#define LIMITE      12

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'
 *
 ****/
static int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeString(): L um string via teclado deixando o buffer
 *             zerado
 *
 * Parmetros:
 *      ar[] (sada): array que armazenar o string lido
 *      nElementos (entrada): nmero mximo de elementos que
 *                            sero armazenados no array,
 *                            incluindo '\0'
 *
 * Retorno: O nmero de caracteres que o usurio digitou a
 *          mais, sem incluir '\n'
 *
 ****/
int LeString(char *ar, int nElementos)
{
   char *p; /* Usado para remover '\n' do string */
   int   nCarDeixados = 0; /* Conta o nmero de     */
                           /* caracteres excedentes */
                           /* no buffer             */

      /* L no mximo nElementos - 1 caracteres ou at   */
      /* encontrar '\n'. Quando fgets() retorna NULL,    */
      /* ocorreu erro ou tentativa de leitura alm do    */
      /* final do arquivo. Nesses casos, o lao continua */
      /* efetuando novas tentativas de leitura.          */
   while ( fgets(ar, nElementos, stdin) == NULL )  {
      printf( "\a\n\t>>> %s", "Erro de leitura. "
              "Tente novamente\n\t> " );
   }

      /* Faz p apontar para o caractere que      */
      /* antecede o caractere terminal do string */
   p = strchr(ar, '\0') - 1;

      /* Se o caractere '\n' foi lido, remove-o do   */
      /* string. Caso contrrio, remove-o do buffer. */
   if (*p == '\n') {
         /* Caractere '\n' faz parte do string */
      *p = '\0'; /* Sobrescreve '\n' com '\0' */
   } else { /* Usurio digitou caracteres demais */
      nCarDeixados = LimpaBuffer();
   }

      /* Retorna o nmero de caracteres que o usurio */
      /* digitou alm do esperado, sem incluir '\n'   */
   return nCarDeixados;
}

/****
 *
 * MenorDe3(): Determina o menor dentre trs nmeros inteiros
 *
 * Parmetros:
 *      d1, d2, d3 (entrada) - os trs nmeros inteiros
 *
 * Retorno: O menor dos trs nmeros recebidos como parmetros
 *
 ****/
int MenorDe3(int d1, int d2, int d3)
{
   if(d1 < d2 && d1 < d3)
      return d1;
   else if(d1 < d3)
      return d2;
   else if(d2 < d3)
      return d2;
   else
      return d3;
}

/****
 *
 * DistanciaDeEdicao(): Calcula a distncia de edio
 *                      de dois strings
 *
 * Parmetros:
 *      s1, s2 (entrada) - os strings
 *
 * Retorno: a distncia de edio dos dois strings s1 e s2
 *
 ****/
int DistanciaDeEdicao(const char *s1, const char *s2)
{
   int da, /* Distncia de edio devido a alterao */
       di, /* Distncia de edio devido a insero  */
       dr; /* Distncia de edio devido a remoo   */

   /* Se um dos strings  vazio, a distncia */
   /* de edio  o tamanho do outro string  */

   if(*s1 == 0)
      return strlen(s2);

   if(*s2 == 0)
      return strlen(s1);

      /* Se os dois primeiros caracteres correspondentes nos dois  */
      /* strings so iguais, passa-se para os caracteres seguintes */
   if(*s1 == *s2)
      da = DistanciaDeEdicao(s1 + 1, s2 + 1);
   else /* Alterao */
      da = 1 + DistanciaDeEdicao(s1 + 1, s2 + 1);

      /* Insero */
   di = 1 + DistanciaDeEdicao(s1, s2 + 1);

      /* Remoo */
   dr = 1 + DistanciaDeEdicao(s1 + 1, s2);

   return MenorDe3(da, di, dr);
}

/****
 *
 * main(): Testa a funo DistanciaDeEdicao()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main()
{
   char s1[TAMANHO_MAX],
        s2[TAMANHO_MAX];
   int  tam1,
        tam2;

      /* Apresenta o programa ao usurio */
      /* e explica seu funcionamento     */
   printf( "\n\t>>> Este programa calcula a distancia de edicao"
           "\n\t>>> de dois strings. Digite apenas [ENTER] para"
           "\n\t>>> encerrar o programa. O numero maximo de"
           "\n\t>>> caracteres permitido num string e' %d.\n",
           TAMANHO_MAX );

   while(1) {
      printf( "\n>>> Digite o primeiro string: " );
      LeString(s1, TAMANHO_MAX);

      if (!*s1)
         break;

      printf( "\n>>> Digite o segundo string: " );
      LeString(s2, TAMANHO_MAX);

      if (!*s2)
         break;

      tam1 = strlen(s1);
      tam2 = strlen(s2);

      if ( tam1 > LIMITE || tam2 > LIMITE) {
         MedidaDeTempo();
         printf("\nCalculando distancia de edicao...");

         printf("\n\n>>> Distancia de edicao entre\n\"%s\" "
                "(%d caracteres) e\n\"%s\" (%d caracteres): %d",
                s1, tam1, s2, tam2, DistanciaDeEdicao(s1, s2));

         MedidaDeTempo();
      } else {
         printf("\n\n>>> Distancia de edicao entre\n\"%s\" "
                "(%d caracteres) e\n\"%s\" (%d caracteres): %d\n",
                s1, tam1, s2, tam2, DistanciaDeEdicao(s1, s2));
      }
   }

   return 0;
}

/******************* Exemplo de Execuo ******************

        >>> Este programa calcula a distancia de edicao
        >>> de dois strings. Digite apenas [ENTER] para
        >>> encerrar o programa. O numero maximo de
        >>> caracteres permitido num string e' 80.

>>> Digite o primeiro string: Otorrinolaring

>>> Digite o segundo string: Otorrinolarig

Calculando distancia de edicao...

>>> Distancia de edicao entre
"Otorrinolaring" (14 caracteres) e
"Otorrinolarig" (13 caracteres): 1

        *** Tempo gasto na operacao: 44.00 segundos ***

>>> Digite o primeiro string: Otorrinolaringo

>>> Digite o segundo string: Otorrinolarino

Calculando distancia de edicao...

>>> Distancia de edicao entre
"Otorrinolaringo" (15 caracteres) e
"Otorrinolarino" (14 caracteres): 1

        *** Tempo gasto na operacao: 192.00 segundos ***

>>> Digite o primeiro string: Otorrinolaringol

>>> Digite o segundo string: Otorrinolaringl

Calculando distancia de edicao...

>>> Distancia de edicao entre
"Otorrinolaringol" (16 caracteres) e
"Otorrinolaringl" (15 caracteres): 1

        *** Tempo gasto na operacao: 1183.00 segundos ***

1183.00 segundos = 19.72 minutos

**********************************************************/
