/****
 *
 * Ttulo: Gera DNA
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2015
 * ltima modificao: 08/10/2017
 *
 * Descrio: Cria um arquivo de texto contendo uma sequncia de
 *            DNA gerada aleatriamente
 *
 * Entrada: Tamanho da sequncia de DNA
 *
 * Sada: Arquivo contendo a sequncia de DNA gerada
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/***************** Constantes Simblicas ******************/

#define N_ACIDOS 4
#define NOME_ARQ "DNA_Rand.txt"

/****************** Definies de Tipos *******************/

typedef enum {A, G, C, T} tAcido;

/***************** Definies de Funes ******************/

/****
 *
 * LimpaBuffer(): Limpa o buffer associado ao meio de
 *                entrada padro (stdin)
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de caracteres descartados, com
 *          exceo do caractere '\n'.
 *
 ****/
static int LimpaBuffer(void)
{
   int carLido, /* Armazena cada caractere lido */
       nCarLidos = 0; /* Conta o nmero de */
                      /* caracteres lidos  */

      /* L e descarta cada caractere lido at    */
      /* encontrar '\n' ou getchar() retornar EOF */
   do {
      carLido = getchar(); /* L um caractere */
      ++nCarLidos; /* Mais um caractere foi lido */
   } while ((carLido != '\n') && (carLido != EOF));

      /* O ltimo caractere lido foi '\n' ou  */
      /* EOF e no deve ser considerado sobra */
   return nCarLidos - 1;
}

/****
 *
 * LeNaturalPositivo(): L um nmero natural positivo via
 *                      teclado e deixa o buffer intacto
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero natural lido
 *
 ****/
int LeNaturalPositivo(void)
{
   int num, /* O nmero lido */
       teste, /* Valor retornado por scanf() */
       nResto = 0; /* Nmero de caracteres */
                   /* excedentes           */

   /* Desvia para c se o valor for invlido */
inicio:
      /* Tenta ler um valor vlido */
   teste = scanf("%d", &num);

      /* Se no ocorreu erro de leitura ou de final */
      /* de arquivo, h caracteres remanescentes    */
      /* que precisam ser removidos                 */
   if (teste != EOF) {
      nResto = LimpaBuffer();
   }

      /* Enquanto o valor retornado por scanf()  */
      /* no indicar que um valor vlido foi     */
      /* lido continua tentando obter esse valor */
   while(teste != 1 || num < 1) {
      printf("\a\n\t>>> O valor digitado e' invalido. "
                      "Tente novamente\n\t> " );
      goto inicio; /* No causa dano algum */
   }

      /* Repreende o usurio se ele digitou demais */
   if (nResto == 1) {
      printf("\t>>> Um caractere foi descartado\n");
   } else if (nResto > 1) {
      printf( "\t>>> %d caracteres foram descartados\n",
              nResto );
   }

      /* O valor retornado certamente  vlido */
   return num;
}

/****
 *
 * main(): Cria um arquivo de texto contendo uma sequncia de DNA
 *         gerada aleatriamente
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
 int main(void)
{
   int   i,
         tamanho; /* Tamanho da sequencia (e do arquivo) */
   FILE *stream;  /* Stream associado ao arquivo */

   srand(time(NULL)); /* Alimenta o gerador de nmeros aleatrios */

   printf("\nIntroduza o tamanho da sequencia: ");
   tamanho = LeNaturalPositivo();

   stream = fopen(NOME_ARQ, "w"); /* Tenta abrir o arquivo */

   if (!stream) { /* Verifica se o arquivo foi aberto */
      printf("\nArquivo nao pode ser criado");
      return 1;
   }

      /* Gera e escreve a sequncia no arquivo */
   for (i = 0; i < tamanho; ++i) {
      switch ((tAcido) (rand()%N_ACIDOS)) {
         case A:
            fputc('A', stream);
            break;

         case G:
            fputc('G', stream);
            break;

         case C:
            fputc('C', stream);
            break;

         case T:
            fputc('T', stream);
            break;

         default:
            printf("\nErro desconhecido");;
            return 1;
      }
   }

   fclose(stream);

   return 0;
}
