/****
 *
 * Arquivo Tipos.h
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 06/10/2017
 * ltima modificao: 06/10/2017
 *
 * Descrio: Tipos utilizados pelo programa de codificao de
 *            Huffman
 *
 ****/

#ifndef _Tipos_H_ /* Previne incluso mltipla deste arquivo */
#define _Tipos_H_

/*********************** Includes *************************/

#include <limits.h>

/***************** Constantes Simblicas ******************/

   /* Nmero mximo de bits num cdigo associado a um byte */
#define MAX_BITS  (UCHAR_MAX + 1)

   /* Nmero mximo de bytes num cdigo associado a um byte */
#define MAX_BYTES MAX_BITS/CHAR_BIT

/****************** Definies de Tipos *******************/

typedef unsigned char tByte; /* Tipo de um byte */

   /* Tipo do campo de informao dos ns da rvore de codificao */
typedef struct {
          tByte  byte;
          double frequencia;
        } tByteFreq;

   /* Tipo de cada n e tipo de ponteiro para n da rvore de codificao */
typedef struct rotNoArvoreHuff {
          tByteFreq                byteFreq;
          struct rotNoArvoreHuff *filhoEsquerda;
          struct rotNoArvoreHuff *filhoDireita;
        } tNoArvoreHuff, *tArvoreHuff;

   /* Tipo de n de um heap que armazena ponteiros para */
   /* os ns da rvore de codificao temporariamente   */
typedef tNoArvoreHuff *tNoHeapHuff;

   /* Tipo de heap que armazena as informaes */
   /* sobre os ns da rvore de codificao    */
typedef struct {
           tNoHeapHuff *itens;      /* Array de elementos */
           int          capacidade; /* Quantos elementos o heap pode conter */
           int          nItens;     /* Nmero de elementos do heap */
        } tHeapHuff;

   /* Tipo de ponteiro para funo de comparao */
   /* que compara elementos do heap              */
typedef int (*tFComparaHuff) (const void *, const void *);

   /* Tipo de n de uma lista de cdigos cannicos */
typedef struct {
           tByte valor;          /* Valor do byte */
           tByte nBits;          /* Nmero de bits na codificao */
           tByte cod[MAX_BYTES]; /* Codificao do byte */
        } tNoListaCanHuff;

   /* Tipo de estrutura que armazena os dados de um cabealho */
typedef struct {
           int              tamLista; /* Nmero de elementos da lista */
                                      /* de cdigos cannicos         */
           tNoListaCanHuff *lista; /* A lista de cdigos cannicos */
        } tCabecalhoHuff;

#endif
