/****
 *
 * Exemplo: InsertionSort
 *
 * Seo: 11.2.3
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void InsertionSort(int tabela[], int nElem);
extern int* CriaTabela(int *nElementos, int min, int max);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern void LiberaTabela(int **tabela);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * InsertionSort(): Ordena uma tabela usando ordenao por
 *                  insero simples
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void InsertionSort(int tabela[], int  nElem)
{
   int i, j, aux;

      /*                                                                     */
      /* Inicialmente, o array tabela[]  considerado uma tabela ordenada.   */
      /* Ento, a cada iterao, os elementos entre 0 e j estaro ordenados. */
      /*                                                                     */

      /* Insere tabela[j] na parte ordenada da tabela */
   for (j = 1; j < nElem; ++j) {
      aux = tabela[j];

         /* Move cada elemento maior do que aux uma posio adiante */
      for (i = j-1; i >= 0 && aux < tabela[i]; --i)
         tabela[i + 1] = tabela[i];

      tabela[i + 1] = aux; /* Insere aux na posio correta */
   }
}

/****
 *
 * CriaTabela(): Cria uma tabela (array de inteiros) aleatoriamente
 *
 * Parmetros:
 *      nElementos (sada) - nmero de elementos na
 *                           tabela criada
 *      min (entrada) - nmero mnimo de elementos
 *                      na tabela criada
 *      max (entrada) - nmero mximo de elementos na
 *                      tabela criada
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
int* CriaTabela(int *nElementos, int min, int max)
{
   int *tabela, nElem, i, maiorElemento;

      /* Gera nmero de ns da tabela */
   nElem = (int) rand()%(max-min+1) + min;

   ASSEGURA( nElem >= min && nElem <= max,
             "Numero de elementos fora do intervalo" );

   tabela = calloc(nElem, sizeof(int));
   ASSEGURA( tabela, "Nao foi possivel alocar a tabela" );

      /* Isto no  necessrio e  feito apenas para */
      /* facilitar a visualizao dos resultados     */
   maiorElemento = 2*max;


      /* Gera valores dos elementos entre 0 e maiorElemento */
   for (i = 0; i < nElem; ++i)
      tabela[i] = (int) rand()%maiorElemento;

   *nElementos = nElem;

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando InsertionSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int *tabela; /* Ponteiro para as tabela utilizada */
   int  nElementos;

   srand((int) time(NULL));

   tabela = CriaTabela(&nElementos, MIN_EL, MAX_EL);

   printf("\n\n>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<\n");

   printf("\n************ Antes da Ordenacao *************\n");
   ApresentaTabela(tabela, nElementos);

   InsertionSort(tabela, nElementos);

   if (!EstaOrdenada(tabela, nElementos)) {
      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
      return 1;
   }

   printf("\n\n************ Depois da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   LiberaTabela(&tabela);

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<

************ Antes da Ordenacao *************

Numero de elementos: 12
{10, 11, 18, 11, 24, 2, 16, 28, 3, 8, 14, 0}

************ Depois da Ordenacao ************

Numero de elementos: 12
{0, 2, 3, 8, 10, 11, 11, 14, 16, 18, 24, 28}

**********************************************************/
