/*****
 * Exemplo: QuickSort com mediana de trs e ponto de corte
 *
 * Seo: 11.3.1
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */
#include <string.h> /* memmove()                      */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

   /* Ponto de corte */
#define PONTO_DE_CORTE 3

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void QuickSort2(int tabela[], int nElem);
extern int* CriaTabela(int *nElementos, int min, int max);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern void LiberaTabela(int **tabela);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * TrocaGenerica(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis
 *                                     que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void TrocaGenerica(void *var1, void *var2, size_t tam)
{
   void *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, var1, tam);
   memmove(var1, var2, tam);
   memmove(var2, p, tam);

   free(p);
}

/****
 *
 * InsertionSort(): Ordena uma tabela usando ordenao por
 *                  insero simples
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      inf (entrada) - limite inferior da tabela
 *      sup (entrada) - limite superior da tabela
 *
 * Retorno: Nada
 *
 ****/
void InsertionSort(int tabela[], int  nElem)
{
   int i, j, aux;

      /*                                              */
      /* Inicialmente, o array tabela[]  considerado */
      /* uma tabela ordenada. Ento, a cada iterao, */
      /* os elementos entre 0 e j estaro ordenados.  */
      /*                                              */

      /* Insere tabela[j] na parte ordenada da tabela */
   for (j = 1; j < nElem; ++j) {
      aux = tabela[j];

         /* Move cada elemento maior do */
         /* que aux uma posio adiante */
         /*      */
      for (i = j-1; i >= 0 && aux < tabela[i]; --i)
         tabela[i + 1] = tabela[i];

      tabela[i + 1] = aux; /* Insere aux na posio corsreta */
   }
}

/****
 *
 * Quick2(): Ordena uma tabela usando QuickSort com mediana de
 *           trs e ponto de corte
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser ordenada
 *      inf (entrada) - limite inferior da tabela
 *      sup (entrada) - limite superior da tabela
 *
 * Retorno: Nada
 *
 * Observaes:
 *   1. Por simplicidade, i.e., com o objetivo didtico
 *      de demonstrar o mtodo esta funo  implementada
 *      recursivamente, mas evidentemente uma implementao
 *      no recursiva seria mais eficiente
 *   2. O piv  escolhido usando a tcnica mediana de trs
 *   3. Quando o ponto de corte  atingido a tabela  ordenada
 *      usando InsertionSort
 *
 ****/
static void Quick2(int tabela[], int inf, int sup)
{
   int meio,  /* ndice do elemento no meio da tabela */
       esq,   /* ndice da esquerda */
       dir,   /* ndice da direita  */
       pivo;  /* O piv             */

      /* Se o tamanho da tabela ficar abaixo do ponto de */
      /* corte, chama InsertionSort() para orden-la     */
   if (sup - inf + 1 <= PONTO_DE_CORTE)
      InsertionSort(tabela + inf, sup - inf + 1);
   else {
      /*                                             */
      /* Escolhe o piv usando mediana de trs. Alm */
      /* disso, ao final, os elementos nas posies  */
      /* inf, sup e meio estaro ordenados.          */
      /*                                             */

         /* Obtm o ndice do elemento que */
         /* se encontra no meio da tabela  */
      meio = (inf + sup)/2;

         /* Ordena o elemento do meio e o primeiro elemento */
      if (tabela[meio] < tabela[inf])
         TrocaGenerica(tabela + inf, tabela + meio, sizeof(tabela[0]));

         /* Ordena o primeiro e o ltimo elementos */
      if (tabela[sup] < tabela[inf])
         TrocaGenerica(tabela + inf, tabela + sup, sizeof(tabela[0]));

         /* Ordena o elemento do meio e o ltimo elemento */
      if (tabela[sup] < tabela[meio])
         TrocaGenerica(tabela + meio, tabela + sup, sizeof(tabela[0]));

         /* O piv ser a mediana dos trs elementos ordenados */
      pivo = tabela[meio];

         /* Troca o piv de posio com o primeiro elemento */
      TrocaGenerica(tabela + meio, tabela + inf, sizeof(tabela[0]));

      /*                          */
      /* Final de mediana de trs */
      /*                          */

         /*                                */
         /* Incio da operao de partio */
         /*                                */

      for (esq = inf, dir = sup - 1; ; ) {
            /* Enquanto os elementos forem menores do que */
            /* o piv, incrementa-se o ndice esquerdo    */
         while (tabela[esq] <= pivo && esq < sup)
            ++esq;

            /* Enquanto os elementos forem maiores do que */
            /* o piv, decrementa-se o ndice direito     */
         while (tabela[dir] > pivo)
            --dir;

            /* Se os ndices se cruzarem, */
            /* a diviso foi concluda    */
         if (esq >= dir)
            break;

            /* Os elementos nos ndices esq e dir esto  */
            /* em parties erradas e devem ser trocados */
         TrocaGenerica(tabela + esq, tabela + dir, sizeof(tabela[0]));
      }

         /*                             */
         /* Fim da operao de partio */
         /*                             */

         /* Coloca o piv em sua correta posio */
      TrocaGenerica(tabela + inf, tabela + dir, sizeof(tabela[0]));

         /* Ordena recursivamente a primeira partio da tabela */
      Quick2(tabela, inf, dir - 1);

         /* Ordena recursivamente a segunda partio da tabela */
      Quick2(tabela, dir + 1, sup);
   }
}

/****
 *
 * QuickSort2(): Ordena uma tabela usando quick sort. Esta funo
 *              simplesmente prov uma interface para a funo
 *              Quick2() que realmente faz o servio.
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void QuickSort2(int tabela[], int nElem)
{
      /* Simplesmente chama Quick2() para fazer o servio */
   Quick2(tabela, 0, nElem - 1);
}

/****
 *
 * CriaTabela(): Cria uma tabela (array de inteiros)
 *               aleatoriamente
 *
 * Parmetros:
 *      nElementos (sada) - nmero de elementos na
 *                           tabela criada
 *      min (entrada) - nmero mnimo de elementos
 *                      na tabela criada
 *      max (entrada) - nmero mximo de elementos na
 *                      tabela criada
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
int* CriaTabela(int *nElementos, int min, int max)
{
   int *tabela, nElem, i, maiorElemento;

      /* Gera nmero de ns da tabela */
   nElem = (int) rand()%(max-min+1) + min;

   ASSEGURA( nElem >= min && nElem <= max,
             "Numero de elementos fora do intervalo" );

   tabela = calloc(nElem, sizeof(int));
   ASSEGURA( tabela, "Nao foi possivel alocar a tabela" );

      /* Isto no  necessrio e  feito apenas para */
      /* facilitar a visualizao dos resultados     */
   maiorElemento = 2*max;


      /* Gera valores dos elementos entre 0 e maiorElemento */
   for (i = 0; i < nElem; ++i)
      tabela[i] = (int) rand()%maiorElemento;

   *nElementos = nElem;

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero
 *                         exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na
 *                             tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser
 *                               liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando QuickSort com mediana de trs e
 *         ponto de corte
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int *tabela; /* Ponteiro para as tabela utilizada */
   int  nElementos;

   srand((int) time(NULL));

   tabela = CriaTabela(&nElementos, MIN_EL, MAX_EL);

   printf("\n\n>>>>>>>>>>>>>>>> QuickSort2 <<<<<<<<<<<<<<<<\n");

   printf("\n*********** Antes da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);

   QuickSort2(tabela, nElementos);

   if (!EstaOrdenada(tabela, nElementos)) {
      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
      return 1;
   }

   printf("\n\n*********** Depois da Ordenacao ***********\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   LiberaTabela(&tabela);

   return 0;
}

/******************* Exemplo de Execuo ******************

BUGGY??? Ponto de corte no funciona

**********************************************************/
