/****
 *
 * Exemplo: QuickSort3 com piv escolhido aleatoriamente
 *
 * Seo: 11.3.1
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Alocao de memria, rand() e srand() */
#include <time.h>   /* Funes de cronometragem       */
#include <string.h> /* memmove()                      */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void QuickSort3(int tabela[], int nElem);
extern int* CriaTabela(int *nElementos, int min, int max);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern void LiberaTabela(int **tabela);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * TrocaGenerica(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis
 *                                     que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void TrocaGenerica(void *var1, void *var2, size_t tam)
{
   void *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, var1, tam);
   memmove(var1, var2, tam);
   memmove(var2, p, tam);

   free(p);
}

/****
 *
 * NAleatorio(): Retorna um nmero aleatrio entre os valores
 *               recebidos como parmetros
 *
 * Parmetros:
 *      m, n (entrada) - valores que definem o intervalo de
 *                       nmeros sorteados
 *
 * Retorno: O nmero aleatrio gerado
 *
 * Observaes:
 *      1.  indiferente se m <= n ou n <= m
 *      2. Quando chamada pela primeira vez, esta funo
 *         alimenta o gerador de nmeros aleatrios
 *
 ****/
static int NAleatorio(int m, int n)
{
   static int primeiraChamada = 1; /* Esta varivel checa se a  */
                                   /* funo est sendo chamada */
                                   /* pela primeira vez e deve  */
                                   /* ter durao fixa          */
   int        menor, maior;

      /* Se esta for a primeira chamada da funo, */
      /* alimenta o gerador de nmeros aleatrios  */
   if (primeiraChamada) {
         /* Alimenta o gerador de nmeros aleatrios */
      srand(time(NULL));

         /* A prxima chamada no ser mais a primeira */
      primeiraChamada = 0;
   }

      /* Leva em considerao o fato de n  */
      /* poder ser menor ou maior do que m */
   if (n < m) {
      menor = n;
      maior = m;
   } else {
      menor = m;
      maior = n;
   }

   return rand()%(maior - menor + 1) + menor;
}

/****
 *
 * Quick3(): Ordena uma tabela usando o algoritmo QuickSort com
 *           o piv escolhido aleatoriamente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser ordenada
 *      inf (entrada) - limite inferior da tabela
 *      sup (entrada) - limite superior da tabela
 *
 * Retorno: Nada
 *
 * Observaes:
 *   1. Por simplicidade, i.e., com o objetivo didtico
 *      de demonstrar o mtodo esta funo  implementada
 *      recursivamente, mas evidentemente uma implementao
 *      no recursiva seria mais eficiente
 *   2. O piv  escolhido aleatoriamente
 *
 ****/
static void Quick3(int tabela[], int inf, int sup)
{
   int esq,   /* ndice da esquerda */
       dir,   /* ndice da direita  */
       iPivo; /* ndice do piv     */

      /* Se a tabela estiver vazia ou contiver    */
      /* apenas um elemento, ela j est ordenada */
   if (inf >= sup)
      return; /* A tabela j est ordenada */

      /* Nesta implementao, o piv  escolhido ao acaso */
   iPivo = NAleatorio(inf, sup);

      /* Troca o piv de lugar com o primeiro elemento */
   TrocaGenerica(tabela + iPivo, tabela + inf, sizeof(tabela[0]));

      /* O piv agora  o primeiro elemento */
   iPivo = inf;

      /* Cria as parties */
   for (esq = inf, dir = sup; esq < dir; ) {
         /* Enquanto os elementos da tabela forem menores */
         /* do que o piv, avana-se o ndice inferior    */
      while (tabela[esq] <= tabela[iPivo] && esq < sup)
         ++esq;

         /* Enquanto os elementos da tabela forem maiores */
         /* do que o piv, recua-se o ndice superior     */
      while (tabela[dir] > tabela[iPivo])
         --dir;

         /* Os elementos nos ndices esq e dir esto em parties */
         /* erradas e  preciso troc-los de lugar            */
      if (esq < dir)
         TrocaGenerica(tabela + esq, tabela + dir, sizeof(tabela[0]));
   }

      /* Coloca o piv em sua correta posio */
   TrocaGenerica(tabela + iPivo, tabela + dir, sizeof(tabela[0]));

      /* Ordena recursivamente a primeira metade da tabela */
   Quick3(tabela, inf, dir - 1);

      /* Ordena recursivamente a segunda metade da tabela */
   Quick3(tabela, dir + 1, sup);
}

/****
 *
 * QuickSort3(): Ordena uma tabela usando quick sort. Esta funo
 *               simplesmente prov uma interface para a funo
 *               Quick3() que realmente faz o servio.
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void QuickSort3(int tabela[], int nElem)
{
      /* Simplesmente chama Quick3() para fazer o servio */
   Quick3(tabela, 0, nElem - 1);
}

/****
 *
 * CriaTabela(): Cria uma tabela (array de inteiros)
 *               aleatoriamente
 *
 * Parmetros:
 *      nElementos (sada) - nmero de elementos na
 *                           tabela criada
 *      min (entrada) - nmero mnimo de elementos
 *                      na tabela criada
 *      max (entrada) - nmero mximo de elementos na
 *                      tabela criada
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
int* CriaTabela(int *nElementos, int min, int max)
{
   int *tabela, nElem, i, maiorElemento;

      /* Gera nmero de ns da tabela */
   nElem = (int) rand()%(max-min+1) + min;

   ASSEGURA( nElem >= min && nElem <= max,
             "Numero de elementos fora do intervalo" );

   tabela = calloc(nElem, sizeof(int));
   ASSEGURA( tabela, "Nao foi possivel alocar a tabela" );

      /* Isto no  necessrio e  feito apenas para */
      /* facilitar a visualizao dos resultados     */
   maiorElemento = 2*max;


      /* Gera valores dos elementos entre 0 e maiorElemento */
   for (i = 0; i < nElem; ++i)
      tabela[i] = (int) rand()%maiorElemento;

   *nElementos = nElem;

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *     tabela (entrada) - a tabela cujos elementos sero exibidos
 *     nElementos (entrada) - nmero mnimo de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *     tabela (entrada/sada) - a tabela cujo espao ser liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando QuickSort3 com piv escolhido
 *         aleatoriamente
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int *tabela; /* Ponteiro para as tabela utilizada */
   int  nElementos;

   srand((int) time(NULL));

   tabela = CriaTabela(&nElementos, MIN_EL, MAX_EL);

   printf("\n\n>>>>>>>>>>>>>>>> QuickSort3 <<<<<<<<<<<<<<<\n");

   printf("\n*********** Antes da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);

   QuickSort3(tabela, nElementos);

   printf("\n\n*********** Depois da Ordenacao ***********\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   if (!EstaOrdenada(tabela, nElementos)) {
      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
      return 1;
   }

   LiberaTabela(&tabela);

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>>> QuickSort3 <<<<<<<<<<<<<<<

*********** Antes da Ordenacao ************

Numero de elementos: 16
{32, 3, 9, 21, 38, 5, 6, 35, 39, 12, 15, 32, 1, 36, 4, 32}

*********** Depois da Ordenacao ***********

Numero de elementos: 16
{38, 1, 21, 6, 3, 4, 5, 9, 12, 15, 32, 39, 32, 35, 632490599, 32}

>>> A tabela NAO foi ordenada <<<

**********************************************************/
