/****
 *
 * Exemplo: CountingSort
 *
 * Seo: 11.4.1
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */
#include <limits.h> /* Limites inteiros               */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   5

   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   10

   /* Valor mximo de um elemento na tabela a ser ordenada */
#define VALOR_MAX 20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/************************* Aluses ************************/

extern void CountingSort(int tabela[], int nElem);
extern int* CriaTabela(int *nElementos, int min, int max);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern void LiberaTabela(int **tabela);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * CountingSort(): Ordena uma tabela usando CountingSort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void CountingSort(int tabela[], int nElem)
{
   int *resultado, /* Apontar para o array auxiliar que */
                   /* conter os elementos ordenados     */
       *contagem,  /* Apontar para o array de contagem  */
       maiorValor, /* Maior valor da tabela */
       tamCont,    /* Tamanho do array de contagem */
       i;

      /* Encontra o maior valor da tabela */
   maiorValor = tabela[0];
   for (i = 1; i < nElem; ++i)
      if (tabela[i] > maiorValor)
         maiorValor = tabela[i];

            /*** Passo 1 ***/

   tamCont = maiorValor + 1; /* Calcula o tamanho do array de contagem */

      /* Aloca o array de contagem */
   ASSEGURA( contagem = calloc(tamCont, sizeof(int)),
             "Array de contagem nao foi alocado" );

            /*** Passo 2 ***/

      /* Efetua a contagem de elementos e armazena-a no array de contagem */
   for (i = 0; i < nElem; i++)
      ++contagem[tabela[i]];

            /*** Passo 3 ***/

      /* Altera o array de contagem de modo que contagem[i] passe */
      /* a conter as posies  desse elemento no array ordenado   */
   for (i = 1; i < tamCont; ++i)
      contagem[i] += contagem[i - 1];

            /*** Passo 4 ***/

      /* Aloca o array que conter o resultado ordenado */
   ASSEGURA( resultado = calloc(nElem, sizeof(int)),
             "Array auxiliar nao foi alocado" );

      /* Copia cada elemento da lista original para sua posio ordenada    */
      /* no array que conter o resultado. Esse posicionamento  feito do   */
      /* ltimo para o primeiro elemento para que a ordenao seja estvel. */
   for (i = nElem - 1; i >= 0; --i) {
         /* Armazena um elemento da lista original em sua     */
         /* posio ordenada no array que conter o resultado */
      resultado[contagem[tabela[i]] - 1] = tabela[i];
         /* Decrementa o valor deste elemento do array de contagem para que */
         /* o prximo elemento da lista que seja igual ao elemento corrente */
         /* seja colocado uma posio atrs                                 */
      --contagem[tabela[i]];
   }

            /*** Passo 5 ***/

      /* Copia o resultado para a tabela original */
   for (i = 0; i < nElem; ++i)
      tabela[i] = resultado[i];

      /* Libera o espao ocupado pelos arrays auxiliares */
   free(contagem);
   free(resultado);
}

/****
 *
 * CriaTabela(): Cria uma tabela (array de inteiros) aleatoriamente
 *
 * Parmetros:
 *     nElementos (sada) - nmero de elementos na tabela criada
 *     min (entrada) - nmero mnimo de elementos na tabela criada
 *     max (entrada) - nmero mximo de elementos na tabela criada
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
int* CriaTabela(int *nElementos, int min, int max)
{
   int *tabela, nElem, i;

      /* Gera nmero de elementos da tabela */
   nElem = rand()%(max - min + 1) + min;

   ASSEGURA( nElem >= min && nElem <= max,
             "Numero de elementos fora do intervalo" );

   tabela = calloc(nElem, sizeof(int));
   ASSEGURA( tabela, "Nao foi possivel alocar a tabela" );

      /* Gera valores dos elementos entre 0 e VALOR_MAX */
   for (i = 0; i < nElem; ++i)
      tabela[i] = (int) rand()%VALOR_MAX;

   *nElementos = nElem;

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);

   if (nElementos <= 0)
      return;

   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando CountingSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int *tabela; /* Ponteiro para as tabela utilizada */
   int  nElementos;

   srand((int) time(NULL));

   tabela = CriaTabela(&nElementos, MIN_EL, MAX_EL);

   printf("\n\n>>>>>>>>>>>>>>>> CountingSort <<<<<<<<<<<<<<<<\n");

   printf("\n*********** Antes da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);

   CountingSort(tabela, nElementos);

   printf("\n\n*********** Depois da Ordenacao ***********\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   if (!EstaOrdenada(tabela, nElementos)) {
      printf("\n\n\a>>> Erro: A tabela NAO foi ordenada <<<\n");
      return 1;
   }


   LiberaTabela(&tabela);

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>>> CountingSort <<<<<<<<<<<<<<<<

*********** Antes da Ordenacao ************

Numero de elementos: 6
{1, 17, 17, 9, 0, 8}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 2}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{1, 2, 2, 2, 2, 2, 2, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 6}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{1, 1, 2, 2, 2, 2, 2, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 6}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{1, 1, 2, 2, 2, 2, 2, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{1, 1, 2, 2, 2, 2, 2, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4}

        ***** Debug: Array de contagem *****
Numero de elementos: 18
{0, 1, 2, 2, 2, 2, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4}

*********** Depois da Ordenacao ***********

Numero de elementos: 6
{0, 1, 8, 9, 17, 17}

**********************************************************/
