/****
 *
 * Ttulo: Convertendo um Arquivo de Registros de Texto para Binrio
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 21/05/2012
 * ltima modificao: 21/05/2012
 *
 * Descrio: L cada linha de um arquivo de texto, converte-a
 *            numa estrutura do tipo tRegistroCEP e armazena-a num arquivo
 *            binrio
 *
 * Entrada: Arquivo cujo nome  especificado pela constante
 *          NOME_ARQUIVO_TEXTO
 *
 * Sada: Arquivo cujo nome  especificado pela constante
 *        NOME_ARQUIVO_BIN
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Entrada e Sada              */
#include <string.h>    /* Processamento de strings     */
#include <stdlib.h>    /* strtol() e alocao dinmica */
#include "ASSEGURA.h"  /* Macro ASSEGURA               */
#include "Registros.h" /* Interface deste mdulo       */
#include "Leitura.h"   /* LeLinhaIlimitada()           */

/******************** Funes de Locais *******************/

/****
 *
 * ObtemTokens(): Separa um string em partes (tokens)
 *
 * Parmetros:
 *     str (entrada/sada) - o string que ser separado
 *                           ou NULL para continuar uma
 *                           separao anterior
 *     sep (entrada) - string contendo os separadores de tokens
 *
 * Retorno: Endereo do prximo token do string, se este
 *          existir, ou NULL, em caso contrrio
 *
 * Observao: Esta funo NO  funcionalmente equivalente 
 *             funo strtok() da biblioteca padro de C
 *
****/
static char *ObtemTokens(char *str, char const *sep)
{
   static char *proximoToken; /* Aponta para o prximo */
                              /* token, se ele existir */
   char        *s, /* Apontar para o string no qual a */
                   /* busca pelo token ser efetuada   */
               *inicio; /* Guardar o incio */
                        /* do token corrente */

   if (str != NULL)
      proximoToken = str;

   if (proximoToken == NULL)
      return NULL;

   if ((s = strpbrk(proximoToken, sep)) != NULL) {
      *s  = '\0';
      inicio = proximoToken;
      proximoToken = ++s;
   } else if (*proximoToken) {
      inicio = proximoToken;
      proximoToken = NULL;
   }

   return inicio;
}

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa
 *                    estrutura do tipo tRegistroCEP.
 *
 * Parmetros: regCEP (sada) - ponteiro para a estrutura
 *                             resultante da converso
 *             linha (entrada/sada) - linha que ser
 *                                     convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado
 *          da converso
 *
 * Observaes:
 *     1. Para facilitar o entendimento, esta funo assume
 *        que o parmetro 'linha' realmente  um string no
 *        formato das linhas do arquivo. Portanto, os
 *        valores retornados por ObtemTokens() no so testados
 *        como deveriam.
 *
 *     2. O parmetro 'linha'  alterado por ObtemTokens().
 *
 *     3. A funo ObtemTokens()  usada em lugar de strtok()
 *        devido  presena de campos vazios no arquivo de
 *        texto, o que faz com que dois separadores apaream
 *        juntos. Nesse caso, a funo strtok() consideraria
 *        os dois separadores como sendo apenas um.
 *
 ****/
static tRegistroCEP *LinhaEmRegistro(tRegistroCEP *regCEP, char *linha)
{
   char *str; /* Apontar para cada token da linha */
   char *final; /* Final de uma converso numrica */

      /* Obtm o nmero e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(linha, "\t;\n");
   regCEP->numero = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm a UF e acrescenta-o ao  */
      /* respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->UF, str, TAM_UF);

      /* Obtm o nmero de localidade e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->localidadeNumero = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->localidadeNumero < 0,
            "Erro de conversao numerica");

      /* Obtm o nome abreviado e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->nomeAbr, str, MAX_NOME);

      /* Obtm o nome e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->nome, str, MAX_NOME);

      /* Obtm o bairro incio e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->bairroInicio = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->bairroInicio < 0,
            "Erro de conversao numerica");

      /* Obtm o bairro final e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->bairroFim = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->bairroFim < 0,
            "Erro de conversao numerica");

      /* Obtm o CEP e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->CEP = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm o complemento e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->complemento, str, MAX_COMP);

      /* Obtm o tipo de logradouro e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->tipoLogradouro, str, MAX_TIPO_LOG);

      /* Obtm o tipo de logradouro e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->statusTipoLogradouro = *str;

      /* Obtm o nome sem acento e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->nomeSemAcento, str, MAX_NOME);

      /* Obtm a chave DNE e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->chaveDNE, str, TAM_DNE);

   return regCEP;
}

/******************** Funes de Globais ******************/

/****
 *
 * AbreArquivo(): Tenta abrir um arquivo e, quando isso no 
 *                possvel, aborta o programa
 *
 * Parmetros:
 *      nomeArq (entrada) - nome do arquivo ou NULL
 *      modo (entrada) - modo de abertura do arquivo
 *
 * Retorno: Stream associado ao arquivo aberto
 *
 ****/
FILE *AbreArquivo(const char *nomeArq, const char *modo)
{
   FILE *stream = fopen(nomeArq, modo);

      /* Se ocorreu erro de abertura, aborta o programa */
   if (!stream) {
      fprintf( stderr, "\a\n>>> Erro na abertura do arquivo %s."
              "\n>>> O programa sera' encerrado.\n", nomeArq);
      exit(1); /* Aborta o programa */
   }

   return stream;
}

/****
 *
 * FechaArquivo(): Tenta fechar um arquivo e, quando isso
 *                 no  possvel, aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      nomeArq (entrada) - nome do arquivo ou NULL
 *
 * Retorno: Nada
 *
 * Observao: Se o segundo parmetro no for NULL, o nome
 *             do arquivo aparece na mensagem de erro
 *
 ****/
void FechaArquivo(FILE *stream, const char *nomeArq)
{
   /***********************************************************/
   /* Se fclose() retornar um valor diferente de zero ocorreu */
   /* algum erro na tentativa de fechamento do arquivo. Nesse */
   /* caso, apresenta mensagem de erro e aborta o programa.   */
   /***********************************************************/

   if (fclose(stream)) { /* Erro de fechamento */
      fprintf( stderr, "\a\n>>> Ocorreu erro no fechamento "
              "do arquivo %s.\n>>> O programa sera' "
              "encerrado.\n", nomeArq ? nomeArq : "");
      exit(1); /* Aborta o programa */
   }
}

/****
 *
 * TamanhoDeArquivo(): Calcula o tamanho de um arquivo lendo cada
 *                     byte que o constitui e contando quantos
 *                     bytes so lidos
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *
 * Retorno: -1, se ocorrer algum erro.
 *          O tamanho do arquivo, em caso contrrio.
 *
 * Observaes:
 *     1. O arquivo deve estar aberto em modo binrio que permite
 *        leitura
 *     2. A posio atual do indicador de posio do arquivo 
 *        preservada
 *
 ****/
int TamanhoDeArquivo(FILE *stream)
{
   int posicaoAtual, tamanho = 0;

      /* Guarda a posio atual do indicador de posio */
      /* do arquivo para que ele possa ser restaurado   */
   posicaoAtual = ftell(stream);

      /* Verifica se ftell() foi bem sucedida */
   if (posicaoAtual < 0) {
      return -1; /* A chamada de ftell() falhou */
   }

      /* Assegura leitura a partir do incio do arquivo */
   if ( fseek(stream, 0, SEEK_SET) ) {
         /* O movimento do indicador de posio no foi   */
         /* possvel. Neste caso, no precisa restaur-lo */
      return -1;
   }

      /* O lao encerra quando o final do arquivo for */
      /* atingido ou se ocorrer erro de leitura       */
   while (1){
         /* L e descarta o prximo byte */
      fgetc(stream);

         /* Checa se o arquivo j foi inteiramente lido */
      if (feof(stream)) {
         break; /* Encerra lao */
      }

         /* Verifica se ocorreu erro de leitura */
      if (ferror(stream)) {
            /* Restaura a posio original do indicador */
            /* de posio do arquivo  antes de retornar */
         fseek(stream, posicaoAtual, SEEK_SET);
         return -1; /* Ocorreu erro de leitura */
      }

      ++tamanho; /* Mais um byte foi lido */
   }

      /* Restaura a posio original do indicador de posio */
   if ( fseek(stream, posicaoAtual, SEEK_SET) ) {
      return -1; /* O movimento no foi possvel */
   }

   return tamanho;
}

/****
 *
 * NumeroDeRegistros(): Calcula o nmero de registros (i.e.,
 *                      parties de tamanho especificado) de um
 *                      arquivo binrio
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *    tamRegistro (entrada) - tamanho de cada partio do arquivo
 *
 * Retorno: O nmero de parties do arquivo, se no ocorrer
 *          nenhum erro, ou -1, em caso contrrio.
 *
 * Observao: O arquivo deve estar aberto em modo binrio que
 *             permite leitura
 *
 ****/
int NumeroDeRegistros( FILE *stream, int tamRegistro )
{
   int    nBytes = 0; /* Nmero de bytes do arquivo */

      /* Calcula o nmero de bytes do arquivo */
   nBytes = TamanhoDeArquivo(stream);

      /* Verifica se ocorreu erro na contagem de bytes */
   if (nBytes < 0) {
      return -1; /* Erro na contagem de bytes */
   }

      /* O nmero de registros  o nmero de bytes do arquivo */
      /* dividido pelo nmero de bytes de cada registro       */
   return nBytes/tamRegistro;
}

/****
 *
 * CriaArquivoApenasCEPs(): L cada linha de um arquivo no formato
 *                   especificado, converte-a numa estrutura
 *                   do tipo tRegistroCEP e armazena a estrutura
 *                   num arquivo binrio.
 *
 * Parmetros:
 *     arqTexto (entrada) - nome do arquivo que ser lido
 *     arqBin (entrada) - nome do arquivo que ser escrito
 *     maxRegs (entrada) - nmero mximo de registros que sero
 *                         processados
 *
 * Retorno: Nmero de estruturas escritas no arquivo binrio
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de
 *        texto que permite leitura
 *     2. O stream de sada deve estar aberto em modo "wb"
 *
 ****/
int CriaArquivoApenasCEPs( const char *arqTexto, const char *arqBin,
                    int maxRegs )
{
   char  *linha; /* Apontar para cada linha lida */
   tRegistroCEP   umCEP; /* Dados de uma linha      */
                 /* convertida em estrutura */
   int    nRegistros = 0; /* Nmero de estruturas */
                          /* escritas no arquivo binrio */
   FILE  *streamTexto, /* Associado ao arquivo de texto */
         *streamBin;   /* Associado ao arquivo binrio */

      /* Tenta abrir arquivo para leitura em modo texto */
   streamTexto = AbreArquivo(arqTexto, "r");

      /* Tenta criar arquivo binrio */
   streamBin = AbreArquivo(arqBin, "wb");

      /* L cada linha do arquivo, cria um registro do tipo   */
      /* tRegistroCEP e armazena-o no arquivo binrio. O lao encerra */
      /* quando h tentativa de leitura alm do final do      */
      /* arquivo de entrada, ocorre erro de leitura/escrita   */
      /* ou quando o nmero mximo de registros for lido      */
   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(NULL, streamTexto);

         /* Verifica se o final do arquivo foi atingido */
      if (feof(streamTexto))
         break;

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA(!ferror(streamTexto), "Erro de leitura");

         /* Converte a linha lida em estrutura */
      LinhaEmRegistro(&umCEP, linha);

         /* Escreve a estrutura resultante da */
         /* converso no arquivo binrio      */
      fwrite(&umCEP.CEP, sizeof(umCEP.CEP), 1, streamBin);

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA(!ferror(streamBin), "Erro de escrita");

      ++nRegistros; /* Mais um registro lido */

      if (maxRegs && nRegistros == maxRegs)
         break;
   }

      /* Fecha os arquivos */
   FechaArquivo(streamTexto, arqTexto);
   FechaArquivo(streamBin, arqBin);

   return nRegistros; /* No ocorreu nenhum erro */
}
