/****
 *
 * Ttulo: Ordenao de Ponteiros
 *
 * Seo 11.8.2
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/05/2017
 * ltima modificao: 21/05/2018
 *
 * Descrio: Exemplo de ordenao de listas encadeadas
 *
 * Entrada: Nada
 *
 * Sada: V. Exemplo de Execuo
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string.h>

/***************** Constantes Simblicas ******************/

   /* Nmero mximo de caracteres num nome */
#define MAX_NOME      20

   /* Nmero de caracteres numa matrcula  */
#define TAM_MATR      4

   /* Nmero mximo de elementos na tabela  */
#define MAX_ELEMENTOS 20

   /* Nome do arquivo binrio (diretrio corrente) */
#define NOME_ARQ_BIN "Tudor.bin"

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/****************** Definies de Tipos *******************/

   /* Tipo do nome */
typedef char tNome[MAX_NOME + 1];

   /* Tipo da matrcula */
typedef char tMatricula[TAM_MATR + 1];

typedef struct {
           tNome      nome;   /* Nome do aluno */
           tMatricula matr;   /* Sua matrcula */
           double     n1, n2; /* Suas notas    */
        } tRegistro;

typedef struct {
           tRegistro elementos[MAX_ELEMENTOS];
           int       nElementos;
        } tTabela;

/************************* Aluses ************************/

extern void ExibeElemento(const tRegistro *elemento);
extern void ExibeTabela(const tTabela *tabela);
extern void ConstroiTabela(const char *arq, tTabela *tabela);
extern tRegistro **AssociaPonteiros(tRegistro tabela[], int nElem);
extern void OrdenaPonteiros( tRegistro *ptr[], int nElem,
                             int (*F)(void *e1, void *e2) );
extern int Compara(void *e1, void *e2);
extern void ExibeTabelaOrdenada(tRegistro *ptr[], int nElem);

/***************** Definies de Funes ******************/

/****
 *
 * Troca(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis
 *                                     que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void Troca(void *var1, void *var2, size_t tam)
{
   void *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, var1, tam);
   memmove(var1, var2, tam);
   memmove(var2, p, tam);

   free(p);
}

/****
 *
 * ExibeElemento(): Exibe na tela um elemento de uma tabela
 *
 * Parmetros: *elemento (entrada) - elemento que ser exibido
 *
 * Retorno: Nada
 *
 ****/
void ExibeElemento(const tRegistro *elemento)
{
   printf("\n%s", elemento->nome);
   printf("\t%s", elemento->matr);
   printf("\t%3.1f", elemento->n1);
   printf("\t%3.1f", elemento->n2);
}

/****
 *
 * ExibeTabela(): Apresenta na tela, em forma de tabela, os
 *                elementos de uma lista indexada de estruturas
 *                do tipo tRegistro
 *
 * Parmetros:
 *     *tabela (entrada) - a tabela
 *
 * Retorno: Nada
 *
 ****/
void ExibeTabela(const tTabela *tabela)
{
   tRegistro elemento;
   int       i, tamanho = tabela->nElementos;

      /* Se a tabela estiver vazia, no h o que exibir */
   if (!tamanho) {
      printf("\n*** Lista Vazia *** \n");
      return;
   }

      /* Cabealho de apresentao na tela */
   printf( "\n   Nome   \tMatr\tN1\tN2"
           "\n   ====   \t====\t==\t==\n");

      /* Apresenta cada estrutura numa linha separada */
   for (i = 0; i < tamanho; ++i) {
      elemento = tabela->elementos[i];
      ExibeElemento(&elemento);
   }

   putchar('\n'); /* Embelezamento apenas */
}

/****
 *
 * ConstroiTabela(): L arquivo de dados binrio e o armazena
 *                   numa lista indexada SEM ordenao
 *
 * Parmetros:
 *      arq (entrada) - nome do arquivo
 *     *tabela (sada) - tabela que conter os registros lidos
 *
 * Retorno: Nada
 *
 ****/
void ConstroiTabela( const char *arq, tTabela *tabela )
{
   FILE      *stream;
   tRegistro  umRegistro;
   int        erroLeitura = 0,
              erroInsercao = 0;

      /* Abre arquivo em formato binrio para leitura */
   stream = fopen(arq, "rb");

      /* Se o arquivo no puder ser aberto, */
      /* nada mais pode ser feito           */
   ASSEGURA(stream, "Arquivo nao pode ser aberto");

      /* L cada registro do arquivo e o insere na tabela */
   while (1) {
         /* Obtm o registro*/
      fread(&umRegistro, sizeof(umRegistro), 1, stream);

         /* Se ocorreu erro ou o final do arquivo */
         /* foi atingido, encerra o lao          */
      if (feof(stream) || ferror(stream))
         break;

         /* Acrescenta na tabela o registro lido */
      tabela->elementos[tabela->nElementos++] = umRegistro;

         /* Se ocorreu erro durante a insero, encerra lao */
      if (erroInsercao)
         break;
   }

      /* Verifica se ocorreu erro de leitura */
      /* antes de fechar o arquivo           */
   erroLeitura = ferror(stream);

   fclose(stream); /* Fecha o arquivo. */

      /* Se ocorreu erro de leitura aborta o programa */
   ASSEGURA( !erroLeitura,
             "Erro de leitura no arquivo binario" );

      /* Se ocorreu erro de insero aborta o programa */
   ASSEGURA( !erroInsercao,
             "Impossivel inserir um elemento na tabela" );
}

/****
 *
 * AssociaPonteiros(): Associa um array de ponteiros a uma
 *                     tabela indexada
 *
 * Parmetros:
 *      tabela (entrada) - a tabela
 *      nElem (entrada) - nmero de elementos da tabela
 *
 * Retorno: O nmero de linhas do arquivo ou
 *          um valor negativo se ocorrer erro
 *
 ****/
tRegistro **AssociaPonteiros(tRegistro tabela[], int nElem)
{
   tRegistro **ponteiros;
   int         i;

      /* Aloca o array de ponteiros e verifica */
      /* se a alocao foi bem sucedida        */
   ponteiros = malloc(nElem*sizeof(tRegistro *));
   ASSEGURA(ponteiros, "Impossivel alocar array de ponteiros");

      /* Faz cada ponteiro apontar para o  */
      /* elemento correspondente da tabela */
   for (i = 0; i < nElem; ++i)
      ponteiros[i] = tabela + i;

   return ponteiros;
}

/****
 *
 * OrdenaPonteiros(): Ordena ponteiros associados aos elementos
 *                    de uma tabela indexada
 *
 * Parmetros:
 *    ptr (entrada/sada) - array de ponteiros associado  tabela
 *    nElem (entrada) - nmero de elementos do array
 *    F (entrada) - endereo da funo que compara dois elementos
 *                  da tabela
 *
 * Retorno: Nada
 *
 * Observao: Usa-se BubbleSort para facilitar o entendimento, mas
 *             qualquer outro mtodo de ordenao pode ser usado
 *
 ****/
void OrdenaPonteiros( tRegistro *ptr[], int nElem,
                      int (*F)(void *e1, void *e2) )
{
   int i, ordenado = 0;

   while (!ordenado){
      ordenado = 1; /* Supe que a tabela est ordenada */

      for (i = 0; i < nElem - 1; i++)
            /* Compara elementos adjacentes */
         if (F(ptr[i], ptr[i + 1]) > 0) {
               /* Pelo menos um par de elementos */
               /* est fora de ordem             */
            ordenado = 0;

               /* Troca ponteiros para os elementos adjacentes */
            Troca(ptr + i, ptr + i + 1, sizeof(ptr[0]));
         }
   }
}

/****
 *
 * Compara(): Compara dois registros
 *
 * Parmetros:
 *     e1, e2 (entrada) - os dois registros
 *
 * Retorno: O mesmo de strcmp()
 *
 ****/
int Compara(void *e1, void *e2)
{
   tRegistro *p1 = (tRegistro *)e1,
             *p2 = (tRegistro *)e2;

   return strcmp(p1->nome, p2->nome);
}

/****
 *
 * ExibeTabelaOrdenada(): Apresenta na tela, em forma de tabela,
 *                        os elementos de uma lista indexada de
 *                        estruturas do tipo tRegistro
 *
 * Parmetros:
 *     *ptr[] (entrada) - array ordenado de ponteiros para registros
 *     nElem (entrada) - nmero de elementos do array
 *
 * Retorno: Nada
 *
 ****/
void ExibeTabelaOrdenada(tRegistro *ptr[], int nElem)
{
   int i;

      /* Se a tabela estiver vazia, no h o que exibir */
   if (!nElem) {
      printf("\n*** Lista Vazia *** \n");
      return;
   }

      /* Cabealho de apresentao na tela */
   printf( "\n   Nome   \tMatr\tN1\tN2"
           "\n   ====   \t====\t==\t==\n");

      /* Apresenta cada estrutura numa linha separada */
   for (i = 0; i < nElem; ++i)
      ExibeElemento(ptr[i]);

   putchar('\n'); /* Embelezamento apenas */
}

/****
 *
 * main(): Testa ordenao de ponteiros
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   tTabela     tabela; /* Lista que armazena a tabela */
   tRegistro **ponteiros;

      /* A tabela deve ser iniciada! */
   tabela.nElementos = 0;

         /* L contedo do arquivo de dados */
         /* e o armazena na tabela          */
   ConstroiTabela(NOME_ARQ_BIN, &tabela);

   printf("\n****** Tabela Original Antes da Ordenacao *******\n");
   ExibeTabela(&tabela);

   ponteiros = AssociaPonteiros(tabela.elementos,
                                tabela.nElementos);

   OrdenaPonteiros(ponteiros, tabela.nElementos, Compara);

   printf("\n****** Tabela Original Depois da Ordenacao *******\n");
   ExibeTabela(&tabela);

   printf("\n************* Tabela Ordenada **************\n");
   ExibeTabelaOrdenada(ponteiros, tabela.nElementos);
   putchar('\n');

   return 0;
}

/******************* Exemplo de Execuo ******************

****** Tabela Original Antes da Ordenacao *******

   Nome         Matr    N1      N2
   ====         ====    ==      ==

Catarina Parr   1440    4.0     6.0
Catarina Howard 1340    6.0     7.7
Henrique VIII   1029    9.5     9.0
Ana de Cleves   1022    4.5     6.0
Joana Seymour   1017    7.7     8.7
Catarina Aragon 1014    5.5     6.5
Ana Bolena      1012    7.8     8.0

****** Tabela Original Depois da Ordenacao *******

   Nome         Matr    N1      N2
   ====         ====    ==      ==

Catarina Parr   1440    4.0     6.0
Catarina Howard 1340    6.0     7.7
Henrique VIII   1029    9.5     9.0
Ana de Cleves   1022    4.5     6.0
Joana Seymour   1017    7.7     8.7
Catarina Aragon 1014    5.5     6.5
Ana Bolena      1012    7.8     8.0

************* Tabela Ordenada **************

   Nome         Matr    N1      N2
   ====         ====    ==      ==

Ana Bolena      1012    7.8     8.0
Ana de Cleves   1022    4.5     6.0
Catarina Aragon 1014    5.5     6.5
Catarina Howard 1340    6.0     7.7
Catarina Parr   1440    4.0     6.0
Henrique VIII   1029    9.5     9.0
Joana Seymour   1017    7.7     8.7

**********************************************************/

