/****
 *
 * Exemplo: Intercalao Multplice de Arrays
 *
 * Seo: 12.3.3
 *
 ****/

/*********************** Includes *************************/

#include <limits.h> /* Limites de inteiros      */
#include <string.h> /* memmove()                */
#include <stdio.h>  /* Funes de Entrada/Sada */
#include <stdlib.h> /* Alocao dinmica        */

/***************** Constantes Simblicas ******************/

   /* Nmero de elementos dos arrays que sero intercalados */
#define N_ELEMENTOS 4

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/****************** Definies de Tipos *******************/

typedef int tElemArray; /* Tipo de cada elemento dos arrays */

   /* Tipo de n de um heap */
typedef struct {
           tElemArray elemento; /* Elemento do array */
           int        i;      /* ndice do elemento em seu array */
           int        iArray; /* ndice do array ao qual o elemento pertence */
        } tNoHeapIM_Arr;

   /* Tipo que representa um heap */
typedef struct {
           tNoHeapIM_Arr *itens;  /* Array de elementos */
           int            nItens; /* Nmero de elementos do heap */
        } tHeapIM_Arr;

   /* Tipo de ponteiro para funo de comparao que compara elementos do array */
typedef int (*tFCompara) (const void *, const void *);

/******************* Aluses de Funes *******************/

extern void OrdenaHeapIM_Arr(tHeapIM_Arr *heap, int indice, tFCompara Compara);
extern tHeapIM_Arr *IniciaHeapIM_Arr(tHeapIM_Arr *heap, int nElementos);
extern void DestroiHeapIM_Arr(tHeapIM_Arr *heap);
extern tNoHeapIM_Arr ObtemMinimoIM_Arr(const tHeapIM_Arr *heap);
extern void SubstituiMinimoIM_Arr( tHeapIM_Arr *heap, const tNoHeapIM_Arr *item, tFCompara Compara );
extern tNoHeapIM_Arr ReduzHeapIM_Arr(tHeapIM_Arr *heap);
extern tElemArray *IntercalaNArrays( tElemArray ar[][N_ELEMENTOS], int k );
extern void ExibeArray(const int *array, int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * ComparaInts(): Compara dois elementos do tipo int
 *
 * Parmetros:
 *      e1, e2 (entrada) - endereos dos elementos que sero comparados
 *
 * Retorno: Nada
 *
 ****/
int ComparaInts(const void *e1, const void *e2)
{
   ASSEGURA(e1 && e2, "Elemento nulo recebido por ComparaInts()");

   return *(int *)e1 - *(int *)e2;
}

/****
 *
 * TrocaGenerica(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void TrocaGenerica(void *var1, void *var2, size_t tam)
{
   void *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, var1, tam);
   memmove(var1, var2, tam);
   memmove(var2, p, tam);

   free(p);
}

/****
 *
 * OrdenaHeapIM_Arr(): Restaura a propriedade de ordenao de um
 *               heap de mnimo aps ele ter sido alterado
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap
 *      indice (entrada) - ndice do elemento do heap a partir
 *                         do qual ser efetuada a restaurao
 *      Compara (entrada) - ponteiro para uma funo que compara
 *                          dois elementos de array armazenados no heap
 *
 * Retorno: Nada
 *
 ****/
void OrdenaHeapIM_Arr(tHeapIM_Arr *heap, int indice, tFCompara Compara)
{
   int iEsq = 2*indice + 1, /* ndice do filho esquerdo */
       iDir = 2*indice + 2, /* ndice do filho direito  */
       iMenor = indice; /* ndice do menor elemento entre pai e filhos */

      /* Compara o filho esquerdo com seu pai */
   if ( iEsq < heap->nItens &&
        Compara(&heap->itens[iEsq].elemento,
                &heap->itens[indice].elemento) < 0 )
         /* O filho esquerdo  menor do que o pai */
      iMenor = iEsq;

      /* Compara o filho direito com o   */
      /* menor entre seu pai e seu irmo */
   if ( iDir < heap->nItens &&
        Compara(&heap->itens[iDir].elemento,
                &heap->itens[iMenor].elemento) < 0 )
      iMenor = iDir;

      /* Se o n no for menor do que seus filhos, */
      /* troca-o de posio com o menor deles      */
   if (iMenor != indice) {
         /* O pai e o menor filho trocam de posio */
      TrocaGenerica( &heap->itens[indice], &heap->itens[iMenor],
             sizeof(tNoHeapIM_Arr) );

         /* Reordena o heap a partir do filho que se tornou pai */
      OrdenaHeapIM_Arr(heap, iMenor, Compara);
   }
}

/****
 *
 * IniciaHeapIM_Arr(): Inicia um heap
 *
 * Parmetros:
 *      *heap (sada) - heap que ser iniciado
 *      nElementos (entrada) - nmero de elementos do heap
 *
 * Retorno: O primeiro parmetro
 *
 ****/
tHeapIM_Arr *IniciaHeapIM_Arr(tHeapIM_Arr *heap, int nElementos)
{
   heap->nItens = nElementos;
   heap->itens = calloc(nElementos, sizeof(tNoHeapIM_Arr));

   ASSEGURA(heap->itens, "Impossivel alocar heap");

   return heap;
}

/****
 *
 * DestroiHeapIM_Arr(): Libera o espao ocupado por um heap
 *
 * Parmetros: heap (entrada/sada) - o heap
 *
 * Retorno: Nada
 *
 ****/
void DestroiHeapIM_Arr(tHeapIM_Arr *heap)
{
      /* Libera o espao ocupado pelos ns do heap */
   free(heap->itens);
   heap->nItens = 0;
}

/****
 *
 * ObtemMinimoIM_Arr(): Obtm o menor elemento de um heap de mnimo
 *
 * Parmetros:
 *      heap (entrada) - ponteiro para o heap
 *
 * Retorno: O menor elemento do heap recebido como parmetro
 *
 ****/
tNoHeapIM_Arr ObtemMinimoIM_Arr(const tHeapIM_Arr *heap)
{
   return heap->itens[0]; /* O menor elemento de um heap de mnimo se acha na raiz */
}

/****
 *
 * SubstituiMinimoIM_Arr(): Substitui o menor elemento de um heap de mnimo
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap que ter
 *                               seu menor elemento substitudo
 *      item (entrada) - ponteiro para o contedo do elemento que
 *                       substituir o menor elemento do heap
 *      Compara (entrada) - ponteiro para uma funo que compara
 *                          dois elementos de array armazenados
 *                          no heap
 *
 * Retorno: Nada
 *
 ****/
void SubstituiMinimoIM_Arr( tHeapIM_Arr *heap, const tNoHeapIM_Arr *item,
                            tFCompara Compara )
{
   heap->itens[0] = *item; /* Efetua a substituio */

      /* Talvez a substituio da raiz faa com que o heap */
      /* deixe de satisfazer sua propriedade de ordenao  */
   OrdenaHeapIM_Arr(heap, 0, Compara);
}

/****
 *
 * ReduzHeapIM_Arr(): Reduz o tamanho de um heap e retorna seu ltimo elemento
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap
 *
 * Retorno: O ltimo elemento do heap recebido como parmetro
 *
 ****/
tNoHeapIM_Arr ReduzHeapIM_Arr(tHeapIM_Arr *heap)
{
   return heap->itens[--heap->nItens];
}

/****
 *
 * IntercalaNArrays(): Intercala k arrays ordenados de n elementos
 *                     cada usando um heap de mnimo
 *
 * Parmetros:
 *      ar[][] (entrada) - os arrays ordenados
 *      k (entrada) - o nmero de arrays ordenados
 *
 * Retorno: Endereo do array que contm o resultado da intercalao
 *
 ****/
tElemArray *IntercalaNArrays(tElemArray ar[][N_ELEMENTOS], int k)
{
   tElemArray   *resultado;
   tHeapIM_Arr   heap;
   int           i;
   tNoHeapIM_Arr raiz;

      /* Aloca espao para o array que conter o resultado */
   resultado = calloc(N_ELEMENTOS*k, sizeof(tElemArray));
   ASSEGURA(resultado, "Impossivel alocar array");

   IniciaHeapIM_Arr(&heap, k); /* Inicia o heap */

      /* Associa o primeiro elemento de cada array a um n do heap */
   for (i = 0; i < k; i++) {
      heap.itens[i].elemento = ar[i][0];
      heap.itens[i].i = 0;
      heap.itens[i].iArray = i;
   }

      /* Ordena o heap */
   for (i = (k - 1)/2; i >= 0; --i)
      OrdenaHeapIM_Arr(&heap, i, ComparaInts);

      /* Obtm o menor elemento do heap e o substitui pelo prximo elemento */
      /* do array que o contm at que todos os arrays estejam esgotados    */
   for (i = 0; i < N_ELEMENTOS*k; ++i) {
      raiz = ObtemMinimoIM_Arr(&heap); /* Obtm o menor valor armazenado no heap */

         /* Armazena o menor valor no array que */
         /* conter o resultado da intercalao */
      resultado[i] = raiz.elemento;

         /* O elemento que substituir a raiz pertence ao mesmo array ao */
         /* qual pertence raiz, a no ser que esse array esteja esgotado */
      if (++raiz.i < N_ELEMENTOS)
            /* Substitui a raiz por um novo elemento */
            /* do array ao qual ela pertencia        */
         raiz.elemento = ar[raiz.iArray][raiz.i];
      else
            /* Os elementos desse array esto esgotados. A nova raiz ser o   */
            /* ltimo elemento do heap, que passar a ter um elemento a menos */
         raiz = ReduzHeapIM_Arr(&heap);

         /* Substitui a raiz do heap pelo novo elemento */
      SubstituiMinimoIM_Arr(&heap, &raiz, ComparaInts);
   }

   return resultado; /* Retorna o array resultante da intercalao */
}

/****
 *
 * ExibeArray(): Apresenta na tela um array recebida como parmetro
 *
 * Parmetros:
 *      array (entrada) - o array cujos elementos sero exibidos
 *      nElementos (entrada) - nmero mnimo de elementos no array
 *
 * Retorno: Nada
 *
 ****/
void ExibeArray(const int *array, int nElementos)
{
   int i;

   printf("\n>>> Numero de elementos: %d", nElementos);

   if (nElementos <= 0)
      return;

   printf("\n>>> {");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", array[i]);

   printf("%d}", array[nElementos - 1]);

   putchar('\n');
}

/****
 *
 * main(): Intercala mltiplos arrays ordenado usando um heap
 *         de mnimo
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int arrays[][N_ELEMENTOS] =  { {2, 6, 12, 34},
                                  {13, 17, 32, 51},
                                  {1, 9, 20, 1000},
                                  {23, 34, 90, 2000}
                                };
   int k = sizeof(arrays)/sizeof(arrays[0]);/* Nmero de arrays */
                                            /* ordenados        */

   int *arOrdenado; /* Ponteiro para o array que conter */
                    /* o resultado da operao           */

      /* Efetua a intercalao */
   arOrdenado = IntercalaNArrays(arrays, k);

      /* Apresenta o resultado na tela */
   printf("\n\t>>> Array ordenado <<<\n");
   ExibeArray(arOrdenado, N_ELEMENTOS*k);

   return 0;
}

/****************** Resultado do Programa *****************

        >>> Array ordenado <<<

>>> Numero de elementos: 16
>>> {1, 2, 6, 9, 12, 13, 17, 20, 23, 32, 34, 34, 51, 90, 1000, 2000}

**********************************************************/
