/****
 *
 * Arquivo Macros.h: Macros utilizadas pelo programa de bulkloading
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 21/09/2017
 * ltima modificao: 10/05/2018
 *
 ****/

#ifndef _Macros_H_  /* Previne incluso mltipla deste arquivo */
#define _Macros_H_

/*********************** Includes *************************/

#include <stdlib.h> /* Funo exit()   */
#include <stdio.h>  /* Funo printf() */

/****************** Definies de Macros ******************/

   /****                                ****/
   /* As macros de tratamento de excees */
   /* esto no aquivo ASSEGURA.h          */
   /****                               ****/

   /* Maior de dois nmeros */
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

   /* Menor de dois valores */
#define MIN(a, b) ((a) <= (b)) ? (a) : (b)

/****************** Constantes Simblicas *****************/

/* >>>>>>>>>>>>>>>>>>>>>> Arquivos <<<<<<<<<<<<<<<<<<<<<< */

   /* Nome do arquivo de texto contendo os registros */
#define NOME_ARQ_REGS "../CensoMEC2012.txt"

   /* Nome do arquivo binrio de registros (Teste) */
#define NOME_ARQ_BIN "../CensoMEC2012.bin"

      /* Nome do arquivo contendo as sries */
#define NOME_ARQ_SERIES "../CensoMEC2012_TMP.bin"

      /* Nome do arquivo de dados ordenado */
#define NOME_ARQ_ORD "../CensoMEC2012_Ord.bin"

/* Nome do arquivo contendo a rvore */
#define NOME_ARQ_ARV "ArvoreBM.dat"

   /* Nome do arquivo de texto que contm as chaves ordenadas */
#define NOME_ARQ_CHAVES "ChavesOrdenadasBM.txt"

   /* Nome do arquivo de texto que contm o */
   /* resultado de uma busca de intervalo   */
#define NOME_ARQ_INTER "BuscaIntervaloBM.txt"

   /* Nome inicial do arquivo de texto contendo os ns da rvore */
#define NOME_ARQ_NOS "NosArvoreBM"

   /* Nome do arquivo de texto contendo registros em formato legvel */
#define NOME_ARQ_DEBUG   "RegistrosDebug.txt"

/* >>>>>>>>>>>>>>>>>>>>> Registros <<<<<<<<<<<<<<<<<<<<< */

#define TAM_CODIGO_ALUNO  12 /* Nmero de dgitos num */
                             /* cdigo de aluno       */
#define TAM_ALUNO_CURSO   8 /* Nmero de dgitos num         */
                            /* cdigo de aluno/curso (chave) */
#define MAX_NOME_IES     200 /* Tamanho mximo de */
                             /* um nome de IES    */
#define MAX_NOME_CURSO   200 /* Tamanho mximo de */
                             /* um nome de curso  */
#define MAX_UF_NASCIMENTO 30 /* Tamanho mximo   */
                             /* de um nome de UF */

   /* Tamanho de um par chave/ndice */
#define TAM_PAR sizeof(tChaveIndice)

/* >>>>>>>>>>>>> Compilao Condidicional >>>>>>>>>>>>>> */

#define DEPURACAO 0

/* >>>>>>>>>>>>>>>>>>>>> Ordenao <<<<<<<<<<<<<<<<<<<<< */

   /* Nmero mximo de pares em cada srie */
#define MAX_PARES 80000

/* >>>>>>>>>>>>>>>>>>>>> rvore B+ <<<<<<<<<<<<<<<<<<<<< */

   /* Posio nula (invlida) em arquivo */
#define POSICAO_NULA  -1

#define TB   4096 /* Tamanho do bloco lido/escrito  */
#define TC   sizeof(tChave) /* Tamanho de uma chave */
#define TCI  sizeof(tChaveIndice)  /* Tamanho de um    */
                                   /* par chave/ndice */
#define TI   sizeof(int) /* Tamanho de um filho e do inteiro */
                         /* que representa o grau do n      */
#define TT   sizeof(tTipoDoNo) /* Tamanho de uma constante */
                               /* de enumerao            */

   /* Correo devido a preenchimento (tentativa e erro) */
#define CORRECAO 5

   /* Clculo do grau da rvore */
#define G   ((TB - CORRECAO - TT - TI + TC)/(TC + TI))

   /* Metade do grau da rvore */
#define TG   G%2 ? G/2 + 1 : G/2

   /* Nmero mximo de pares chave/ndice em cada folha */
#define F   ((TB - CORRECAO - TT - 2*TI)/TCI)

   /* Metade do mximo de pares chave/ndice numa folha */
#define TF   F%2 ? F/2 + 1 : F/2

#endif
